/*
 * aarbology.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>
#include <vector>

#include <factory/XmlDataFactory.hpp>
#include <exception/AlibException.h>
#include <tree/Tree.h>
#include <container/Container.h>
#include <container/ObjectsSet.h>

#include <arbology/exact/ExactSubtreeMatch.h>
#include <arbology/exact/ExactSubtreeMatchingAutomaton.h>
#include <chrono>

int main(int argc, char* argv[]) {
	try {
		TCLAP::CmdLine cmd("Arbology algorithm access binary", ' ', "0.01");

		std::vector<std::string> allowed;
		allowed.push_back("exactSubtreeMatch");
		allowed.push_back("exactSubtreeMatchingAutomaton");
		TCLAP::ValuesConstraint<std::string> allowedVals( allowed );

		TCLAP::ValueArg<std::string> algorithm(	"a",	"algorithm",	"Execute algorithm",		false,	"exactSubtreeMatch",	&allowedVals);
		cmd.add(algorithm);

		TCLAP::MultiArg<std::string> subject(	"s",	"subject",	"Subject tree from file",	false,	"file");
		cmd.add( subject );

		TCLAP::MultiArg<std::string> pattern(	"p",	"pattern",	"Pattern tree from file",	false,	"file");
		cmd.add( pattern );

		TCLAP::SwitchArg measure(		"m",	"measure",	"Measure times",		false);
		cmd.add( measure );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc,argv);

		int needPattern = 0;
		int needSubject = 0;
		if( algorithm.getValue() == "exactSubtreeMatch") {
			needPattern = needSubject = 1;
		} else if( algorithm.getValue() == "exactSubtreeMatchingAutomaton") {
			needPattern = 1;
		} else {
		}

		std::chrono::measurements::start("Overal", std::chrono::measurements::Type::OVERALL);
		std::chrono::measurements::start("Input read", std::chrono::measurements::Type::AUXILARY);

		std::deque<std::deque<sax::Token>> subjectTokens;
		if(subject.isSet()) {
			for(const std::string& fileName : subject.getValue()) {
				std::deque<sax::Token> tmp;
				if(fileName == "-") {
					sax::SaxParseInterface::parseStdin(tmp);
				} else {
					sax::SaxParseInterface::parseFile(fileName, tmp);
				}
				subjectTokens.emplace_back(std::move(tmp));
			}
		} else if(needSubject) {
			std::deque<sax::Token> tmp;
			sax::SaxParseInterface::parseStdin(tmp);
			subjectTokens.emplace_back(std::move(tmp));
		}

		std::deque<std::deque<sax::Token>> patternTokens;
		if(pattern.isSet()) {
			for(const std::string& fileName : pattern.getValue()) {
				std::deque<sax::Token> tmp;
				if(fileName == "-") {
					sax::SaxParseInterface::parseStdin(tmp);
				} else {
					sax::SaxParseInterface::parseFile(fileName, tmp);
				}
				patternTokens.emplace_back(std::move(tmp));
			}
		} else if(needPattern) {
			std::deque<sax::Token> tmp;
			sax::SaxParseInterface::parseStdin(tmp);
			patternTokens.emplace_back(std::move(tmp));
		}

		if( algorithm.getValue() == "exactSubtreeMatch") {
			tree::Tree subject = alib::XmlDataFactory::fromTokens<tree::Tree>(subjectTokens.front());
			tree::Tree pattern = alib::XmlDataFactory::fromTokens<tree::Tree>(patternTokens.front());

			std::chrono::measurements::end();
			std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

			std::set<unsigned> res = arbology::exact::ExactSubtreeMatch::match(subject, pattern);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout(res);
		} else if( algorithm.getValue() == "exactSubtreeMatchingAutomaton") {
			tree::Tree pattern = alib::XmlDataFactory::fromTokens<tree::Tree>(patternTokens.front());

			std::chrono::measurements::end();
			std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

			automaton::Automaton res = arbology::exact::ExactSubtreeMatchingAutomaton::construct(pattern);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout(res);
		} else {
			throw exception::AlibException( "Invalid algorithm" );
			return 1;
		}

		std::chrono::measurements::end();
		std::chrono::measurements::end();

		if(measure.getValue()) std::clog << std::chrono::measurements::results() << std::endl;

		return 0;
	} catch( const exception::AlibException & exception ) {
		alib::XmlDataFactory::toStdout( exception );
		return 1;
	} catch(const TCLAP::ArgException& exception) {
		std::cout << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch(...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
