/*
 * AllEpsilonClosure.cpp
 *
 *  Created on: 12. 4. 2015
 *	  Author: Jan Travnicek
 */

#include "AllEpsilonClosure.h"

#include <exception/AlibException.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <deque>
#include <iostream>

#include "../../regexp/properties/RegExpEpsilon.h"


namespace automaton {

namespace properties {

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::EpsilonNFA & fsm) {
	std::deque<std::map<automaton::State, std::set<automaton::State>>> Qi;

	Qi.push_back({});
	for(const automaton::State& state : fsm.getStates())
		Qi.at(0)[state].insert(state);

	int i = 1;
	while(true) {
		Qi.push_back( Qi.at( i - 1 ) );

		for( const std::pair<automaton::State, std::set<automaton::State>> & p : Qi.at( i - 1 ) )
			for(const automaton::State& pclosure : p.second )
				for(const auto& transition : fsm.getEpsilonTransitionsFromState(pclosure)) {
					Qi.at( i )[p.first].insert(transition.second.begin(), transition.second.end());
				}

		if( Qi.at( i ) == Qi.at( i - 1 ) )
			break;

		i = i + 1;
	};

	return Qi.at(i);
}

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::MultiInitialStateNFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> closure;
	for(const automaton::State& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::NFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> closure;
	for(const automaton::State& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::DFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> closure;
	for(const automaton::State& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::ExtendedNFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> res;
	std::map<automaton::State, std::set<automaton::State>> step;

	for(const std::pair<std::pair<State, regexp::RegExp>, std::set<State> >& transition : fsm.getTransitions() )
		if( regexp::properties::RegExpEpsilon::languageContainsEpsilon( transition.first.second ) )
			step[transition.first.first].insert(transition.second.begin(), transition.second.end());

	for(const automaton::State& state : fsm.getStates())
		step[state].insert(state);

	do {
		res = step;

		for(const std::pair<automaton::State, std::set<automaton::State>>& item : res)
			for(const automaton::State& to : item.second)
				step[item.first].insert(res[to].begin(), res[to].end());
	} while(res != step);

	return res;
}

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::CompactNFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> res;
	std::map<automaton::State, std::set<automaton::State>> step;

	for(const std::pair<std::pair<State, string::LinearString>, std::set<State> >& transition : fsm.getTransitions() )
		if( transition.first.second.getContent().size() == 0 )
			step[transition.first.first].insert(transition.second.begin(), transition.second.end());

	for(const automaton::State& state : fsm.getStates())
		step[state].insert(state);

	do {
		res = step;

		for(const std::pair<automaton::State, std::set<automaton::State>>& item : res)
			for(const automaton::State& to : item.second)
				step[item.first].insert(res[to].begin(), res[to].end());
	} while(res != step);

	return res;
}

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure(const Automaton& automaton) {
	std::map<automaton::State, std::set<automaton::State>> out;
	automaton.getData().Accept((void*) &out, AllEpsilonClosure::ALL_EPSILON_CLOSURE);
	return std::move(out);
}

void AllEpsilonClosure::Visit(void* data, const EpsilonNFA& automaton) const {
	std::map<automaton::State, std::set<automaton::State>> & out = *((std::map<automaton::State, std::set<automaton::State>>*) data);
	out = std::move(this->allEpsilonClosure(automaton));
}

void AllEpsilonClosure::Visit(void* data, const MultiInitialStateNFA& automaton) const {
	std::map<automaton::State, std::set<automaton::State>> & out = *((std::map<automaton::State, std::set<automaton::State>>*) data);
	out = std::move(this->allEpsilonClosure(automaton));
}

void AllEpsilonClosure::Visit(void* data, const NFA& automaton) const {
	std::map<automaton::State, std::set<automaton::State>> & out = *((std::map<automaton::State, std::set<automaton::State>>*) data);
	out = std::move(this->allEpsilonClosure(automaton));
}

void AllEpsilonClosure::Visit(void* data, const DFA& automaton) const {
	std::map<automaton::State, std::set<automaton::State>> & out = *((std::map<automaton::State, std::set<automaton::State>>*) data);
	out = std::move(this->allEpsilonClosure(automaton));
}

void AllEpsilonClosure::Visit(void* data, const ExtendedNFA& automaton) const {
	std::map<automaton::State, std::set<automaton::State>> & out = *((std::map<automaton::State, std::set<automaton::State>>*) data);
	out = std::move(this->allEpsilonClosure(automaton));
}

void AllEpsilonClosure::Visit(void* data, const CompactNFA& automaton) const {
	std::map<automaton::State, std::set<automaton::State>> & out = *((std::map<automaton::State, std::set<automaton::State>>*) data);
	out = std::move(this->allEpsilonClosure(automaton));
}

void AllEpsilonClosure::Visit(void*, const DFTA&) const {
	throw exception::AlibException("Unsupported automaton type DFTA");
}

void AllEpsilonClosure::Visit(void*, const NFTA&) const {
	throw exception::AlibException("Unsupported automaton type NFTA");
}

void AllEpsilonClosure::Visit(void*, const DPDA&) const {
	throw exception::AlibException("Unsupported automaton type DPDA");
}

void AllEpsilonClosure::Visit(void*, const SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void AllEpsilonClosure::Visit(void*, const InputDrivenDPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenDPDA");
}

void AllEpsilonClosure::Visit(void*, const InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void AllEpsilonClosure::Visit(void*, const VisiblyPushdownDPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void AllEpsilonClosure::Visit(void*, const VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void AllEpsilonClosure::Visit(void*, const RealTimeHeightDeterministicDPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void AllEpsilonClosure::Visit(void*, const RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void AllEpsilonClosure::Visit(void*, const NPDA&) const {
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void AllEpsilonClosure::Visit(void*, const SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void AllEpsilonClosure::Visit(void*, const OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

const AllEpsilonClosure AllEpsilonClosure::ALL_EPSILON_CLOSURE;

} /* namespace properties */

} /* namespace automaton */

