/*
 * EpsilonClosure.cpp
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "EpsilonClosure.h"

#include <exception/AlibException.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <queue>

#include "../../regexp/properties/RegExpEpsilon.h"

namespace automaton {

namespace properties {

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::EpsilonNFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	std::queue<automaton::State> queue;
	std::map<automaton::State, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) )
	{
		automaton::State p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		auto tos = fsm.getTransitions( ).find(std::make_pair(p, std::variant<string::Epsilon, alphabet::Symbol>(string::Epsilon::EPSILON) ) );
		if(tos == fsm.getTransitions().end()) continue;

		for( const auto & to : tos->second )
			if( visited [ to ] == false )
				queue.push( to );
	}

	return closure;
}

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::MultiInitialStateNFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	closure.insert(q);
	return closure;
}

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::NFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	closure.insert(q);
	return closure;
}

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::DFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	closure.insert(q);
	return closure;
}

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::ExtendedNFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	std::queue<automaton::State> queue;
	std::map<automaton::State, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) )
	{
		automaton::State p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		for( const auto & transition : fsm.getTransitionsFromState( p ) )
			if( regexp::properties::RegExpEpsilon::languageContainsEpsilon( transition.first.second ) )
				for (const auto & to : transition.second )
					if( visited [ to ] == false )
						queue.push( to );
	}

	return closure;
}

std::set<automaton::State> EpsilonClosure::epsilonClosure( const automaton::CompactNFA & fsm, const automaton::State & q ) {
	if(! fsm.getStates().count(q) ) throw exception::AlibException("State is not in the automaton");

	std::set<automaton::State> closure;
	std::queue<automaton::State> queue;
	std::map<automaton::State, bool> visited;

	for( const auto & p : fsm.getStates( ) )
		visited[ p ] = false;

	queue.push( q );
	while( ! queue.empty( ) )
	{
		automaton::State p = queue.front( );
		queue.pop( );
		visited[ p ] = true;
		closure.insert( p );

		for( const auto & transition : fsm.getTransitionsFromState( p ) )
			if( transition.first.second.getContent().size() == 0 )
				for (const auto & to : transition.second )
					if( visited [ to ] == false )
						queue.push( to );
	}

	return closure;
}

std::set<automaton::State> EpsilonClosure::epsilonClosure(const Automaton& automaton, const automaton::State& q) {
	std::pair<automaton::State, std::set<automaton::State>> out = std::make_pair(q, std::set<automaton::State> {});
	automaton.getData().Accept((void*) &out, EpsilonClosure::EPSILON_CLOSURE);
	return std::move(out.second);
}

void EpsilonClosure::Visit(void* data, const EpsilonNFA& automaton) const {
	std::pair<automaton::State, std::set<automaton::State>> & out = *((std::pair<automaton::State, std::set<automaton::State>>*) data);
	out.second = std::move(this->epsilonClosure(automaton, out.first));
}

void EpsilonClosure::Visit(void* data, const MultiInitialStateNFA& automaton) const {
	std::pair<automaton::State, std::set<automaton::State>> & out = *((std::pair<automaton::State, std::set<automaton::State>>*) data);
	out.second = std::move(this->epsilonClosure(automaton, out.first));
}

void EpsilonClosure::Visit(void* data, const NFA& automaton) const {
	std::pair<automaton::State, std::set<automaton::State>> & out = *((std::pair<automaton::State, std::set<automaton::State>>*) data);
	out.second = std::move(this->epsilonClosure(automaton, out.first));
}

void EpsilonClosure::Visit(void* data, const DFA& automaton) const {
	std::pair<automaton::State, std::set<automaton::State>> & out = *((std::pair<automaton::State, std::set<automaton::State>>*) data);
	out.second = std::move(this->epsilonClosure(automaton, out.first));
}

void EpsilonClosure::Visit(void* data, const ExtendedNFA& automaton) const {
	std::pair<automaton::State, std::set<automaton::State>> & out = *((std::pair<automaton::State, std::set<automaton::State>>*) data);
	out.second = std::move(this->epsilonClosure(automaton, out.first));
}

void EpsilonClosure::Visit(void* data, const CompactNFA& automaton) const {
	std::pair<automaton::State, std::set<automaton::State>> & out = *((std::pair<automaton::State, std::set<automaton::State>>*) data);
	out.second = std::move(this->epsilonClosure(automaton, out.first));
}

void EpsilonClosure::Visit(void*, const DPDA&) const {
	throw exception::AlibException("Unsupported automaton type DPDA");
}

void EpsilonClosure::Visit(void*, const SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void EpsilonClosure::Visit(void*, const InputDrivenDPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenDPDA");
}

void EpsilonClosure::Visit(void*, const InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void EpsilonClosure::Visit(void*, const VisiblyPushdownDPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void EpsilonClosure::Visit(void*, const VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void EpsilonClosure::Visit(void*, const RealTimeHeightDeterministicDPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void EpsilonClosure::Visit(void*, const RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void EpsilonClosure::Visit(void*, const NPDA&) const {
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void EpsilonClosure::Visit(void*, const SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void EpsilonClosure::Visit(void*, const OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

void EpsilonClosure::Visit(void*, const DFTA&) const {
	throw exception::AlibException("Unsupported automaton type DFTA");
}

void EpsilonClosure::Visit(void*, const NFTA&) const {
	throw exception::AlibException("Unsupported automaton type NFTA");
}

const EpsilonClosure EpsilonClosure::EPSILON_CLOSURE;

} /* namespace properties */

} /* namespace automaton */

