/*
 * Occurrences.cpp
 *
 *  Created on: 9. 2. 2014
 *	  Author: Jan Travnicek
 */

#include "Occurrences.h"
#include <exception/AlibException.h>
#include <automaton/FSM/DFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/TA/DFTA.h>

#include <deque>

namespace automaton {

namespace run {

std::set<unsigned> Occurrences::occurrences(const automaton::Automaton& automaton, const alib::Object& object) {
	std::pair<const alib::Object*, std::set<unsigned>> data(&object, {});
	automaton.getData().Accept((void*) &data, Occurrences::OCCURRENCES);
	return data.second;
}

std::set<unsigned> Occurrences::occurrences(const automaton::Automaton& automaton, const string::LinearString& string) {
	alib::Object obj(string);
	std::pair<const alib::Object*, std::set<unsigned>> data(&obj, {});
	automaton.getData().Accept((void*) &data, Occurrences::OCCURRENCES);
	return data.second;
}

std::set<unsigned> Occurrences::occurrences(const automaton::Automaton& automaton, const tree::RankedTree& tree) {
	alib::Object obj(tree);
	std::pair<const alib::Object*, std::set<unsigned>> data(&obj, {});
	automaton.getData().Accept((void*) &data, Occurrences::OCCURRENCES);
	return data.second;
}

std::set<unsigned> Occurrences::occurrences(const automaton::DFA& automaton, const string::LinearString& string) {
	automaton::State state = automaton.getInitialState();
	unsigned i = 0;
	std::set<unsigned> occ;

	for(const alphabet::Symbol& symbol : string.getContent()) {
		if(automaton.getFinalStates().count(state))
			occ.insert(i);

		auto transitions = automaton.getTransitionsFromState(state);
		auto next = transitions.find(std::make_pair(state, symbol));
		if(next == transitions.end()) { //makes sence to reset the automaton when there is "unknown" symbol on the input
			state = automaton.getInitialState();
		} else {
			state = next->second;
		}

		i++;
	}

	if(automaton.getFinalStates().count(state))
		occ.insert(i);

	return occ;
}

State Occurrences::occurences(const automaton::DFTA & automaton, const tree::RankedNode & node, std::set<unsigned> & occ, unsigned & i) {
	std::vector<State> states;
	unsigned tmp = i;
	i++;
	states.reserve(node.getSymbol().getRank().getData());
	bool fail = false;
	for (const auto & child : node.getChildren()) {
		State state = occurences(automaton, *child, occ, i);
		if (state == FAILSTATE) fail = true;
		else if (!fail) states.push_back(state);
	}
	if (fail) return FAILSTATE;
	const auto & it = automaton.getTransitions().find(std::make_pair(node.getSymbol(), states));
	if (it == automaton.getTransitions().end()) return FAILSTATE;
	State state = it -> second;
	if (automaton.getFinalStates().count(state)) occ.insert(tmp);
	return state;
}

std::set<unsigned> Occurrences::occurrences(const automaton::DFTA& automaton, const tree::RankedTree & tree) {
	std::set<unsigned> occ;
	unsigned i = 0;
	occurences(automaton, tree.getRoot(), occ, i);
	return occ;
}

std::set<unsigned> Occurrences::occurrences(const automaton::DPDA& automaton, const string::LinearString& string) {
	automaton::State state = automaton.getInitialState();
	std::deque<alphabet::Symbol> pushdownStore;
	unsigned i = 0;
	std::set<unsigned> occ;

	for(const alphabet::Symbol& symbol : string.getContent()) {
		if(automaton.getFinalStates().count(state))
			occ.insert(i);

		auto transitions = automaton.getTransitionsFromState(state);
		auto transition = transitions.begin();
		unsigned pushdownStoreSize = pushdownStore.size();
		for(auto transition = transitions.begin(); transition != transitions.end(); transition++) {
			if(std::get<1>(transition->first) != symbol) continue;

			const std::vector<alphabet::Symbol> & pop = std::get<2>(transition->first);
			unsigned popSize = pop.size();
			if(pushdownStoreSize < popSize) continue;
			bool sign = true;
			for(unsigned i = 1; i <= popSize; i++) {
				if(pop[popSize - i] != pushdownStore[pushdownStoreSize - i]) {
					sign = false;
					break;
				}
			}
			if(!sign) continue;

			break;
		}
		if(transition == transitions.end()) return {};
		for(auto pop : std::get<2>(transition->first)) pushdownStore.pop_back();

		state = transition->second.first;
		for(const alphabet::Symbol& push : transition->second.second) pushdownStore.push_back(push);

		i++;
	}

	if(automaton.getFinalStates().count(state))
		occ.insert(i);

	if(pushdownStore.empty()) {
		return occ;
	} else {
		return {};
	}
}

void Occurrences::Visit(void*, const automaton::EpsilonNFA& ) const {
	throw exception::AlibException("Unsupported automaton type EpsilonNFA");
}

void Occurrences::Visit(void*, const automaton::MultiInitialStateNFA& ) const {
	throw exception::AlibException("Unsupported automaton type MultiInitialStateNFA");
}

void Occurrences::Visit(void*, const automaton::NFA& ) const {
	throw exception::AlibException("NFA");
}

void Occurrences::Visit(void* data, const automaton::DFA& automaton) const {
	std::pair<const alib::Object*, std::set<unsigned>> & res = *((std::pair<const alib::Object*, std::set<unsigned>>*) data);
	res.second = this->occurrences(automaton, static_cast<const string::LinearString&>(res.first->getData()));
}

void Occurrences::Visit(void*, const automaton::ExtendedNFA& ) const {
	throw exception::AlibException("Unsupported automaton type ExtendedNFA");
}

void Occurrences::Visit(void*, const automaton::CompactNFA& ) const {
	throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void Occurrences::Visit(void* data, const DPDA& automaton) const {
	std::pair<const alib::Object*, std::set<unsigned>> & res = *((std::pair<const alib::Object*, std::set<unsigned>>*) data);
	res.second = this->occurrences(automaton, static_cast<const string::LinearString&>(res.first->getData()));
}

void Occurrences::Visit(void*, const SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void Occurrences::Visit(void*, const InputDrivenDPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenDPDA");
}

void Occurrences::Visit(void*, const InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void Occurrences::Visit(void*, const VisiblyPushdownDPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void Occurrences::Visit(void*, const VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void Occurrences::Visit(void*, const RealTimeHeightDeterministicDPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void Occurrences::Visit(void*, const RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void Occurrences::Visit(void*, const NPDA&) const {
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void Occurrences::Visit(void*, const SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void Occurrences::Visit(void*, const OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

void Occurrences::Visit(void* data, const automaton::DFTA& automaton) const {
	std::pair<const alib::Object*, std::set<unsigned>> & res = *((std::pair<const alib::Object*, std::set<unsigned>>*) data);
	res.second = this->occurrences(automaton, static_cast<const tree::RankedTree&>(res.first->getData()));
}

void Occurrences::Visit(void*, const NFTA&) const {
	throw exception::AlibException("Unsupported automaton type NFTA");
}

const Occurrences Occurrences::OCCURRENCES;

} /* namespace run */

} /* namespace automaton */
