/*
 * Result.cpp
 *
 *  Created on: 9. 2. 2014
 *	  Author: Jan Travnicek
 */

#include "Result.h"
#include <exception/AlibException.h>
#include <automaton/FSM/DFA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/TA/DFTA.h>
#include <label/LabelSetLabel.h>

#include <deque>

namespace automaton {

namespace run {

label::Label Result::result(const automaton::Automaton& automaton, const alib::Object& object) {
	std::pair<const alib::Object*, label::Label*> data(&object, NULL);
	automaton.getData().Accept((void*) &data, Result::RESULT);
	label::Label res = std::move(*data.second);
	delete data.second;
	return res;
}

label::Label Result::result(const automaton::DFA& automaton, const string::LinearString& string) {
	automaton::State state = automaton.getInitialState();
	for(const alphabet::Symbol& symbol : string.getContent()) {
		auto transitions = automaton.getTransitionsFromState(state);
		auto next = transitions.find(std::make_pair(state, symbol));
		if(next == transitions.end()) return label::Label(label::LabelSetLabel(std::set<label::Label>{}));
		state = next->second;
	}
	return state.getName();
}

label::Label Result::result(const automaton::DPDA& automaton, const string::LinearString& string) {
	automaton::State state = automaton.getInitialState();
	std::deque<alphabet::Symbol> pushdownStore;

	for(const alphabet::Symbol& symbol : string.getContent()) {
		auto transitions = automaton.getTransitionsFromState(state);
		auto transition = transitions.begin();
		int pushdownStoreSize = pushdownStore.size();
		for(auto transition = transitions.begin(); transition != transitions.end(); transition++) {
			if(std::get<1>(transition->first) != symbol) continue;

			const std::vector<alphabet::Symbol> & pop = std::get<2>(transition->first);
			int popSize = pop.size();
			if(pushdownStoreSize < popSize) continue;
			bool sign = true;
			for(int i = 1; i <= popSize; i++) {
				if(pop[popSize - i] != pushdownStore[pushdownStoreSize - i]) {
					sign = false;
					break;
				}
			}
			if(!sign) continue;

			break;
		}
		if(transition == transitions.end()) return label::Label(label::LabelSetLabel(std::set<label::Label>{}));
		for(auto pop : std::get<2>(transition->first)) pushdownStore.pop_back();

		state = transition->second.first;
		for(const alphabet::Symbol& push : transition->second.second) pushdownStore.push_back(push);
	}
	if(automaton.getFinalStates().empty()) {
		return label::Label(label::LabelSetLabel(std::set<label::Label>{}));
	} else {
		return state.getName();
	}
}

automaton::State Result::calculateState(const automaton::DFTA& automaton, const tree::RankedNode & node) {
	std::vector<automaton::State> states;
	states.reserve(node.getSymbol().getRank().getData());
	for (const auto & child : node.getChildren()) {
		automaton::State state = calculateState(automaton, *child);
		if (state == FAILSTATE) return FAILSTATE;
		states.push_back(state);
	}
	const auto & it = automaton.getTransitions().find(std::make_pair(node.getSymbol(), states));
	if (it == automaton.getTransitions().end()) return FAILSTATE;
	return it -> second;
}

label::Label Result::result(const automaton::DFTA& automaton, const tree::RankedTree& tree) {
	State state = calculateState(automaton, tree.getRoot());
	return state.getName();
}

void Result::Visit(void*, const automaton::EpsilonNFA& ) const {
	throw exception::AlibException("Unsupported automaton type EpsilonNFA");
}

void Result::Visit(void*, const automaton::MultiInitialStateNFA& ) const {
	throw exception::AlibException("Unsupported automaton type MultiInitialStateNFA");
}

void Result::Visit(void*, const automaton::NFA& ) const {
	throw exception::AlibException("Unsupported automaton type NFA");
}

void Result::Visit(void* data, const automaton::DFA& automaton) const {
	std::pair<const alib::Object*, label::Label*> & res = *((std::pair<const alib::Object*,  label::Label*>*) data);
	res.second = new label::Label(std::move(this->result(automaton, static_cast<const string::LinearString&>(res.first->getData()))));
}

void Result::Visit(void*, const automaton::ExtendedNFA& ) const {
	throw exception::AlibException("Unsupported automaton type ExtendedNFA");
}

void Result::Visit(void*, const automaton::CompactNFA& ) const {
	throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void Result::Visit(void* data, const DPDA& automaton) const {
	std::pair<const alib::Object*, label::Label*> & res = *((std::pair<const alib::Object*,  label::Label*>*) data);
	res.second = new label::Label(std::move(this->result(automaton, static_cast<const string::LinearString&>(res.first->getData()))));
}

void Result::Visit(void*, const SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void Result::Visit(void*, const InputDrivenDPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenDPDA");
}

void Result::Visit(void*, const InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void Result::Visit(void*, const VisiblyPushdownDPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void Result::Visit(void*, const VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void Result::Visit(void*, const RealTimeHeightDeterministicDPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void Result::Visit(void*, const RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void Result::Visit(void*, const NPDA&) const {
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void Result::Visit(void*, const SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void Result::Visit(void*, const OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

void Result::Visit(void* data, const automaton::DFTA& automaton) const {
	std::pair<const alib::Object*, label::Label*> & res = *((std::pair<const alib::Object*,  label::Label*>*) data);
	res.second = new label::Label(std::move(this->result(automaton, static_cast<const tree::RankedTree&>(res.first->getData()))));
}

void Result::Visit(void*, const NFTA&) const {
	throw exception::AlibException("Unsupported automaton type NFTA");
}

const Result Result::RESULT;

} /* namespace run */

} /* namespace automaton */
