/*
 * Rename.cpp
 *
 *  Created on: Dec 9, 2013
 *	Author: Jan Travnicek
 */

#include "Rename.h"

#include <map>
#include <deque>
#include <vector>
#include <set>
#include <algorithm>
#include <sstream>
#include <iostream>
#include "exception/AlibException.h"

#include "alphabet/Symbol.h"

#include "automaton/Automaton.h"

namespace automaton {

namespace simplify {

automaton::Automaton Rename::rename(const automaton::Automaton& automaton) {
	automaton::Automaton* out = NULL;
	automaton.getData().Accept((void*) &out, Rename::RENAME);
	automaton::Automaton res = std::move(*out);
	delete out;
	return res;
}

automaton::DFA Rename::rename(const automaton::DFA& fsm) {
	int counter = 0;
	std::map<automaton::State, int > renamingData;

	for(auto iter = fsm.getStates().begin(); iter != fsm.getStates().end(); iter++) {
		renamingData.insert(std::make_pair(*iter, counter++));
	}

	automaton::DFA result(automaton::State(renamingData.find(fsm.getInitialState())->second));

	result.setInputAlphabet(fsm.getInputAlphabet());

	for(auto iter = fsm.getStates().begin(); iter != fsm.getStates().end(); iter++) {
		result.addState(automaton::State(renamingData.find(*iter)->second));
	}

	for(auto iter = fsm.getFinalStates().begin(); iter != fsm.getFinalStates().end(); iter++) {
		result.addFinalState(automaton::State(renamingData.find(*iter)->second));
	}

	for(auto iter = fsm.getTransitions().begin(); iter != fsm.getTransitions().end(); iter++) {
		result.addTransition(automaton::State(renamingData.find(iter->first.first)->second), iter->first.second, automaton::State(renamingData.find(iter->second)->second));
	}

	return result;
}

automaton::DPDA Rename::rename(const automaton::DPDA& pda) {
	int counterState = 0;
	std::map<automaton::State, int > renamingDataState;
	int counterSymbol = 0;
	std::map<alphabet::Symbol, int > renamingDataSymbol;

	for(auto iter = pda.getStates().begin(); iter != pda.getStates().end(); iter++) {
		renamingDataState.insert(std::make_pair(*iter, counterState++));
	}

	for(auto iter = pda.getStackAlphabet().begin(); iter != pda.getStackAlphabet().end(); iter++) {
		renamingDataSymbol.insert(std::make_pair(*iter, counterSymbol++));
	}

	automaton::DPDA result(automaton::State(renamingDataState.find(pda.getInitialState())->second), alphabet::symbolFrom(renamingDataSymbol.find(pda.getInitialSymbol())->second));

	result.setInputAlphabet(pda.getInputAlphabet());

	for(auto iter = pda.getStackAlphabet().begin(); iter != pda.getStackAlphabet().end(); iter++) {
		result.addStackSymbol(alphabet::symbolFrom(renamingDataSymbol.find(*iter)->second));
	}

	for(auto iter = pda.getStates().begin(); iter != pda.getStates().end(); iter++) {
		result.addState(automaton::State(renamingDataState.find(*iter)->second));
	}

	for(auto iter = pda.getFinalStates().begin(); iter != pda.getFinalStates().end(); iter++) {
		result.addFinalState(automaton::State(renamingDataState.find(*iter)->second));
	}

	for(auto iter = pda.getTransitions().begin(); iter != pda.getTransitions().end(); iter++) {
		std::vector<alphabet::Symbol> pop;
		for(const auto& elem : std::get<2>(iter->first)) {
			pop.push_back(alphabet::symbolFrom(renamingDataSymbol.find(elem)->second));
		}
		std::vector<alphabet::Symbol> push;
		for(const auto& elem : iter->second.second) {
			push.push_back(alphabet::symbolFrom(renamingDataSymbol.find(elem)->second));
		}
		result.addTransition(automaton::State(renamingDataState.find(std::get<0>(iter->first))->second), std::get<1>(iter->first), pop, automaton::State(renamingDataState.find(iter->second.first)->second), push);
	}

	return result;
}

automaton::SinglePopDPDA Rename::rename(const automaton::SinglePopDPDA& pda) {
	int counterState = 0;
	std::map<automaton::State, int > renamingDataState;
	int counterSymbol = 0;
	std::map<alphabet::Symbol, int > renamingDataSymbol;

	for(auto iter = pda.getStates().begin(); iter != pda.getStates().end(); iter++) {
		renamingDataState.insert(std::make_pair(*iter, counterState++));
	}

	for(auto iter = pda.getStackAlphabet().begin(); iter != pda.getStackAlphabet().end(); iter++) {
		renamingDataSymbol.insert(std::make_pair(*iter, counterSymbol++));
	}

	automaton::SinglePopDPDA result(automaton::State(renamingDataState.find(pda.getInitialState())->second), alphabet::symbolFrom(renamingDataSymbol.find(pda.getInitialSymbol())->second));

	result.setInputAlphabet(pda.getInputAlphabet());

	for(auto iter = pda.getStackAlphabet().begin(); iter != pda.getStackAlphabet().end(); iter++) {
		result.addStackSymbol(alphabet::symbolFrom(renamingDataSymbol.find(*iter)->second));
	}

	for(auto iter = pda.getStates().begin(); iter != pda.getStates().end(); iter++) {
		result.addState(automaton::State(renamingDataState.find(*iter)->second));
	}

	for(auto iter = pda.getFinalStates().begin(); iter != pda.getFinalStates().end(); iter++) {
		result.addFinalState(automaton::State(renamingDataState.find(*iter)->second));
	}

	for(auto iter = pda.getTransitions().begin(); iter != pda.getTransitions().end(); iter++) {
		std::vector<alphabet::Symbol> push;
		for(const auto& elem : iter->second.second) {
			push.push_back(alphabet::symbolFrom(renamingDataSymbol.find(elem)->second));
		}
		result.addTransition(automaton::State(renamingDataState.find(std::get<0>(iter->first))->second), std::get<1>(iter->first), alphabet::symbolFrom(renamingDataSymbol.find(std::get<2>(iter->first))->second), automaton::State(renamingDataState.find(iter->second.first)->second), push);
	}

	return result;
}

automaton::InputDrivenDPDA Rename::rename(const automaton::InputDrivenDPDA& pda) {
	int counter = 0;
	std::map<automaton::State, int > renamingData;
	int counterSymbol = 0;
	std::map<alphabet::Symbol, int > renamingDataSymbol;

	for(auto iter = pda.getStates().begin(); iter != pda.getStates().end(); iter++) {
		renamingData.insert(std::make_pair(*iter, counter++));
	}

	for(auto iter = pda.getStackAlphabet().begin(); iter != pda.getStackAlphabet().end(); iter++) {
		renamingDataSymbol.insert(std::make_pair(*iter, counterSymbol++));
	}

	automaton::InputDrivenDPDA result(automaton::State(renamingData.find(pda.getInitialState())->second), alphabet::symbolFrom(renamingDataSymbol.find(pda.getInitialSymbol())->second));

	result.setInputAlphabet(pda.getInputAlphabet());
	result.setStackAlphabet(pda.getStackAlphabet());

	for(const std::pair<alphabet::Symbol, std::pair<std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol>>> operation : pda.getPushdownStoreOperations()) {
		std::vector<alphabet::Symbol> pop;
		for(const auto& elem : operation.second.first) {
			pop.push_back(alphabet::symbolFrom(renamingDataSymbol.find(elem)->second));
		}
		std::vector<alphabet::Symbol> push;
		for(const auto& elem : operation.second.second) {
			push.push_back(alphabet::symbolFrom(renamingDataSymbol.find(elem)->second));
		}
		result.setPushdownStoreOperation(operation.first, pop, push);
	}

	for(auto iter = pda.getStates().begin(); iter != pda.getStates().end(); iter++) {
		result.addState(automaton::State(renamingData.find(*iter)->second));
	}

	for(auto iter = pda.getFinalStates().begin(); iter != pda.getFinalStates().end(); iter++) {
		result.addFinalState(automaton::State(renamingData.find(*iter)->second));
	}

	for(auto iter = pda.getTransitions().begin(); iter != pda.getTransitions().end(); iter++) {
		result.addTransition(automaton::State(renamingData.find(iter->first.first)->second), iter->first.second, automaton::State(renamingData.find(iter->second)->second));
	}

	return result;
}

automaton::VisiblyPushdownDPDA Rename::rename(const automaton::VisiblyPushdownDPDA& pda) {
	int counterState = 0;
	std::map<automaton::State, int > renamingDataState;
	int counterSymbol = 0;
	std::map<alphabet::Symbol, int > renamingDataSymbol;

	for(auto iter = pda.getStates().begin(); iter != pda.getStates().end(); iter++) {
		renamingDataState.insert(std::make_pair(*iter, counterState++));
	}

	for(auto iter = pda.getStackAlphabet().begin(); iter != pda.getStackAlphabet().end(); iter++) {
		renamingDataSymbol.insert(std::make_pair(*iter, counterSymbol++));
	}

	automaton::VisiblyPushdownDPDA result(automaton::State(renamingDataState.find(pda.getInitialState())->second), alphabet::symbolFrom(renamingDataSymbol.find(pda.getBottomOfTheStackSymbol())->second));

	result.setCallInputAlphabet(pda.getCallInputAlphabet());
	result.setLocalInputAlphabet(pda.getLocalInputAlphabet());
	result.setReturnInputAlphabet(pda.getReturnInputAlphabet());

	for(auto iter = pda.getStackAlphabet().begin(); iter != pda.getStackAlphabet().end(); iter++) {
		result.addStackSymbol(alphabet::symbolFrom(renamingDataSymbol.find(*iter)->second));
	}

	for(auto iter = pda.getStates().begin(); iter != pda.getStates().end(); iter++) {
		result.addState(automaton::State(renamingDataState.find(*iter)->second));
	}

	for(auto iter = pda.getFinalStates().begin(); iter != pda.getFinalStates().end(); iter++) {
		result.addFinalState(automaton::State(renamingDataState.find(*iter)->second));
	}

	for(auto iter = pda.getCallTransitions().begin(); iter != pda.getCallTransitions().end(); iter++) {
		result.addCallTransition(automaton::State(renamingDataState.find(iter->first.first)->second), iter->first.second, automaton::State(renamingDataState.find(iter->second.first)->second), alphabet::symbolFrom(renamingDataSymbol.find(iter->second.second)->second));
	}

	for(auto iter = pda.getLocalTransitions().begin(); iter != pda.getLocalTransitions().end(); iter++) {
		result.addLocalTransition(automaton::State(renamingDataState.find(iter->first.first)->second), iter->first.second, automaton::State(renamingDataState.find(iter->second)->second));
	}

	for(auto iter = pda.getReturnTransitions().begin(); iter != pda.getReturnTransitions().end(); iter++) {
		result.addReturnTransition(automaton::State(renamingDataState.find(std::get<0>(iter->first))->second), std::get<1>(iter->first), alphabet::symbolFrom(renamingDataSymbol.find(std::get<2>(iter->first))->second), automaton::State(renamingDataState.find(iter->second)->second));
	}

	return result;
}

automaton::RealTimeHeightDeterministicDPDA Rename::rename(const automaton::RealTimeHeightDeterministicDPDA& pda) {
	int counterState = 0;
	std::map<automaton::State, int > renamingDataState;
	int counterSymbol = 0;
	std::map<alphabet::Symbol, int > renamingDataSymbol;

	for(auto iter = pda.getStates().begin(); iter != pda.getStates().end(); iter++) {
		renamingDataState.insert(std::make_pair(*iter, counterState++));
	}

	for(auto iter = pda.getStackAlphabet().begin(); iter != pda.getStackAlphabet().end(); iter++) {
		renamingDataSymbol.insert(std::make_pair(*iter, counterSymbol++));
	}

	automaton::RealTimeHeightDeterministicDPDA result(automaton::State(renamingDataState.find(pda.getInitialState())->second), alphabet::symbolFrom(renamingDataSymbol.find(pda.getBottomOfTheStackSymbol())->second));

	result.setInputAlphabet(pda.getInputAlphabet());

	for(auto iter = pda.getStackAlphabet().begin(); iter != pda.getStackAlphabet().end(); iter++) {
		result.addStackSymbol(alphabet::symbolFrom(renamingDataSymbol.find(*iter)->second));
	}

	for(auto iter = pda.getStates().begin(); iter != pda.getStates().end(); iter++) {
		result.addState(automaton::State(renamingDataState.find(*iter)->second));
	}

	for(auto iter = pda.getFinalStates().begin(); iter != pda.getFinalStates().end(); iter++) {
		result.addFinalState(automaton::State(renamingDataState.find(*iter)->second));
	}

	for(auto iter = pda.getCallTransitions().begin(); iter != pda.getCallTransitions().end(); iter++) {
		result.addCallTransition(automaton::State(renamingDataState.find(iter->first.first)->second), iter->first.second, automaton::State(renamingDataState.find(iter->second.first)->second), alphabet::symbolFrom(renamingDataSymbol.find(iter->second.second)->second));
	}

	for(auto iter = pda.getLocalTransitions().begin(); iter != pda.getLocalTransitions().end(); iter++) {
		result.addLocalTransition(automaton::State(renamingDataState.find(iter->first.first)->second), iter->first.second, automaton::State(renamingDataState.find(iter->second)->second));
	}

	for(auto iter = pda.getReturnTransitions().begin(); iter != pda.getReturnTransitions().end(); iter++) {
		result.addReturnTransition(automaton::State(renamingDataState.find(std::get<0>(iter->first))->second), std::get<1>(iter->first), alphabet::symbolFrom(renamingDataSymbol.find(std::get<2>(iter->first))->second), automaton::State(renamingDataState.find(iter->second)->second));
	}

	return result;
}

void Rename::Visit(void*, const automaton::EpsilonNFA&) const {
	throw exception::AlibException("Unsupported automaton type EpsilonNFA");
}

void Rename::Visit(void* data, const automaton::DFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->rename(automaton));
}

void Rename::Visit(void*, const automaton::MultiInitialStateNFA&) const {
	throw exception::AlibException("Unsupported automaton type MultiInitialStateNFA");
}

void Rename::Visit(void*, const automaton::NFA&) const {
	throw exception::AlibException("Unsupported automaton type DFA");
}

void Rename::Visit(void*, const automaton::ExtendedNFA& ) const {
	throw exception::AlibException("Unsupported automaton type ExtendedNFA");
}

void Rename::Visit(void*, const automaton::CompactNFA& ) const {
	throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void Rename::Visit(void* data, const automaton::DPDA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->rename(automaton));
}

void Rename::Visit(void* data, const automaton::SinglePopDPDA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->rename(automaton));
}

void Rename::Visit(void* data, const automaton::InputDrivenDPDA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->rename(automaton));
}

void Rename::Visit(void*, const automaton::InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void Rename::Visit(void* data, const automaton::VisiblyPushdownDPDA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->rename(automaton));
}

void Rename::Visit(void*, const automaton::VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void Rename::Visit(void* data, const automaton::RealTimeHeightDeterministicDPDA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->rename(automaton));
}

void Rename::Visit(void*, const automaton::RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void Rename::Visit(void*, const automaton::NPDA&) const {
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void Rename::Visit(void*, const automaton::SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void Rename::Visit(void*, const automaton::OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

void Rename::Visit(void*, const automaton::DFTA&) const {
	throw exception::AlibException("Unsupported automaton type DFTA");
}

void Rename::Visit(void*, const automaton::NFTA&) const {
	throw exception::AlibException("Unsupported automaton type NFTA");
}

const Rename Rename::RENAME;

} /* namespace simplify */

} /* namespace automaton */
