/*
 * SingleInitialState.cpp
 *
 *  Created on: 20. 9. 2014
 *	  Author: Tomas Pecka
 */

#include "SingleInitialState.h"

#include <algorithm>
#include <set>

#include <exception/AlibException.h>
#include <automaton/FSM/DFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/CompactNFA.h>

#include <label/LabelSetLabel.h>

namespace automaton {

namespace simplify {

automaton::Automaton SingleInitialState::convert(const Automaton& automaton) {
	automaton::Automaton* out = NULL;
	automaton.getData().Accept((void*) &out, SingleInitialState::SINGLE_INITIAL_STATE);
	automaton::Automaton res = std::move(*out);
	delete out;
	return res;
}

automaton::NFA SingleInitialState::convert(const automaton::MultiInitialStateNFA& automaton) {
	// step 1, 3
	std::set<label::Label> labelSet;
	for(const auto& q : automaton.getInitialStates()) {
		labelSet.insert(q.getName());
	}
	automaton::State q0(label::Label(label::LabelSetLabel(std::move(labelSet))));

	automaton::NFA res(q0);
	res.setInputAlphabet(automaton.getInputAlphabet());
	for(const auto& q : automaton.getStates())
		res.addState(q);

	// step 2
	for(const auto& q : automaton.getInitialStates()) {
		for(const auto& kv: automaton.getTransitionsFromState(q)) {
			for(const auto& to : kv.second) {
				res.addTransition(q0, kv.first.second, to);
			}
		}
	}

	for(const auto& t : automaton.getTransitions()) {
		for(const auto& q : t.second) {
			res.addTransition(t.first.first, t.first.second, q);
		 }
	}

	// step 4, 5
	std::set<automaton::State> intersection;
	std::set_intersection(res.getFinalStates().begin(), res.getFinalStates().end(), automaton.getInitialStates().begin(), automaton.getInitialStates().end(), std::inserter(intersection, intersection.begin()));

	res.setFinalStates(automaton.getFinalStates());
	if(intersection.size() != 0) {
		res.addFinalState(q0);
	}

	return res;
}

automaton::DFA SingleInitialState::convert(const automaton::DFA& automaton) {
	return automaton;
}

automaton::EpsilonNFA SingleInitialState::convert(const automaton::EpsilonNFA& automaton) {
	return automaton;
}

automaton::NFA SingleInitialState::convert(const automaton::NFA& automaton) {
	return automaton;
}

automaton::CompactNFA SingleInitialState::convert(const automaton::CompactNFA& automaton) {
	return automaton;
}

automaton::ExtendedNFA SingleInitialState::convert(const automaton::ExtendedNFA& automaton) {
	return automaton;
}

void SingleInitialState::Visit(void* data, const EpsilonNFA& automaton) const {
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

void SingleInitialState::Visit(void* data, const MultiInitialStateNFA& automaton) const {
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

void SingleInitialState::Visit(void* data, const NFA& automaton) const {
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

void SingleInitialState::Visit(void* data, const DFA& automaton) const {
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

void SingleInitialState::Visit(void* data, const ExtendedNFA& automaton) const {
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

void SingleInitialState::Visit(void* data, const CompactNFA& automaton) const {
	automaton::Automaton*& out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->convert(automaton));
}

void SingleInitialState::Visit(void*, const DPDA&) const {
	throw exception::AlibException("Unsupported automaton type DPDA");
}

void SingleInitialState::Visit(void*, const SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void SingleInitialState::Visit(void*, const InputDrivenDPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void SingleInitialState::Visit(void*, const InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void SingleInitialState::Visit(void*, const VisiblyPushdownDPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void SingleInitialState::Visit(void*, const VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void SingleInitialState::Visit(void*, const RealTimeHeightDeterministicDPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void SingleInitialState::Visit(void*, const RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void SingleInitialState::Visit(void*, const NPDA&) const {
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void SingleInitialState::Visit(void*, const SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void SingleInitialState::Visit(void*, const OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

void SingleInitialState::Visit(void*, const automaton::DFTA&) const {
	throw exception::AlibException("Unsupported automaton type DFTA");
}

void SingleInitialState::Visit(void*, const automaton::NFTA&) const {
	throw exception::AlibException("Unsupported automaton type NFTA");
}

const SingleInitialState SingleInitialState::SINGLE_INITIAL_STATE;

} /* namespace simplify */

} /* namespace automaton */
