/*
 * AutomataConcatenation.cpp
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#include "AutomataConcatenation.h"
#include <exception/AlibException.h>
#include <label/Label.h>
#include "common/PairLabel.h"

#define AUTOMATON_FIRST  (label::labelFrom(1))
#define AUTOMATON_SECOND (label::labelFrom(2))

namespace automaton
{

namespace transform
{

automaton::Automaton AutomataConcatenation::concatenation(const automaton::Automaton& first, const automaton::Automaton& second)
{
	AutomatonBase* out;
	Accept((void*) &out, first.getData(), second.getData(), AutomataConcatenation::AUTOMATA_CONCATENATION);
	automaton::Automaton res(*out);
	delete out;
	return res;
}

automaton::NFA AutomataConcatenation::concatenation(const automaton::NFA& first, const automaton::NFA& second)
{
	automaton::State q01q02(pairLabel(first.getInitialState(), second.getInitialState()));
	automaton::NFA res(pairLabel(AUTOMATON_FIRST, first.getInitialState()));

	for(const auto& q : first.getStates())
		res.addState(pairLabel(AUTOMATON_FIRST, q));
	for(const auto& q : second.getStates())
		res.addState(pairLabel(AUTOMATON_SECOND, q));
	res.addState(q01q02);

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.setInitialState(q01q02);

	for(const auto& t : first.getTransitions())
	{
		for(const auto& q : t.second)
		{
			res.addTransition(pairLabel(AUTOMATON_FIRST, t.first.first), t.first.second, pairLabel(AUTOMATON_FIRST, q));

			if(first.getFinalStates().count(q) > 0)
				res.addTransition(pairLabel(AUTOMATON_FIRST, t.first.first), t.first.second, pairLabel(AUTOMATON_SECOND, second.getInitialState()));
		}
	}
	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(AUTOMATON_SECOND, t.first.first), t.first.second, pairLabel(AUTOMATON_SECOND, q));

	for(const auto& t : first.getTransitionsFromState(first.getInitialState()))
		for(const auto& q : t.second)
			res.addTransition(q01q02, t.first.second, pairLabel(AUTOMATON_FIRST, q));
	for(const auto& t : second.getTransitionsFromState(second.getInitialState()))
		for(const auto& q : t.second)
			res.addTransition(q01q02, t.first.second, pairLabel(AUTOMATON_SECOND, q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(AUTOMATON_SECOND, q));
	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.addFinalState(q01q02);

	return res;
}

automaton::NFA AutomataConcatenation::concatenation(const automaton::DFA& first, const automaton::DFA& second)
{
	automaton::State q01q02(pairLabel(first.getInitialState(), second.getInitialState()));
	automaton::NFA res(pairLabel(AUTOMATON_FIRST, first.getInitialState()));

	for(const auto& q : first.getStates())
		res.addState(pairLabel(AUTOMATON_FIRST, q));
	for(const auto& q : second.getStates())
		res.addState(pairLabel(AUTOMATON_SECOND, q));
	res.addState(q01q02);

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.setInitialState(q01q02);

	for(const auto& t : first.getTransitions())
	{
		res.addTransition(pairLabel(AUTOMATON_FIRST, t.first.first), t.first.second, pairLabel(AUTOMATON_FIRST, t.second));

		if(first.getFinalStates().count(t.second) > 0)
			res.addTransition(pairLabel(AUTOMATON_FIRST, t.first.first), t.first.second, pairLabel(AUTOMATON_SECOND, second.getInitialState()));
	}
	for(const auto& t : second.getTransitions())
		res.addTransition(pairLabel(AUTOMATON_SECOND, t.first.first), t.first.second, pairLabel(AUTOMATON_SECOND, t.second));

	for(const auto& t : first.getTransitionsFromState(first.getInitialState()))
		res.addTransition(q01q02, t.first.second, pairLabel(AUTOMATON_FIRST, t.second));
	for(const auto& t : second.getTransitionsFromState(second.getInitialState()))
		res.addTransition(q01q02, t.first.second, pairLabel(AUTOMATON_SECOND, t.second));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(AUTOMATON_SECOND, q));
	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.addFinalState(q01q02);

	return res;
}

void AutomataConcatenation::Visit(void*, const automaton::EpsilonNFA&, const automaton::EpsilonNFA&) const
{
	throw exception::AlibException("Unsupported automaton type EpsilonNFA");
}

void AutomataConcatenation::Visit(void*, const automaton::CompactNFA&, const automaton::CompactNFA&) const
{
	throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void AutomataConcatenation::Visit(void* data, const automaton::DFA& first, const automaton::DFA& second) const
{
	AutomatonBase* &ret = *(AutomatonBase**) data;
	ret = std::move(AutomataConcatenation::concatenation(first, second)).plunder();
}

void AutomataConcatenation::Visit(void*, const automaton::ExtendedNFA&, const automaton::ExtendedNFA&) const
{
	throw exception::AlibException("Unsupported automaton type ExtendedNFA");
}

void AutomataConcatenation::Visit(void*, const automaton::MultiInitialStateNFA&, const automaton::MultiInitialStateNFA&) const
{
	throw exception::AlibException("Unsupported automaton type MultiInitialStateNFA");
}

void AutomataConcatenation::Visit(void* data, const automaton::NFA& first, const automaton::NFA& second) const
{
	AutomatonBase* &ret = *(AutomatonBase**) data;
	ret = std::move(AutomataConcatenation::concatenation(first, second)).plunder();
}

void AutomataConcatenation::Visit(void*, const automaton::DPDA&, const automaton::DPDA&) const
{
	throw exception::AlibException("Unsupported automaton type DPDA");
}

void AutomataConcatenation::Visit(void*, const automaton::NPDA&, const automaton::NPDA&) const
{
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void AutomataConcatenation::Visit(void*, const automaton::InputDrivenDPDA&, const automaton::InputDrivenDPDA&) const
{
	throw exception::AlibException("Unsupported automaton type InputDrivenDPDA");
}

void AutomataConcatenation::Visit(void*, const automaton::InputDrivenNPDA&, const automaton::InputDrivenNPDA&) const
{
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void AutomataConcatenation::Visit(void*, const automaton::RealTimeHeightDeterministicDPDA&, const automaton::RealTimeHeightDeterministicDPDA&) const
{
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void AutomataConcatenation::Visit(void*, const automaton::RealTimeHeightDeterministicNPDA&, const automaton::RealTimeHeightDeterministicNPDA&) const
{
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void AutomataConcatenation::Visit(void*, const automaton::SinglePopNPDA&, const automaton::SinglePopNPDA&) const
{
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void AutomataConcatenation::Visit(void*, const automaton::SinglePopDPDA&, const automaton::SinglePopDPDA&) const
{
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void AutomataConcatenation::Visit(void*, const automaton::VisiblyPushdownDPDA&, const automaton::VisiblyPushdownDPDA&) const
{
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void AutomataConcatenation::Visit(void*, const automaton::VisiblyPushdownNPDA&, const automaton::VisiblyPushdownNPDA&) const
{
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void AutomataConcatenation::Visit(void*, const automaton::OneTapeDTM&, const automaton::OneTapeDTM&) const
{
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

void AutomataConcatenation::Visit(void*, const automaton::DFTA&, const automaton::DFTA&) const
{
	throw exception::AlibException("Unsupported automaton type DFTA");
}

void AutomataConcatenation::Visit(void*, const automaton::NFTA&, const automaton::NFTA&) const
{
	throw exception::AlibException("Unsupported automaton type NFTA");
}

const AutomataConcatenation AutomataConcatenation::AUTOMATA_CONCATENATION;

} /* namespace transform */

} /* namespace automaton */
