/*
 * AutomataIntersectionCartesianProduct.cpp
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#include "AutomataIntersectionCartesianProduct.h"
#include <exception/AlibException.h>
#include "common/PairLabel.h"

#define AUTOMATON_FIRST  1
#define AUTOMATON_SECOND 2

namespace automaton
{

namespace transform
{

automaton::Automaton AutomataIntersectionCartesianProduct::intersection(const automaton::Automaton& first, const automaton::Automaton& second)
{
	AutomatonBase* out;
	Accept((void*) &out, first.getData(), second.getData(), AutomataIntersectionCartesianProduct::AUTOMATA_INTERSECTION_CARTESIAN_PRODUCT);
	automaton::Automaton res(*out);
	delete out;
	return res;
}

automaton::DFA AutomataIntersectionCartesianProduct::intersection(const automaton::DFA& first, const automaton::DFA& second)
{
	automaton::State q0(pairLabel(first.getInitialState(), second.getInitialState()));
	automaton::DFA res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& p : first.getStates())
		for(const auto& q : second.getStates())
			res.addState(automaton::State(pairLabel(p, q)));

	for(const auto& p : first.getFinalStates())
		for(const auto& q : second.getFinalStates())
			res.addFinalState(automaton::State(pairLabel(p, q)));

	for(const auto& state : res.getStates())
	{
		const label::Label& label_p = static_cast<const label::LabelPairLabel&>(state.getName().getData()).getData().first;
		const label::Label& label_q = static_cast<const label::LabelPairLabel&>(state.getName().getData()).getData().second;

		for(const auto& tp : first.getTransitionsFromState(automaton::State(label_p)))
			for(const auto& tq : second.getTransitionsFromState(automaton::State(label_q)))
				if(tp.first.second == tq.first.second)
					res.addTransition(state, tp.first.second, automaton::State(pairLabel(tp.second, tq.second)));
	}

	return res;
}

automaton::NFA AutomataIntersectionCartesianProduct::intersection(const automaton::NFA& first, const automaton::NFA& second)
{
	automaton::State q0(pairLabel(first.getInitialState(), second.getInitialState()));
	automaton::NFA res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& p : first.getStates())
		for(const auto& q : second.getStates())
			res.addState(automaton::State(pairLabel(p, q)));

	for(const auto& p : first.getFinalStates())
		for(const auto& q : second.getFinalStates())
			res.addFinalState(automaton::State(pairLabel(p, q)));

	for(const auto& state : res.getStates())
	{
		const label::Label& label_p = static_cast<const label::LabelPairLabel&>(state.getName().getData()).getData().first;
		const label::Label& label_q = static_cast<const label::LabelPairLabel&>(state.getName().getData()).getData().second;

		for(const auto& tp : first.getTransitionsFromState(automaton::State(label_p)))
			for(const auto& tq : second.getTransitionsFromState(automaton::State(label_q)))
				if(tp.first.second == tq.first.second)
					for(const auto& p : tp.second)
						for(const auto& q : tq.second)
							res.addTransition(state, tp.first.second, automaton::State(pairLabel(p, q)));
	}

	return res;
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::EpsilonNFA&, const automaton::EpsilonNFA&) const
{
	throw exception::AlibException("Unsupported automaton type EpsilonNFA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::CompactNFA&, const automaton::CompactNFA&) const
{
	throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void AutomataIntersectionCartesianProduct::Visit(void* data, const automaton::DFA& first, const automaton::DFA& second) const
{
	AutomatonBase* &ret = *(AutomatonBase**) data;
	ret = std::move(AutomataIntersectionCartesianProduct::intersection(first, second)).plunder();
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::ExtendedNFA&, const automaton::ExtendedNFA&) const
{
	throw exception::AlibException("Unsupported automaton type ExtendedNFA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::MultiInitialStateNFA&, const automaton::MultiInitialStateNFA&) const
{
	throw exception::AlibException("Unsupported automaton type MultiInitialStateNFA");
}

void AutomataIntersectionCartesianProduct::Visit(void* data, const automaton::NFA& first, const automaton::NFA& second) const
{
	AutomatonBase* &ret = *(AutomatonBase**) data;
	ret = std::move(AutomataIntersectionCartesianProduct::intersection(first, second)).plunder();
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::DPDA&, const automaton::DPDA&) const
{
	throw exception::AlibException("Unsupported automaton type DPDA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::NPDA&, const automaton::NPDA&) const
{
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::InputDrivenDPDA&, const automaton::InputDrivenDPDA&) const
{
	throw exception::AlibException("Unsupported automaton type InputDrivenDPDA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::InputDrivenNPDA&, const automaton::InputDrivenNPDA&) const
{
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::RealTimeHeightDeterministicDPDA&, const automaton::RealTimeHeightDeterministicDPDA&) const
{
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::RealTimeHeightDeterministicNPDA&, const automaton::RealTimeHeightDeterministicNPDA&) const
{
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::SinglePopNPDA&, const automaton::SinglePopNPDA&) const
{
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::SinglePopDPDA&, const automaton::SinglePopDPDA&) const
{
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::VisiblyPushdownDPDA&, const automaton::VisiblyPushdownDPDA&) const
{
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::VisiblyPushdownNPDA&, const automaton::VisiblyPushdownNPDA&) const
{
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::OneTapeDTM&, const automaton::OneTapeDTM&) const
{
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::DFTA&, const automaton::DFTA&) const
{
	throw exception::AlibException("Unsupported automaton type DFTA");
}

void AutomataIntersectionCartesianProduct::Visit(void*, const automaton::NFTA&, const automaton::NFTA&) const
{
	throw exception::AlibException("Unsupported automaton type NFTA");
}

const AutomataIntersectionCartesianProduct AutomataIntersectionCartesianProduct::AUTOMATA_INTERSECTION_CARTESIAN_PRODUCT;

} /* namespace transform */

} /* namespace automaton */
