/*
 * AutomataUnionEpsilonTransition.cpp
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#include "AutomataUnionEpsilonTransition.h"
#include <exception/AlibException.h>
#include "common/PairLabel.h"

#define AUTOMATON_FIRST  1
#define AUTOMATON_SECOND 2

namespace automaton
{

namespace transform
{

automaton::Automaton AutomataUnionEpsilonTransition::unification(const automaton::Automaton& first, const automaton::Automaton& second)
{
	AutomatonBase* out;
	Accept((void*) &out, first.getData(), second.getData(), AutomataUnionEpsilonTransition::AUTOMATA_UNION_EPSILON_TRANSITION);
	automaton::Automaton res(*out);
	delete out;
	return res;
}

automaton::EpsilonNFA AutomataUnionEpsilonTransition::unification(const automaton::EpsilonNFA& first, const automaton::EpsilonNFA& second)
{
	std::set<automaton::State> states;
	for(const auto& q : first.getStates())
		states.insert(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getStates())
		states.insert(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	automaton::State q0 = automaton::createUniqueState(automaton::State("q0"), states);
	automaton::EpsilonNFA res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& q : states)
		res.addState(q);

	for(const auto& q : first.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	res.addTransition(q0, pairLabel(label::labelFrom(AUTOMATON_FIRST), first.getInitialState()));
	res.addTransition(q0, pairLabel(label::labelFrom(AUTOMATON_SECOND), second.getInitialState()));

	for(const auto& t : first.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(label::labelFrom(AUTOMATON_FIRST), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_FIRST), q));

	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(label::labelFrom(AUTOMATON_SECOND), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	return res;
}

automaton::EpsilonNFA AutomataUnionEpsilonTransition::unification(const automaton::NFA& first, const automaton::NFA& second)
{
	std::set<automaton::State> states;
	for(const auto& q : first.getStates())
		states.insert(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getStates())
		states.insert(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	automaton::State q0 = automaton::createUniqueState(automaton::State("q0"), states);
	automaton::EpsilonNFA res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& q : states)
		res.addState(q);

	for(const auto& q : first.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	res.addTransition(q0, pairLabel(label::labelFrom(AUTOMATON_FIRST), first.getInitialState()));
	res.addTransition(q0, pairLabel(label::labelFrom(AUTOMATON_SECOND), second.getInitialState()));

	for(const auto& t : first.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(label::labelFrom(AUTOMATON_FIRST), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_FIRST), q));

	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(pairLabel(label::labelFrom(AUTOMATON_SECOND), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	return res;
}

automaton::EpsilonNFA AutomataUnionEpsilonTransition::unification(const automaton::DFA& first, const automaton::DFA& second)
{
	std::set<automaton::State> states;
	for(const auto& q : first.getStates())
		states.insert(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getStates())
		states.insert(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	automaton::State q0 = automaton::createUniqueState(automaton::State("q0"), states);
	automaton::EpsilonNFA res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& q : states)
		res.addState(q);

	for(const auto& q : first.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_FIRST), q));
	for(const auto& q : second.getFinalStates())
		res.addFinalState(pairLabel(label::labelFrom(AUTOMATON_SECOND), q));

	res.addTransition(q0, pairLabel(label::labelFrom(AUTOMATON_FIRST), first.getInitialState()));
	res.addTransition(q0, pairLabel(label::labelFrom(AUTOMATON_SECOND), second.getInitialState()));

	for(const auto& t : first.getTransitions())
		res.addTransition(pairLabel(label::labelFrom(AUTOMATON_FIRST), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_FIRST), t.second));

	for(const auto& t : second.getTransitions())
		res.addTransition(pairLabel(label::labelFrom(AUTOMATON_SECOND), t.first.first), t.first.second, pairLabel(label::labelFrom(AUTOMATON_SECOND), t.second));

	return res;
}

void AutomataUnionEpsilonTransition::Visit(void* data, const automaton::EpsilonNFA& first, const automaton::EpsilonNFA& second) const
{
	AutomatonBase* &ret = *(AutomatonBase**) data;
	ret = std::move(AutomataUnionEpsilonTransition::unification(first, second)).plunder();
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::CompactNFA&, const automaton::CompactNFA&) const
{
	throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void AutomataUnionEpsilonTransition::Visit(void* data, const automaton::DFA& first, const automaton::DFA& second) const
{
	AutomatonBase* &ret = *(AutomatonBase**) data;
	ret = std::move(AutomataUnionEpsilonTransition::unification(first, second)).plunder();
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::ExtendedNFA&, const automaton::ExtendedNFA&) const
{
	throw exception::AlibException("Unsupported automaton type ExtendedNFA");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::MultiInitialStateNFA&, const automaton::MultiInitialStateNFA&) const
{
	throw exception::AlibException("Unsupported automaton type MultiInitialStateNFA");
}

void AutomataUnionEpsilonTransition::Visit(void* data, const automaton::NFA& first, const automaton::NFA& second) const
{
	AutomatonBase* &ret = *(AutomatonBase**) data;
	ret = std::move(AutomataUnionEpsilonTransition::unification(first, second)).plunder();
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::DPDA&, const automaton::DPDA&) const
{
	throw exception::AlibException("Unsupported automaton type DPDA");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::NPDA&, const automaton::NPDA&) const
{
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::InputDrivenDPDA&, const automaton::InputDrivenDPDA&) const
{
	throw exception::AlibException("Unsupported automaton type InputDrivenDPDA");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::InputDrivenNPDA&, const automaton::InputDrivenNPDA&) const
{
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::RealTimeHeightDeterministicDPDA&, const automaton::RealTimeHeightDeterministicDPDA&) const
{
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::RealTimeHeightDeterministicNPDA&, const automaton::RealTimeHeightDeterministicNPDA&) const
{
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::SinglePopNPDA&, const automaton::SinglePopNPDA&) const
{
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::SinglePopDPDA&, const automaton::SinglePopDPDA&) const
{
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::VisiblyPushdownDPDA&, const automaton::VisiblyPushdownDPDA&) const
{
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::VisiblyPushdownNPDA&, const automaton::VisiblyPushdownNPDA&) const
{
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::OneTapeDTM&, const automaton::OneTapeDTM&) const
{
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::DFTA&, const automaton::DFTA&) const
{
	throw exception::AlibException("Unsupported automaton type DFTA");
}

void AutomataUnionEpsilonTransition::Visit(void*, const automaton::NFTA&, const automaton::NFTA&) const
{
	throw exception::AlibException("Unsupported automaton type NFTA");
}

const AutomataUnionEpsilonTransition AutomataUnionEpsilonTransition::AUTOMATA_UNION_EPSILON_TRANSITION;

} /* namespace transform */

} /* namespace automaton */
