/*
 * NonterminalUnitRuleCycle.cpp
 *
 *  Created on: 22. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "NonterminalUnitRuleCycle.h"

#include <grammar/ContextFree/CFG.h>
#include <grammar/ContextFree/EpsilonFreeCFG.h>
#include <grammar/ContextFree/GNF.h>
#include <grammar/ContextFree/CNF.h>
#include <grammar/ContextFree/LG.h>
#include <grammar/Regular/LeftLG.h>
#include <grammar/Regular/LeftRG.h>
#include <grammar/Regular/RightLG.h>
#include <grammar/Regular/RightRG.h>

#include <set>
#include <deque>

namespace grammar {

namespace properties {

template<class T>
std::set<alphabet::Symbol> NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(const T& grammar, const alphabet::Symbol& nonterminal) {
	if(grammar.getNonterminalAlphabet().count(nonterminal) == 0) {
		throw exception::AlibException("Nonterminal symbol \"" + (std::string) nonterminal + "\" is not present in grammar.");
	}

	std::deque<std::set<alphabet::Symbol>> Ni;
	Ni.push_back(std::set<alphabet::Symbol>{nonterminal});
	int i = 1;

	while(true) {
		Ni.push_back(Ni.at(i-1));
		for(const auto&rule : grammar.getRawRules()) {
			const alphabet::Symbol& lhs = rule.first;

			for(const auto& rhs : rule.second) {
				if(Ni.at(i-1).count(lhs) && rhs.size() == 1 && grammar.getNonterminalAlphabet().count(rhs.front())) {
					Ni.at(i).insert(rhs.front());
				}
			}
		}

		if(Ni.at(i) == Ni.at(i-1))
			break;
		
		i += 1;
	}

	return Ni.at(i);
}

template std::set<alphabet::Symbol> NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(const grammar::CFG& grammar, const alphabet::Symbol& nonterminal);
template std::set<alphabet::Symbol> NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(const grammar::EpsilonFreeCFG& grammar, const alphabet::Symbol& nonterminal);
template std::set<alphabet::Symbol> NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(const grammar::GNF& grammar, const alphabet::Symbol& nonterminal);
template std::set<alphabet::Symbol> NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(const grammar::CNF& grammar, const alphabet::Symbol& nonterminal);
template std::set<alphabet::Symbol> NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(const grammar::LG& grammar, const alphabet::Symbol& nonterminal);
template std::set<alphabet::Symbol> NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(const grammar::LeftLG& grammar, const alphabet::Symbol& nonterminal);
template std::set<alphabet::Symbol> NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(const grammar::LeftRG& grammar, const alphabet::Symbol& nonterminal);
template std::set<alphabet::Symbol> NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(const grammar::RightLG& grammar, const alphabet::Symbol& nonterminal);
template std::set<alphabet::Symbol> NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(const grammar::RightRG& grammar, const alphabet::Symbol& nonterminal);

std::set<alphabet::Symbol> NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(const grammar::Grammar& grammar, const alphabet::Symbol& nonterminal) {
	std::pair<alphabet::Symbol, std::set<alphabet::Symbol>> out(nonterminal, {});
	grammar.getData().Accept((void*) &out, NonterminalUnitRuleCycle::NONTERMINAL_UNIT_RULE_CYCLE);
	return out.second;
}

void NonterminalUnitRuleCycle::Visit(void* data, const grammar::LeftLG& grammar) const {
	std::pair<alphabet::Symbol, std::set<alphabet::Symbol>> & out = *((std::pair<alphabet::Symbol, std::set<alphabet::Symbol>>*) data);
	out.second = std::move(this->getNonterminalUnitRuleCycle(grammar, out.first));
}

void NonterminalUnitRuleCycle::Visit(void* data, const grammar::LeftRG& grammar) const {
	std::pair<alphabet::Symbol, std::set<alphabet::Symbol>> & out = *((std::pair<alphabet::Symbol, std::set<alphabet::Symbol>>*) data);
	out.second = std::move(this->getNonterminalUnitRuleCycle(grammar, out.first));
}

void NonterminalUnitRuleCycle::Visit(void* data, const grammar::RightLG& grammar) const {
	std::pair<alphabet::Symbol, std::set<alphabet::Symbol>> & out = *((std::pair<alphabet::Symbol, std::set<alphabet::Symbol>>*) data);
	out.second = std::move(this->getNonterminalUnitRuleCycle(grammar, out.first));
}

void NonterminalUnitRuleCycle::Visit(void* data, const grammar::RightRG& grammar) const {
	std::pair<alphabet::Symbol, std::set<alphabet::Symbol>> & out = *((std::pair<alphabet::Symbol, std::set<alphabet::Symbol>>*) data);
	out.second = std::move(this->getNonterminalUnitRuleCycle(grammar, out.first));
}

void NonterminalUnitRuleCycle::Visit(void* data, const grammar::LG& grammar) const {
	std::pair<alphabet::Symbol, std::set<alphabet::Symbol>> & out = *((std::pair<alphabet::Symbol, std::set<alphabet::Symbol>>*) data);
	out.second = std::move(this->getNonterminalUnitRuleCycle(grammar, out.first));
}

void NonterminalUnitRuleCycle::Visit(void* data, const grammar::CFG& grammar) const {
	std::pair<alphabet::Symbol, std::set<alphabet::Symbol>> & out = *((std::pair<alphabet::Symbol, std::set<alphabet::Symbol>>*) data);
	out.second = std::move(this->getNonterminalUnitRuleCycle(grammar, out.first));
}

void NonterminalUnitRuleCycle::Visit(void* data, const grammar::EpsilonFreeCFG& grammar) const {
	std::pair<alphabet::Symbol, std::set<alphabet::Symbol>> & out = *((std::pair<alphabet::Symbol, std::set<alphabet::Symbol>>*) data);
	out.second = std::move(this->getNonterminalUnitRuleCycle(grammar, out.first));
}

void NonterminalUnitRuleCycle::Visit(void* data, const grammar::CNF& grammar) const {
	std::pair<alphabet::Symbol, std::set<alphabet::Symbol>> & out = *((std::pair<alphabet::Symbol, std::set<alphabet::Symbol>>*) data);
	out.second = std::move(this->getNonterminalUnitRuleCycle(grammar, out.first));
}

void NonterminalUnitRuleCycle::Visit(void* data, const grammar::GNF& grammar) const {
	std::pair<alphabet::Symbol, std::set<alphabet::Symbol>> & out = *((std::pair<alphabet::Symbol, std::set<alphabet::Symbol>>*) data);
	out.second = std::move(this->getNonterminalUnitRuleCycle(grammar, out.first));
}

void NonterminalUnitRuleCycle::Visit(void*, const grammar::CSG&) const {
	throw exception::AlibException("Unsupported grammar type CSG");
}

void NonterminalUnitRuleCycle::Visit(void*, const grammar::NonContractingGrammar&) const {
	throw exception::AlibException("Unsupported grammar type NonConctractingGrammar");
}

void NonterminalUnitRuleCycle::Visit(void*, const grammar::ContextPreservingUnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type ContextPreservingUnrestrictedGrammar");
}

void NonterminalUnitRuleCycle::Visit(void*, const grammar::UnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type UnrestrictedGrammar");
}

const NonterminalUnitRuleCycle NonterminalUnitRuleCycle::NONTERMINAL_UNIT_RULE_CYCLE;

} /* namespace properties */

} /* namespace grammar */
