/*
 * EpsilonRemover.cpp
 *
 *  Created on: 24. 11. 2014
 *	  Author: Jan Travnicek
 */

#include "EpsilonRemover.h"

#include "../properties/NullableNonterminals.h"

namespace grammar {

namespace simplify {

void removeNullableNonterminals(grammar::EpsilonFreeCFG& grammar, const std::set<alphabet::Symbol>& nullableNonterminals, const alphabet::Symbol& lhs, const std::vector<alphabet::Symbol>& rhs, unsigned i, std::vector<alphabet::Symbol> clear) {
	if(rhs.size() == i) {
		if(clear.size() == 0) return;
		grammar.addRule(lhs, clear);
	} else if(nullableNonterminals.find(rhs[i]) != nullableNonterminals.end()) {
		removeNullableNonterminals(grammar, nullableNonterminals, lhs, rhs, i+1, clear);
		clear.push_back(rhs[i]);
		removeNullableNonterminals(grammar, nullableNonterminals, lhs, rhs, i+1, clear);
	} else {
		clear.push_back(rhs[i]);
		removeNullableNonterminals(grammar, nullableNonterminals, lhs, rhs, i+1, clear);
	}
}

template<class T>
grammar::EpsilonFreeCFG removeInternal( const T & origGrammar ) {
	grammar::EpsilonFreeCFG grammar(origGrammar.getInitialSymbol());

	for( const auto & symbol : origGrammar.getNonterminalAlphabet() )
		grammar.addNonterminalSymbol( symbol );

	for( const auto & symbol : origGrammar.getTerminalAlphabet() )
		grammar.addTerminalSymbol( symbol );

	std::set<alphabet::Symbol> nullableNonterminals = grammar::properties::NullableNonterminals::getNullableNonterminals(origGrammar);

	for( const auto & rule : origGrammar.getRawRules( ) )
		for(const auto & rhs : rule.second) {
			if(rhs.size() == 0) continue;

			removeNullableNonterminals(grammar, nullableNonterminals, rule.first, rhs, 0, {});
		}

	if(nullableNonterminals.find(origGrammar.getInitialSymbol()) != nullableNonterminals.end())
		grammar.setGeneratesEpsilon(true);

	return grammar;
}

grammar::EpsilonFreeCFG EpsilonRemover::remove(const grammar::CFG& origGrammar)
{
	return removeInternal(origGrammar);
}

grammar::EpsilonFreeCFG EpsilonRemover::remove(const grammar::EpsilonFreeCFG& origGrammar)
{
	return origGrammar;
}

grammar::CNF EpsilonRemover::remove(const grammar::CNF& origGrammar)
{
	return origGrammar;
}

grammar::GNF EpsilonRemover::remove(const grammar::GNF& origGrammar)
{
	return origGrammar;
}

grammar::EpsilonFreeCFG EpsilonRemover::remove(const grammar::LG& origGrammar)
{
	return removeInternal(origGrammar);
}

grammar::EpsilonFreeCFG EpsilonRemover::remove(const grammar::LeftLG& origGrammar)
{
	return removeInternal(origGrammar);
}

grammar::LeftRG EpsilonRemover::remove(const grammar::LeftRG& origGrammar)
{
	return origGrammar;
}

grammar::EpsilonFreeCFG EpsilonRemover::remove(const grammar::RightLG& origGrammar)
{
	return removeInternal(origGrammar);
}

grammar::RightRG EpsilonRemover::remove(const grammar::RightRG& origGrammar)
{
	return origGrammar;
}

grammar::Grammar EpsilonRemover::remove(const grammar::Grammar& grammar) {
	grammar::Grammar* out = NULL;
	grammar.getData().Accept((void*) &out, EpsilonRemover::EPSILON_REMOVER);
	grammar::Grammar res = std::move(*out);
	delete out;
	return res;
}

void EpsilonRemover::Visit(void* data, const grammar::LeftLG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void EpsilonRemover::Visit(void* data, const grammar::LeftRG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void EpsilonRemover::Visit(void* data, const grammar::RightLG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void EpsilonRemover::Visit(void* data, const grammar::RightRG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void EpsilonRemover::Visit(void*, const grammar::LG&) const {
	throw exception::AlibException("Unsupported grammar type LG");
}

void EpsilonRemover::Visit(void* data, const grammar::CFG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void EpsilonRemover::Visit(void* data, const grammar::EpsilonFreeCFG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void EpsilonRemover::Visit(void* data, const grammar::CNF& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void EpsilonRemover::Visit(void* data, const grammar::GNF& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void EpsilonRemover::Visit(void*, const grammar::CSG&) const {
	throw exception::AlibException("Unsupported grammar type CSG");
}

void EpsilonRemover::Visit(void*, const grammar::NonContractingGrammar&) const {
	throw exception::AlibException("Unsupported grammar type NonConctractingGrammar");
}

void EpsilonRemover::Visit(void*, const grammar::ContextPreservingUnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type ContextPreservingUnrestrictedGrammar");
}

void EpsilonRemover::Visit(void*, const grammar::UnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type UnrestrictedGrammar");
}

const EpsilonRemover EpsilonRemover::EPSILON_REMOVER;

} /* namespace simplify */

} /* namespace grammar */
