/*
 * SimpleRulesRemover.cpp
 *
 *  Created on: 24. 11. 2014
 *	  Author: Jan Travnicek
 */

#include "SimpleRulesRemover.h"

#include "../properties/NonterminalUnitRuleCycle.h"

namespace grammar {

namespace simplify {

template<class T>
T removeEpsilonFree( const T & origGrammar ) {
	T grammar(origGrammar.getInitialSymbol());

	for( const auto & symbol : origGrammar.getNonterminalAlphabet() )
		grammar.addNonterminalSymbol( symbol );

	for( const auto & symbol : origGrammar.getTerminalAlphabet() )
		grammar.addTerminalSymbol( symbol );

	auto origRules = origGrammar.getRawRules();

	for( const auto & symbol : origGrammar.getNonterminalAlphabet() ) {
		std::set<alphabet::Symbol> simpleRulesClosure = grammar::properties::NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(origGrammar, symbol);
		for( const auto & closureSymbol : simpleRulesClosure ) {
			auto rules = origRules.find(closureSymbol);
			if(rules != origRules.end()) for( const auto& rawRule : rules->second ) {
				if(rawRule.size() == 0 || (rawRule.size() == 1 && !origGrammar.getNonterminalAlphabet().count(rawRule[0])) || rawRule.size() >= 2)
					grammar.addRawRule(symbol, rawRule);
			}
		}
	}

	grammar.setGeneratesEpsilon(origGrammar.getGeneratesEpsilon());

	return grammar;
}

template<class T>
T removeNonEpsilonFree( const T & origGrammar ) {
	T grammar(origGrammar.getInitialSymbol());

	for( const auto & symbol : origGrammar.getNonterminalAlphabet() )
		grammar.addNonterminalSymbol( symbol );

	for( const auto & symbol : origGrammar.getTerminalAlphabet() )
		grammar.addTerminalSymbol( symbol );

	auto origRules = origGrammar.getRawRules();

	for( const auto & symbol : origGrammar.getNonterminalAlphabet() ) {
		std::set<alphabet::Symbol> simpleRulesClosure = grammar::properties::NonterminalUnitRuleCycle::getNonterminalUnitRuleCycle(origGrammar, symbol);
		for( const auto & closureSymbol : simpleRulesClosure ) {
			auto rules = origRules.find(closureSymbol);
			if(rules != origRules.end()) for( const auto& rawRule : rules->second ) {
				if(rawRule.size() == 0 || (rawRule.size() == 1 && !origGrammar.getNonterminalAlphabet().count(rawRule[0])) || rawRule.size() >= 2)
					grammar.addRawRule(symbol, rawRule);
			}
		}
	}

	return grammar;
}



grammar::CFG SimpleRulesRemover::remove(const grammar::CFG& origGrammar)
{
	return removeNonEpsilonFree(origGrammar);
}

grammar::EpsilonFreeCFG SimpleRulesRemover::remove(const grammar::EpsilonFreeCFG& origGrammar)
{
	return removeEpsilonFree(origGrammar);
}

grammar::CNF SimpleRulesRemover::remove(const grammar::CNF& origGrammar)
{
	return origGrammar;
}

grammar::GNF SimpleRulesRemover::remove(const grammar::GNF& origGrammar)
{
	return origGrammar;
}

grammar::LG SimpleRulesRemover::remove(const grammar::LG& origGrammar)
{
	return removeNonEpsilonFree(origGrammar);
}

grammar::LeftLG SimpleRulesRemover::remove(const grammar::LeftLG& origGrammar)
{
	return removeNonEpsilonFree(origGrammar);
}

grammar::LeftRG SimpleRulesRemover::remove(const grammar::LeftRG& origGrammar)
{
	return origGrammar;
}

grammar::RightLG SimpleRulesRemover::remove(const grammar::RightLG& origGrammar)
{
	return removeNonEpsilonFree(origGrammar);
}

grammar::RightRG SimpleRulesRemover::remove(const grammar::RightRG& origGrammar)
{
	return origGrammar;
}



grammar::Grammar SimpleRulesRemover::remove(const grammar::Grammar& grammar) {
	grammar::Grammar* out = NULL;
	grammar.getData().Accept((void*) &out, SimpleRulesRemover::SIMPLE_RULES_REMOVER);
	grammar::Grammar res = std::move(*out);
	delete out;
	return res;
}

void SimpleRulesRemover::Visit(void* data, const grammar::LeftLG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void SimpleRulesRemover::Visit(void* data, const grammar::LeftRG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void SimpleRulesRemover::Visit(void* data, const grammar::RightLG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void SimpleRulesRemover::Visit(void* data, const grammar::RightRG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void SimpleRulesRemover::Visit(void* data, const grammar::LG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void SimpleRulesRemover::Visit(void* data, const grammar::CFG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void SimpleRulesRemover::Visit(void* data, const grammar::EpsilonFreeCFG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void SimpleRulesRemover::Visit(void* data, const grammar::CNF& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void SimpleRulesRemover::Visit(void* data, const grammar::GNF& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->remove(grammar)));
}

void SimpleRulesRemover::Visit(void*, const grammar::CSG&) const {
	throw exception::AlibException("Unsupported grammar type CSG");
}

void SimpleRulesRemover::Visit(void*, const grammar::NonContractingGrammar&) const {
	throw exception::AlibException("Unsupported grammar type NonConctractingGrammar");
}

void SimpleRulesRemover::Visit(void*, const grammar::ContextPreservingUnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type ContextPreservingUnrestrictedGrammar");
}

void SimpleRulesRemover::Visit(void*, const grammar::UnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type UnrestrictedGrammar");
}

const SimpleRulesRemover SimpleRulesRemover::SIMPLE_RULES_REMOVER;

} /* namespace simplify */

} /* namespace grammar */
