/*
 * ToCNF.cpp
 *
 *  Created on: 24. 11. 2014
 *	  Author: Jan Travnicek
 */

#include "ToCNF.h"

#include "EpsilonRemover.h"
#include "SimpleRulesRemover.h"
#include "alphabet/LabeledSymbol.h"
#include "alphabet/SymbolPairSymbol.h"

namespace grammar {

namespace simplify {

template<class T>
std::pair<alphabet::Symbol, alphabet::Symbol> splitToPairs(T& grammar, const std::vector<alphabet::Symbol>& rhs, unsigned from, unsigned size, std::map<alphabet::Symbol, alphabet::Symbol>& createdSymbols) {
	if(size == 2) {
		return std::make_pair(rhs[from], rhs[from + 1]);
	} else if(size == 3) {
		alphabet::Symbol firstLhs {rhs[from]};

		auto second = splitToPairs(grammar, rhs, from + 1, 2, createdSymbols);
		alphabet::Symbol secondProposal{alphabet::SymbolPairSymbol(second)};
		if(!createdSymbols.count(secondProposal)) {
			createdSymbols.insert(std::make_pair(secondProposal, alphabet::createUniqueSymbol(secondProposal, grammar.getTerminalAlphabet(), grammar.getNonterminalAlphabet())));
		}
		grammar.addNonterminalSymbol(createdSymbols.find(secondProposal)->second);
		grammar.addRawRule(createdSymbols.find(secondProposal)->second, {std::move(second.first), std::move(second.second)});

		return std::make_pair(std::move(firstLhs), createdSymbols.find(secondProposal)->second);
	} else {
		auto first = splitToPairs(grammar, rhs, from, size / 2, createdSymbols);
		alphabet::Symbol firstProposal{alphabet::SymbolPairSymbol(first)};
		if(!createdSymbols.count(firstProposal)) {
			createdSymbols.insert(std::make_pair(firstProposal, alphabet::createUniqueSymbol(firstProposal, grammar.getTerminalAlphabet(), grammar.getNonterminalAlphabet())));
		}
		grammar.addNonterminalSymbol(createdSymbols.find(firstProposal)->second);
		grammar.addRawRule(createdSymbols.find(firstProposal)->second, {std::move(first.first), std::move(first.second)});

		auto second = splitToPairs(grammar, rhs, from + size / 2, size - size / 2, createdSymbols);
		alphabet::Symbol secondProposal{alphabet::SymbolPairSymbol(second)};
		if(!createdSymbols.count(secondProposal)) {
			createdSymbols.insert(std::make_pair(secondProposal, alphabet::createUniqueSymbol(secondProposal, grammar.getTerminalAlphabet(), grammar.getNonterminalAlphabet())));
		}
		grammar.addNonterminalSymbol(createdSymbols.find(secondProposal)->second);
		grammar.addRawRule(createdSymbols.find(secondProposal)->second, {std::move(second.first), std::move(second.second)});

		return std::make_pair(createdSymbols.find(firstProposal)->second, createdSymbols.find(secondProposal)->second);
	}
}

template<class T>
grammar::CNF convertInternal( const T & origGrammar ) {
	T grammarTmp(origGrammar.getInitialSymbol());

	grammarTmp.setNonterminalAlphabet(origGrammar.getNonterminalAlphabet() );
	grammarTmp.setTerminalAlphabet( origGrammar.getTerminalAlphabet() );

	std::map<alphabet::Symbol, alphabet::Symbol> createdSymbols;
	auto origRules = origGrammar.getRawRules();
	for( const auto & origRule : origRules ) {
		for( const auto& origRhs : origRule.second ) {
			if(origRhs.size() == 1 || origRhs.size() == 2)
				grammarTmp.addRawRule(origRule.first, origRhs);
			else if(origRhs.size() > 2) {
				auto second = splitToPairs(grammarTmp, origRhs, 0, origRhs.size(), createdSymbols);
				grammarTmp.addRawRule(origRule.first, {std::move(second.first), std::move(second.second)});
			} else
				throw exception::AlibException("Invalid rule in grammar");
		}
	}

	grammarTmp.setGeneratesEpsilon(origGrammar.getGeneratesEpsilon());

	grammar::CNF grammar(grammarTmp.getInitialSymbol());

	grammar.setNonterminalAlphabet( grammarTmp.getNonterminalAlphabet() );
	grammar.setTerminalAlphabet( grammarTmp.getTerminalAlphabet() );

	std::map<alphabet::Symbol, alphabet::Symbol> terminalToShadowNonterminal;
	for( const auto & symbol : grammarTmp.getTerminalAlphabet() ) {
		alphabet::Symbol shadowSymbol = alphabet::createUniqueSymbol(symbol, grammar.getTerminalAlphabet(), grammar.getNonterminalAlphabet());
		terminalToShadowNonterminal.insert( std::make_pair( symbol, shadowSymbol ));
		grammar.addNonterminalSymbol( shadowSymbol );
		grammar.addRule(std::move(shadowSymbol), symbol);
	}

	auto tmpRules = grammarTmp.getRawRules();
	for( const auto & tmpRule : tmpRules ) {
		for( const auto& tmpRhs : tmpRule.second ) {
			if(tmpRhs.size() == 2) {
				if(grammarTmp.getNonterminalAlphabet().count(tmpRhs[0])) {
					if(grammarTmp.getNonterminalAlphabet().count(tmpRhs[1])) {
						grammar.addRawRule(tmpRule.first, tmpRhs);
					} else {
						grammar.addRawRule(tmpRule.first, {tmpRhs[0], terminalToShadowNonterminal.find(tmpRhs[1])->second});
					}
				} else {
					if(grammarTmp.getNonterminalAlphabet().count(tmpRhs[1])) {
						grammar.addRawRule(tmpRule.first, {terminalToShadowNonterminal.find(tmpRhs[0])->second, tmpRhs[1]});
					} else {
						grammar.addRawRule(tmpRule.first, {terminalToShadowNonterminal.find(tmpRhs[0])->second, terminalToShadowNonterminal.find(tmpRhs[1])->second });
					}
				}
			} else // tmpRhs.size() == 1
				grammar.addRawRule(tmpRule.first, tmpRhs);
		}
	}

	grammar.setGeneratesEpsilon(grammarTmp.getGeneratesEpsilon());

	return grammar;
}



grammar::CNF ToCNF::convert(const grammar::CFG& origGrammar)
{
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::EpsilonRemover::remove(origGrammar)));
}

grammar::CNF ToCNF::convert(const grammar::EpsilonFreeCFG& origGrammar)
{
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(origGrammar));
}

grammar::CNF ToCNF::convert(const grammar::CNF& origGrammar)
{
	return origGrammar;
}

grammar::CNF ToCNF::convert(const grammar::GNF& origGrammar)
{
	return convertInternal(origGrammar);
}

grammar::CNF ToCNF::convert(const grammar::LG& origGrammar)
{
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::EpsilonRemover::remove(origGrammar)));
}

grammar::CNF ToCNF::convert(const grammar::LeftLG& origGrammar)
{
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::EpsilonRemover::remove(origGrammar)));
}

grammar::CNF ToCNF::convert(const grammar::LeftRG& origGrammar)
{
	return convertInternal(origGrammar);
}

grammar::CNF ToCNF::convert(const grammar::RightLG& origGrammar)
{
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::EpsilonRemover::remove(origGrammar)));
}

grammar::CNF ToCNF::convert(const grammar::RightRG& origGrammar)
{
	return convertInternal(origGrammar);
}



grammar::Grammar ToCNF::convert(const grammar::Grammar& grammar) {
	grammar::Grammar* out = NULL;
	grammar.getData().Accept((void*) &out, ToCNF::TO_CNF);
	grammar::Grammar res = std::move(*out);
	delete out;
	return res;
}

void ToCNF::Visit(void* data, const grammar::LeftLG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToCNF::Visit(void* data, const grammar::LeftRG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToCNF::Visit(void* data, const grammar::RightLG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToCNF::Visit(void* data, const grammar::RightRG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToCNF::Visit(void* data, const grammar::LG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToCNF::Visit(void* data, const grammar::CFG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToCNF::Visit(void* data, const grammar::EpsilonFreeCFG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToCNF::Visit(void* data, const grammar::CNF& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToCNF::Visit(void* data, const grammar::GNF& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToCNF::Visit(void*, const grammar::CSG&) const {
	throw exception::AlibException("Unsupported grammar type CSG");
}

void ToCNF::Visit(void*, const grammar::NonContractingGrammar&) const {
	throw exception::AlibException("Unsupported grammar type NonConctractingGrammar");
}

void ToCNF::Visit(void*, const grammar::ContextPreservingUnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type ContextPreservingUnrestrictedGrammar");
}

void ToCNF::Visit(void*, const grammar::UnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type UnrestrictedGrammar");
}

const ToCNF ToCNF::TO_CNF;

} /* namespace simplify */

} /* namespace grammar */
