/*
 * ToGNF.cpp
 *
 *  Created on: 24. 11. 2014
 *	  Author: Jan Travnicek
 */

#include "ToGNF.h"

#include "EpsilonRemover.h"
#include "SimpleRulesRemover.h"
#include "LeftRecursionRemover.h"
#include "../convert/ToGrammarRightRG.h"
#include "alphabet/LabeledSymbol.h"
#include "alphabet/SymbolPairSymbol.h"

namespace grammar {

namespace simplify {

grammar::EpsilonFreeCFG assignNonterminals(const grammar::EpsilonFreeCFG& origGrammar) {
	grammar::EpsilonFreeCFG res(origGrammar.getInitialSymbol());
	res.setNonterminalAlphabet(origGrammar.getNonterminalAlphabet());
	res.setTerminalAlphabet(origGrammar.getTerminalAlphabet());
	res.setGeneratesEpsilon(origGrammar.getGeneratesEpsilon());

	for(const std::pair<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>>& rule : origGrammar.getRules()) {
		for(const std::vector<alphabet::Symbol>& singleRHS : rule.second) {
			if(res.getTerminalAlphabet().count(singleRHS[0])) { //do not substitute terminals
				res.addRule(rule.first, singleRHS);
				continue;
			}
			const alphabet::Symbol& secondLHS = singleRHS[0];
			if(origGrammar.getRules().find(secondLHS) == origGrammar.getRules().end()) { //is there any right hand side to substitue with?
				//if not well this rule does not generate anything anyway
				continue;
			}

			for(const std::vector<alphabet::Symbol>& secondSingleRHS : origGrammar.getRules().find(secondLHS)->second) { // do the substitution
				std::vector<alphabet::Symbol> newRHS(secondSingleRHS);
				newRHS.insert(newRHS.end(), singleRHS.begin() + 1, singleRHS.end());
				res.addRule(rule.first, newRHS);
			}
		}
	}
	return res;
}

grammar::GNF convertInternal( const grammar::EpsilonFreeCFG & origGrammar ) {
	grammar::EpsilonFreeCFG step(origGrammar);
	while(true) {
		grammar::EpsilonFreeCFG nextStep = assignNonterminals(step);

		if(step == nextStep) break;
		step = std::move(nextStep);
	}

	grammar::GNF res(step.getInitialSymbol());
	res.setTerminalAlphabet(step.getTerminalAlphabet());
	res.setNonterminalAlphabet(step.getNonterminalAlphabet());
	res.setGeneratesEpsilon(step.getGeneratesEpsilon());
	std::map<alphabet::Symbol, alphabet::Symbol> terminalToPrimed;
	for(const alphabet::Symbol& terminal : step.getTerminalAlphabet()) {
		alphabet::Symbol primed = alphabet::createUniqueSymbol(terminal, res.getTerminalAlphabet(), res.getNonterminalAlphabet());
		terminalToPrimed.insert(std::make_pair(terminal, primed));
		res.addNonterminalSymbol(primed);
		res.addRule(primed, std::make_pair(terminal, std::vector<alphabet::Symbol> {}));
	}
	for(const auto& rule : step.getRules()) {
		for(const auto& rhs : rule.second) {
			std::vector<alphabet::Symbol> convertedNonterminals;
			bool first = true;
			for(const alphabet::Symbol& rhsSymbol : rhs) {
				if(first) {
					first = false;
					continue;
				}

				if(res.getNonterminalAlphabet().count(rhsSymbol))
					convertedNonterminals.push_back(rhsSymbol);
				else
					convertedNonterminals.push_back(terminalToPrimed.find(rhsSymbol)->second);
			}
			res.addRule(rule.first, std::make_pair(rhs[0], std::move(convertedNonterminals)));
		}
	}
	return res;
}



grammar::GNF ToGNF::convert(const grammar::CFG& origGrammar)
{
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::LeftRecursionRemover::remove(grammar::simplify::EpsilonRemover::remove(origGrammar))));
}

grammar::GNF ToGNF::convert(const grammar::EpsilonFreeCFG& origGrammar)
{
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::LeftRecursionRemover::remove(origGrammar)));
}

grammar::GNF ToGNF::convert(const grammar::CNF& origGrammar)
{
	return convertInternal(grammar::simplify::LeftRecursionRemover::remove(origGrammar));
}

grammar::GNF ToGNF::convert(const grammar::GNF& origGrammar)
{
	return origGrammar;
}

grammar::GNF ToGNF::convert(const grammar::LG& origGrammar)
{
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::LeftRecursionRemover::remove(grammar::simplify::EpsilonRemover::remove(origGrammar))));
}

grammar::GNF ToGNF::convert(const grammar::LeftLG& origGrammar)
{
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::LeftRecursionRemover::remove(grammar::simplify::EpsilonRemover::remove(origGrammar))));
}

grammar::RightRG ToGNF::convert(const grammar::LeftRG& origGrammar)
{
	return convert::ToGrammarRightRG::convert(origGrammar);
}

grammar::GNF ToGNF::convert(const grammar::RightLG& origGrammar)
{
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::LeftRecursionRemover::remove(grammar::simplify::EpsilonRemover::remove(origGrammar))));
}

grammar::RightRG ToGNF::convert(const grammar::RightRG& origGrammar)
{
	return origGrammar;
}



grammar::Grammar ToGNF::convert(const grammar::Grammar& grammar) {
	grammar::Grammar* out = NULL;
	grammar.getData().Accept((void*) &out, ToGNF::TO_GNF);
	grammar::Grammar res = std::move(*out);
	delete out;
	return res;
}

void ToGNF::Visit(void* data, const grammar::LeftLG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToGNF::Visit(void* data, const grammar::LeftRG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToGNF::Visit(void* data, const grammar::RightLG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToGNF::Visit(void* data, const grammar::RightRG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToGNF::Visit(void* data, const grammar::LG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToGNF::Visit(void* data, const grammar::CFG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToGNF::Visit(void* data, const grammar::EpsilonFreeCFG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToGNF::Visit(void* data, const grammar::CNF& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToGNF::Visit(void* data, const grammar::GNF& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new Grammar(std::move(this->convert(grammar)));
}

void ToGNF::Visit(void*, const grammar::CSG&) const {
	throw exception::AlibException("Unsupported grammar type CSG");
}

void ToGNF::Visit(void*, const grammar::NonContractingGrammar&) const {
	throw exception::AlibException("Unsupported grammar type NonConctractingGrammar");
}

void ToGNF::Visit(void*, const grammar::ContextPreservingUnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type ContextPreservingUnrestrictedGrammar");
}

void ToGNF::Visit(void*, const grammar::UnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type UnrestrictedGrammar");
}

const ToGNF ToGNF::TO_GNF;

} /* namespace simplify */

} /* namespace grammar */
