/*
 * UnproductiveSymbolsRemover.cpp
 *
 *  Created on: 22. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "UnproductiveSymbolsRemover.h"

#include <grammar/ContextFree/CFG.h>
#include <grammar/ContextFree/EpsilonFreeCFG.h>
#include <grammar/ContextFree/GNF.h>
#include <grammar/ContextFree/CNF.h>
#include <grammar/ContextFree/LG.h>
#include <grammar/Regular/LeftLG.h>
#include <grammar/Regular/LeftRG.h>
#include <grammar/Regular/RightLG.h>
#include <grammar/Regular/RightRG.h>

#include <set>

#include "../properties/ProductiveNonterminals.h"

namespace grammar {

namespace simplify {

template<class T>
T UnproductiveSymbolsRemover::remove( const T & grammar ) {
	// 1.
	std::set<alphabet::Symbol> Nt = grammar::properties::ProductiveNonterminals::getProductiveNonterminals( grammar );

	T ret(grammar.getInitialSymbol( ) );

	for( const auto & symbol : Nt )
		ret.addNonterminalSymbol( symbol );

	for( const auto & symbol : grammar.getTerminalAlphabet( ) )
		ret.addTerminalSymbol( symbol );

	const std::set<alphabet::Symbol> & terminals = ret.getTerminalAlphabet( );
	for( const auto & rule : grammar.getRawRules( ) ) {
		if( Nt.count( rule.first ) ) {
			for( const auto & rhs : rule.second ) {
				if( all_of( rhs.begin( ), rhs.end( ), [ Nt, terminals ]( const alphabet::Symbol & symbol ) {
						return Nt.count( symbol ) || terminals.count( symbol );
				} ) )
					ret.addRawRule( rule.first, rhs );
			}
		}
	}


	/* if( ! G1.getNonTerminalSymbols( ) . count( grammar.getInitialSymbol( ) ) )
		throw AlibException( "Starting symbol of grammar was marked as unproductive and therefore it was removed." ); */

	// 2.
	return ret;
}

template grammar::CFG UnproductiveSymbolsRemover::remove( const grammar::CFG & grammar );
template grammar::EpsilonFreeCFG UnproductiveSymbolsRemover::remove( const grammar::EpsilonFreeCFG & grammar );
template grammar::GNF UnproductiveSymbolsRemover::remove( const grammar::GNF & grammar );
template grammar::CNF UnproductiveSymbolsRemover::remove( const grammar::CNF & grammar );
template grammar::LG UnproductiveSymbolsRemover::remove( const grammar::LG & grammar );
template grammar::LeftLG UnproductiveSymbolsRemover::remove( const grammar::LeftLG & grammar );
template grammar::LeftRG UnproductiveSymbolsRemover::remove( const grammar::LeftRG & grammar );
template grammar::RightLG UnproductiveSymbolsRemover::remove( const grammar::RightLG & grammar );
template grammar::RightRG UnproductiveSymbolsRemover::remove( const grammar::RightRG & grammar );

grammar::Grammar UnproductiveSymbolsRemover::remove(const grammar::Grammar& grammar) {
	grammar::Grammar* out = NULL;
	grammar.getData().Accept((void*) &out, UnproductiveSymbolsRemover::UNPRODUCTIVE_SYMBOLS_REMOVER);
	grammar::Grammar res = std::move(*out);
	delete out;
	return res;
}

void UnproductiveSymbolsRemover::Visit(void* data, const grammar::LeftLG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnproductiveSymbolsRemover::Visit(void* data, const grammar::LeftRG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnproductiveSymbolsRemover::Visit(void* data, const grammar::RightLG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnproductiveSymbolsRemover::Visit(void* data, const grammar::RightRG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnproductiveSymbolsRemover::Visit(void* data, const grammar::LG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnproductiveSymbolsRemover::Visit(void* data, const grammar::CFG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnproductiveSymbolsRemover::Visit(void* data, const grammar::EpsilonFreeCFG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnproductiveSymbolsRemover::Visit(void* data, const grammar::CNF& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnproductiveSymbolsRemover::Visit(void* data, const grammar::GNF& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnproductiveSymbolsRemover::Visit(void*, const grammar::CSG&) const {
	throw exception::AlibException("Unsupported grammar type CSG");
}

void UnproductiveSymbolsRemover::Visit(void*, const grammar::NonContractingGrammar&) const {
	throw exception::AlibException("Unsupported grammar type NonConctractingGrammar");
}

void UnproductiveSymbolsRemover::Visit(void*, const grammar::ContextPreservingUnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type ContextPreservingUnrestrictedGrammar");
}

void UnproductiveSymbolsRemover::Visit(void*, const grammar::UnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type UnrestrictedGrammar");
}

const UnproductiveSymbolsRemover UnproductiveSymbolsRemover::UNPRODUCTIVE_SYMBOLS_REMOVER;

} /* namespace simplify */

} /* namespace grammar */

