/*
 * UnreachableSymbolsRemover.cpp
 *
 *  Created on: 22. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "UnreachableSymbolsRemover.h"

#include <grammar/ContextFree/CFG.h>
#include <grammar/ContextFree/EpsilonFreeCFG.h>
#include <grammar/ContextFree/GNF.h>
#include <grammar/ContextFree/CNF.h>
#include <grammar/ContextFree/LG.h>
#include <grammar/Regular/LeftLG.h>
#include <grammar/Regular/LeftRG.h>
#include <grammar/Regular/RightLG.h>
#include <grammar/Regular/RightRG.h>

#include <set>

#include "../properties/UnreachableSymbols.h"

namespace grammar {

namespace simplify {

template<class T>
T UnreachableSymbolsRemover::remove( const T & grammar) {
	// 1.
	std::set<alphabet::Symbol> Vt = grammar::properties::UnreachableSymbols::getUnreachableSymbols( grammar );

	T ret(grammar.getInitialSymbol( ) );

	std::set<alphabet::Symbol> newNonTerminals, newTerminals;

	set_intersection( Vt.begin( ), Vt.end( ), grammar.getNonterminalAlphabet( ).begin( ), grammar.getNonterminalAlphabet( ).end( ), std::inserter( newNonTerminals, newNonTerminals.begin( ) ) );
	for( const auto & symbol : newNonTerminals )
		ret.addNonterminalSymbol( symbol );

	set_intersection( Vt.begin( ), Vt.end( ), grammar.getTerminalAlphabet( ).begin( ), grammar.getTerminalAlphabet( ).end( ), std::inserter( newTerminals, newTerminals.begin( ) ) );
	for( const auto & symbol : newTerminals )
		ret.addTerminalSymbol( symbol );

	// A->\alpha: if A \in N' and \alpha in V_i*, then A->\alpha in P
	for( const auto & rule : grammar.getRawRules( ) ) {
		if( newNonTerminals.count( rule.first ) ) {
			for( const auto& rhs : rule.second ) {
				if( all_of( rhs.begin( ), rhs.end( ), [ Vt ]( alphabet::Symbol const& symb ) -> bool {
					return Vt.count( symb );
				} ) )
					ret.addRawRule( rule.first, rhs );
			}
	}	
	}

	// 2.
	return ret;
}

template grammar::CFG UnreachableSymbolsRemover::remove( const grammar::CFG & grammar );
template grammar::EpsilonFreeCFG UnreachableSymbolsRemover::remove( const grammar::EpsilonFreeCFG & grammar );
template grammar::GNF UnreachableSymbolsRemover::remove( const grammar::GNF & grammar );
template grammar::CNF UnreachableSymbolsRemover::remove( const grammar::CNF & grammar );
template grammar::LG UnreachableSymbolsRemover::remove( const grammar::LG & grammar );
template grammar::LeftLG UnreachableSymbolsRemover::remove( const grammar::LeftLG & grammar );
template grammar::LeftRG UnreachableSymbolsRemover::remove( const grammar::LeftRG & grammar );
template grammar::RightLG UnreachableSymbolsRemover::remove( const grammar::RightLG & grammar );
template grammar::RightRG UnreachableSymbolsRemover::remove( const grammar::RightRG & grammar );

grammar::Grammar UnreachableSymbolsRemover::remove(const grammar::Grammar& grammar) {
	grammar::Grammar* out = NULL;
	grammar.getData().Accept((void*) &out, UnreachableSymbolsRemover::UNREACHABLE_SYMBOLS_REMOVER);
	grammar::Grammar res = std::move(*out);
	delete out;
	return res;
}

void UnreachableSymbolsRemover::Visit(void* data, const grammar::LeftLG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnreachableSymbolsRemover::Visit(void* data, const grammar::LeftRG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnreachableSymbolsRemover::Visit(void* data, const grammar::RightLG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnreachableSymbolsRemover::Visit(void* data, const grammar::RightRG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnreachableSymbolsRemover::Visit(void* data, const grammar::LG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnreachableSymbolsRemover::Visit(void* data, const grammar::CFG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnreachableSymbolsRemover::Visit(void* data, const grammar::EpsilonFreeCFG& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnreachableSymbolsRemover::Visit(void* data, const grammar::CNF& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnreachableSymbolsRemover::Visit(void* data, const grammar::GNF& grammar) const {
	grammar::Grammar* & out = *((grammar::Grammar**) data);
	out = new grammar::Grammar(this->remove(grammar));
}

void UnreachableSymbolsRemover::Visit(void*, const grammar::CSG&) const {
	throw exception::AlibException("Unsupported grammar type CSG");
}

void UnreachableSymbolsRemover::Visit(void*, const grammar::NonContractingGrammar&) const {
	throw exception::AlibException("Unsupported grammar type NonConctractingGrammar");
}

void UnreachableSymbolsRemover::Visit(void*, const grammar::ContextPreservingUnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type ContextPreservingUnrestrictedGrammar");
}

void UnreachableSymbolsRemover::Visit(void*, const grammar::UnrestrictedGrammar&) const {
	throw exception::AlibException("Unsupported grammar type UnrestrictedGrammar");
}

const UnreachableSymbolsRemover UnreachableSymbolsRemover::UNREACHABLE_SYMBOLS_REMOVER;

} /* namespace simplify */

} /* namespace grammar */

