#include "BellmanFord.h"

#include <exception/AlibException.h>

namespace graph
{

namespace shortestpath
{

class Data
{
public:
	Node start;
	BellmanFord::Result out;
};

typedef std::unordered_map<Node, std::unordered_map<Node, int>> weights_t;

static BellmanFord::Result bellmanford_impl(const DirectedGraph &graph, const Node &start)
{
	BellmanFord::Result d; // distances

	d[start] = 0;

	for (unsigned i = 0; i < graph.getNodes().size(); ++i) {
		for (const DirectedEdge &e : graph.getEdges()) {
			int ew = graph.getEdgeValue(e);
			const Node &u = e.getFromNode();
			const Node &v = e.getToNode();

			auto du = d.find(u);
			auto dv = d.find(v);

			if (du == d.end()) {
				continue;
			}

			if (dv == d.end() || du->second + ew < dv->second) {
				d[v] = du->second + ew;
			}
		}
	}

	for (const DirectedEdge &e : graph.getEdges()) {
		const Node &u = e.getFromNode();
		const Node &v = e.getToNode();

		if (d.at(u) + graph.getEdgeValue(e) < d.at(v)) {
			throw exception::AlibException("BellmanFord: Found negative-weight cycle!");
		}
	}

	return d;
}

static BellmanFord::Result bellmanford_impl(const UndirectedGraph &graph, const Node &start)
{
	BellmanFord::Result d; // distances

	d[start] = 0;

	for (unsigned i = 0; i < graph.getNodes().size(); ++i) {
		for (const UndirectedEdge &e : graph.getEdges()) {
			int ew = graph.getEdgeValue(e);
			const Node &u = e.getFirstNode();
			const Node &v = e.getSecondNode();

			auto f = [&ew, &d](const Node &u, const Node &v) {
				auto du = d.find(u);
				auto dv = d.find(v);

				if (du == d.end()) {
					return;
				}

				if (dv == d.end() || du->second + ew < dv->second) {
					d[v] = du->second + ew;
				}
			};

			f(u, v);
			f(v, u);
		}
	}

	for (const UndirectedEdge &e : graph.getEdges()) {
		const Node &u = e.getFirstNode();
		const Node &v = e.getSecondNode();

		auto f = [&e, &d, &graph](const Node &u, const Node &v) {
			if (d.at(u) + graph.getEdgeValue(e) < d.at(v)) {
				throw exception::AlibException("BellmanFord: Found negative-weight cycle!");
			}
		};

		f(u, v);
		f(v, u);

	}

	return d;
}

BellmanFord::Result BellmanFord::bellmanford(const Graph &graph, const Node &start)
{
	Data data = {start, BellmanFord::Result {}};
	graph.getData().Accept(static_cast<void*>(&data), BELLMAN_FORD);
	return data.out;
}

BellmanFord::Result BellmanFord::bellmanford(const DirectedGraph &graph, const Node &start)
{
	return bellmanford_impl(graph, start);
}

BellmanFord::Result BellmanFord::bellmanford(const UndirectedGraph &graph, const Node &start)
{
	return bellmanford_impl(graph, start);
}

void BellmanFord::Visit(void *data, const DirectedGraph &graph) const
{
	Data d = *static_cast<Data*>(data);
	d.out = bellmanford(graph, d.start);
}

void BellmanFord::Visit(void *data, const UndirectedGraph &graph) const
{
	Data d = *static_cast<Data*>(data);
	d.out = bellmanford(graph, d.start);
}

const BellmanFord BellmanFord::BELLMAN_FORD;

} // namespace shortestpath

} // namespace graph
