#include "Dijkstra.h"

#include <exception/AlibException.h>

namespace graph
{

namespace shortestpath
{

class Data
{
public:
	Node start;
	Dijkstra::Result out;
};

class Value
{
public:
	Value(const Node &node, int value)
		: node(node)
		, value(value)
	{
	}

	bool operator<(const Value &other) const
	{
		return value < other.value;
	}

	Node node;
	int value;
};

typedef std::unordered_map<Node, std::unordered_map<Node, int>> weights_t;

static weights_t weights(const DirectedGraph &g)
{
	int val;
	weights_t w;

	for (const DirectedEdge &e : g.getEdges()) {
		auto &map = w[e.getFromNode()];
		auto search = map.find(e.getToNode());

		if (search == map.end()) {
			val = g.getEdgeValue(e);
			map[e.getToNode()] = val;
		} else {
			val = std::min(search->second, g.getEdgeValue(e));
			search->second = val;
		}

		if (val < 0) {
			throw exception::AlibException("Dijkstra: Found negative value!");
		}
	}

	return w;
}

static weights_t weights(const UndirectedGraph &g)
{
	int val;
	weights_t w;

	for (const UndirectedEdge &e : g.getEdges()) {
		auto &map = w[e.getFirstNode()];
		auto search = map.find(e.getSecondNode());

		if (search == map.end()) {
			val = g.getEdgeValue(e);
			map[e.getSecondNode()] = val;
			w[e.getSecondNode()][e.getFirstNode()] = val;
		} else {
			val = std::min(search->second, g.getEdgeValue(e));
			search->second = val;
			w[e.getSecondNode()][e.getFirstNode()] = val;
		}

		if (val < 0) {
			throw exception::AlibException("Dijkstra: Found negative value!");
		}
	}

	return w;
}

template <typename T>
static Dijkstra::Result dijkstra_impl(const T &graph, const Node &start)
{
	std::set<Value> q;            // priority queue
	Dijkstra::Result d;           // distances
	weights_t w = weights(graph); // minimum weights

	d[start] = 0;
	q.insert(Value(start, 0));

	while (!q.empty()) {
		Node u = q.begin()->node; q.erase(q.begin());

		for (const Node &v : graph.neighbors(u)) {
			int val = d.at(u) + w[u][v];
			auto search = d.find(v);

			if (search == d.end() || search->second > val) {
				if (search != d.end()) {
					q.erase(q.find(Value(v, search->second)));
				}

				q.insert(Value(v, val));
				d[v] = val;
			}
		}
	}

	return d;
}

Dijkstra::Result Dijkstra::dijkstra(const Graph &graph, const Node &start)
{
	Data data = {start, Dijkstra::Result {}};
	graph.getData().Accept(static_cast<void*>(&data), DIJKSTRA);
	return data.out;
}

Dijkstra::Result Dijkstra::dijkstra(const DirectedGraph &graph, const Node &start)
{
	return dijkstra_impl(graph, start);
}

Dijkstra::Result Dijkstra::dijkstra(const UndirectedGraph &graph, const Node &start)
{
	return dijkstra_impl(graph, start);
}

void Dijkstra::Visit(void *data, const DirectedGraph &graph) const
{
	Data d = *static_cast<Data*>(data);
	d.out = dijkstra(graph, d.start);
}

void Dijkstra::Visit(void *data, const UndirectedGraph &graph) const
{
	Data d = *static_cast<Data*>(data);
	d.out = dijkstra(graph, d.start);
}

const Dijkstra Dijkstra::DIJKSTRA;

} // namespace shortestpath

} // namespace graph
