#include "FloydWarshall.h"

#include <exception/AlibException.h>

#include <limits>

namespace graph
{

namespace shortestpath
{

static void init_distances(const DirectedGraph &g, FloydWarshall::Result &d)
{
	for (const DirectedEdge &e : g.getEdges()) {
		int &val = d[e.getFromNode()][e.getToNode()];
		val = std::min(val, g.getEdgeValue(e));
	}
}

static void init_distances(const UndirectedGraph &g, FloydWarshall::Result &d)
{
	for (const UndirectedEdge &e : g.getEdges()) {
		int &val = d[e.getFirstNode()][e.getSecondNode()];
		val = std::min(val, g.getEdgeValue(e));
		d[e.getSecondNode()][e.getFirstNode()] = val;
	}
}

template <typename T>
static FloydWarshall::Result floydwarshall_impl(const T &graph)
{
	FloydWarshall::Result d;
	const std::set<Node> &nodes = graph.getNodes();
	const int inf = std::numeric_limits<int>::max();

	for (const Node &n : nodes)
		for (const Node &m : nodes)
			d[n][m] = inf;

	for (const Node &n : nodes)
		d[n][n] = 0;

	init_distances(graph, d);

	for (const Node &k : nodes)
		for (const Node &i : nodes)
			if (d[i][k] != inf)
				for (const Node &j : nodes)
					if (d[k][j] != inf)
						d[i][j] = std::min(d[i][j], d[i][k] + d[k][j]);

	for (const Node &n : nodes)
		if (d[n][n] < 0)
			throw exception::AlibException("FloydWarshall: Detected negative-weight cycle");

	return d;
}

FloydWarshall::Result FloydWarshall::floydwarshall(const Graph &graph)
{
	Result res;
	graph.getData().Accept(static_cast<void*>(&res), FLOYD_WARSHALL);
	return res;
}

FloydWarshall::Result FloydWarshall::floydwarshall(const DirectedGraph &graph)
{
	return floydwarshall_impl(graph);
}

FloydWarshall::Result FloydWarshall::floydwarshall(const UndirectedGraph &graph)
{
	return floydwarshall_impl(graph);
}

void FloydWarshall::Visit(void *data, const DirectedGraph &graph) const
{
	Result *r = static_cast<Result*>(data);
	*r = floydwarshall(graph);
}

void FloydWarshall::Visit(void *data, const UndirectedGraph &graph) const
{
	Result *r = static_cast<Result*>(data);
	*r = floydwarshall(graph);
}

const FloydWarshall FloydWarshall::FLOYD_WARSHALL;

} // namespace shortestpath

} // namespace graph
