/*
 * RandomRegExpFactory.cpp
 *
 *  Created on: 27. 3. 2014
 *	  Author: Jan Travnicek
 */

#include "RandomRegExpFactory.h"
#include <regexp/unbounded/UnboundedRegExpElements.h>

#include <algorithm>
#include <random>

namespace regexp {

namespace generate {

regexp::UnboundedRegExp RandomRegExpFactory::generateUnboundedRegExp( size_t leafNodes, size_t height, size_t alphabetSize, bool randomizedAlphabet ) {
	srand( time( NULL ) );

	if(alphabetSize > 26)
		throw exception::AlibException("Too big alphabet.");

	if( alphabetSize <= 0 )
		throw exception::AlibException( "Alphabet size must be greater than 0." );

	std::vector<alphabet::Symbol> symbols;
	for(int i = 0; i < 26; i++) symbols.push_back(alphabet::symbolFrom(i + 'a'));
	if(randomizedAlphabet) shuffle(symbols.begin(), symbols.end(), std::random_devices::semirandom);
	std::set<alphabet::Symbol> alphabet(symbols.begin(), symbols.begin() + alphabetSize);

	return RandomRegExpFactory::generateUnboundedRegExp( leafNodes, height, alphabet );
}

regexp::UnboundedRegExp RandomRegExpFactory::generateUnboundedRegExp( size_t leafNodes, size_t height, std::set<alphabet::Symbol> alphabet) {

	if( alphabet.size() > 26)
		throw exception::AlibException("Too big alphabet.");

	if( alphabet.size() <= 0 )
		throw exception::AlibException( "Alphabet size must be greater than 0." );

	std::vector<const regexp::UnboundedRegExpElement*> elems;

	{
		elems.push_back(new regexp::UnboundedRegExpEmpty());
		elems.push_back(new regexp::UnboundedRegExpEpsilon());
	}
	if(alphabet.size() > 6) {
		elems.push_back(new regexp::UnboundedRegExpEmpty());
		elems.push_back(new regexp::UnboundedRegExpEpsilon());
	}
	if(alphabet.size() > 16) {
		elems.push_back(new regexp::UnboundedRegExpEmpty());
		elems.push_back(new regexp::UnboundedRegExpEpsilon());
	}

	for(const alphabet::Symbol& symbol : alphabet) {
		elems.push_back(new regexp::UnboundedRegExpSymbol(symbol));
	}

	regexp::UnboundedRegExp res = RandomRegExpFactory::SimpleUnboundedRegExp( leafNodes, height, elems );

	for(const regexp::UnboundedRegExpElement* elem : elems ) {
		delete elem;
	}
	return res;
}

regexp::UnboundedRegExp RandomRegExpFactory::SimpleUnboundedRegExp( size_t n, size_t h, const std::vector<const regexp::UnboundedRegExpElement*> & elems) {

	const regexp::UnboundedRegExpElement* elem = SimpleUnboundedRegExpElement(n, h, elems);
	regexp::UnboundedRegExp res(std::move(*elem));
	delete elem;
	return res;
}

const regexp::UnboundedRegExpElement* RandomRegExpFactory::SimpleUnboundedRegExpElement(size_t n, size_t h, const std::vector<const regexp::UnboundedRegExpElement*> & elems) {
	if(h == 0 || n == 0) {
		return elems[ std::random_devices::semirandom() % elems.size( ) ]->clone();
	} else {
		unsigned childNodes = std::random_devices::semirandom() % 10;
		if(childNodes <  4) childNodes = 1;
		else if(childNodes <  6) childNodes = 2;
		else if(childNodes <  8) childNodes = 3;
		else childNodes = 4;

		childNodes = childNodes > n ? n : childNodes;

		int subSizes[4];
		if(childNodes == 4) {
			subSizes[3] = std::random_devices::semirandom() % ( n - 1 );
			subSizes[2] = std::random_devices::semirandom() % ( n - subSizes[3] - 1 );
			subSizes[1] = std::random_devices::semirandom() % ( n - subSizes[2] - subSizes [3] - 1 );

			subSizes[3] += 1;
			subSizes[2] += 1;
			subSizes[1] += 1;

			subSizes[0] = n - subSizes[1] - subSizes[2] - subSizes[3];
		}
		if(childNodes == 3) {
			subSizes[2] = std::random_devices::semirandom() % ( n - 1);
			subSizes[1] = std::random_devices::semirandom() % ( n - subSizes[2] - 1);

			subSizes[2] += 1;
			subSizes[1] += 1;

			subSizes[0] = n - subSizes[1] - subSizes[2];
		}
		if(childNodes == 2) {
			subSizes[1] = std::random_devices::semirandom() % ( n - 1 );

			subSizes[1] += 1;

			subSizes[0] = n - subSizes[1];
		}
		if(childNodes == 1) {
			const regexp::UnboundedRegExpElement* subElem = SimpleUnboundedRegExpElement(n, h - 1, elems);
			regexp::UnboundedRegExpIteration iter(std::move(*subElem));
			delete subElem;
			return new regexp::UnboundedRegExpIteration(iter);
		}

		int nodeType = std::random_devices::semirandom() % 2;
		if(nodeType == 0) {
			regexp::UnboundedRegExpConcatenation con;
			for(unsigned i = 0; i < childNodes; i++) {
				const regexp::UnboundedRegExpElement* subElem = SimpleUnboundedRegExpElement(subSizes[i], h - 1, elems);
				con.appendElement(std::move(*subElem));
				delete subElem;
			}
			return new regexp::UnboundedRegExpConcatenation(con);
		} else {
			regexp::UnboundedRegExpAlternation con;
			for(unsigned i = 0; i < childNodes; i++) {
				const regexp::UnboundedRegExpElement* subElem = SimpleUnboundedRegExpElement(subSizes[i], h - 1, elems);
				con.appendElement(std::move(*subElem));
				delete subElem;
			}
			return new regexp::UnboundedRegExpAlternation(con);
		}
		
	}
}

} /* namespace generate */

} /* namespace regexp */
