/*
 * RegExpIntegral.cpp
 *
 *  Created on: 24. 2. 2014
 *	  Author: Tomas Pecka
 */

#include "RegExpIntegral.h"
#include <regexp/unbounded/UnboundedRegExpElements.h>
#include <regexp/formal/FormalRegExpElements.h>

namespace regexp
{

regexp::RegExp RegExpIntegral::integral(const regexp::RegExp& regexp, const string::LinearString& string)
{
	std::pair<regexp::RegExp*, const string::LinearString> out(nullptr, string);
	regexp.getData().Accept((void*) &out, RegExpIntegral::REGEXP_INTEGRAL);
	regexp::RegExp res = std::move(*(out.first));
	delete out.first;
	return res;
}

regexp::FormalRegExp RegExpIntegral::integral(const regexp::FormalRegExp& regexp, const string::LinearString& string)
{
	const regexp::FormalRegExpElement* newRegExp = regexp.getRegExp().clone();

	for(auto it = string.getContent().begin(); it != string.getContent().end(); it ++)
	{
		const auto& symbol = *it;

		std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> out(symbol, nullptr);
		newRegExp->Accept((void*) &out, RegExpIntegral::REGEXP_INTEGRAL);
		delete newRegExp;
		newRegExp = out.second;
	}

	regexp::FormalRegExp res(std::move(*newRegExp));
	delete newRegExp;
	return res;
}

regexp::UnboundedRegExp RegExpIntegral::integral(const regexp::UnboundedRegExp& regexp, const string::LinearString& string)
{
	const regexp::UnboundedRegExpElement* newRegExp = regexp.getRegExp().clone();

	for(auto it = string.getContent().begin(); it != string.getContent().end(); it ++)
	{
		const auto& symbol = *it;

		std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> out(symbol, nullptr);
		newRegExp->Accept((void*) &out, RegExpIntegral::REGEXP_INTEGRAL);
		delete newRegExp;
		newRegExp = out.second;
	}

	regexp::UnboundedRegExp res(std::move(*newRegExp));
	delete newRegExp;
	return res;
}

// ----------------------------------------------------------------------------

void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExp& regexp) const
{
	std::pair<regexp::RegExp*, const string::LinearString>& out = *((std::pair<regexp::RegExp*, const string::LinearString>*) userData);
	out.first = new regexp::RegExp(this->integral(regexp, out.second));
}

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExp& regexp) const
{
	std::pair<regexp::RegExp*, const string::LinearString>& out = *((std::pair<regexp::RegExp*, const string::LinearString>*) userData);
	out.first = new regexp::RegExp(this->integral(regexp, out.second));
}

// ----------------------------------------------------------------------------

void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExpAlternation& alternation) const
{
	std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::FormalRegExpElement*>*) userData);

	static_cast<const regexp::FormalRegExpElement&>(alternation.getLeftElement()).Accept(userData, *this);
	const regexp::FormalRegExpElement* leftIntegral = out.second;

	static_cast<const regexp::FormalRegExpElement&>(alternation.getRightElement()).Accept(userData, *this);
	const regexp::FormalRegExpElement* rightIntegral = out.second;

	out.second = new regexp::FormalRegExpAlternation(std::move(*leftIntegral), std::move(*rightIntegral));
	delete leftIntegral;
	delete rightIntegral;
}

void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExpConcatenation& concatenation) const
{
	std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::FormalRegExpElement*>*) userData);

	regexp::FormalRegExpSymbol* symb = new regexp::FormalRegExpSymbol(out.first);
	regexp::FormalRegExpElement* concat = concatenation.clone();
	regexp::FormalRegExpConcatenation* ret = new regexp::FormalRegExpConcatenation(std::move(*symb), std::move(*concat));

	delete concat;
	delete symb;

	out.second = ret;
}

void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExpIteration& iteration) const
{
	std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::FormalRegExpElement*>*) userData);

	regexp::FormalRegExpSymbol* symb = new regexp::FormalRegExpSymbol(out.first);
	regexp::FormalRegExpElement* iter = iteration.clone();
	regexp::FormalRegExpConcatenation* ret = new regexp::FormalRegExpConcatenation(std::move(*symb), std::move(*iter));

	delete iter;
	delete symb;

	out.second = ret;
}

void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExpSymbol& symbol) const
{
	std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::FormalRegExpElement*>*) userData);

	regexp::FormalRegExpSymbol* symb = new regexp::FormalRegExpSymbol(out.first);
	regexp::FormalRegExpElement* symb2 = symbol.clone();
	regexp::FormalRegExpConcatenation* ret = new regexp::FormalRegExpConcatenation(std::move(*symb), std::move(*symb2));

	delete symb2;
	delete symb;

	out.second = ret;
}


void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExpEpsilon&) const
{
	std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::FormalRegExpElement*>*) userData);
	regexp::FormalRegExpSymbol* symb = new regexp::FormalRegExpSymbol(out.first);
	out.second = symb;
}

void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExpEmpty&) const
{
	std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::FormalRegExpElement*>*) userData);
	out.second = new regexp::FormalRegExpEmpty();
}

// ----------------------------------------------------------------------------

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExpAlternation& alternation) const
{
	std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*>*) userData);
	regexp::UnboundedRegExpAlternation* ret = new regexp::UnboundedRegExpAlternation();

	for(const auto& child : alternation.getElements())
	{
		static_cast<const regexp::UnboundedRegExpElement&>(*child).Accept(userData, *this);
		ret->appendElement(std::move(*out.second));
		delete out.second;
	}

	out.second = ret;
}

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExpConcatenation& concatenation) const
{
	std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*>*) userData);

	regexp::UnboundedRegExpSymbol* symb = new regexp::UnboundedRegExpSymbol(out.first);
	regexp::UnboundedRegExpElement* concat = concatenation.clone();
	regexp::UnboundedRegExpConcatenation* ret = new regexp::UnboundedRegExpConcatenation();
	ret->appendElement(std::move(*symb));
	ret->appendElement(std::move(*concat));

	delete concat;
	delete symb;

	out.second = ret;
}

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExpIteration& iteration) const
{
	std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*>*) userData);

	regexp::UnboundedRegExpSymbol* symb = new regexp::UnboundedRegExpSymbol(out.first);
	regexp::UnboundedRegExpElement* iter = iteration.clone();
	regexp::UnboundedRegExpConcatenation* ret = new regexp::UnboundedRegExpConcatenation();
	ret->appendElement(std::move(*symb));
	ret->appendElement(std::move(*iter));

	delete iter;
	delete symb;

	out.second = ret;
}

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExpSymbol& symbol) const
{
	std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*>*) userData);

	regexp::UnboundedRegExpSymbol* symb = new regexp::UnboundedRegExpSymbol(out.first);
	regexp::UnboundedRegExpElement* symb2 = symbol.clone();
	regexp::UnboundedRegExpConcatenation* ret = new regexp::UnboundedRegExpConcatenation();
	ret->appendElement(std::move(*symb));
	ret->appendElement(std::move(*symb2));

	delete symb2;
	delete symb;

	out.second = ret;
}

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExpEpsilon&) const
{
	std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*>*) userData);

	regexp::UnboundedRegExpSymbol* symb = new regexp::UnboundedRegExpSymbol(out.first);
	out.second = symb;
}

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExpEmpty&) const
{
	std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*>*) userData);
	out.second = new regexp::UnboundedRegExpEmpty();
}

const RegExpIntegral RegExpIntegral::REGEXP_INTEGRAL;

} /* namespace regexp */
