#include "DfsTest.h"

#include "graph/traverse/Dfs.h"

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( GraphDfsTest, "graph" );
CPPUNIT_TEST_SUITE_REGISTRATION( GraphDfsTest );

void GraphDfsTest::testTraverseAll()
{
	// Common
	int counter;
	graph::Node n1("n1");
	graph::Node n2("n2");
	graph::Node n3("n3");
	graph::Node n4("n4");
	graph::Node n5("n5");
	graph::Node n6("n6");

	// Directed
	graph::DirectedGraph dg;
	dg.addEdge(graph::DirectedEdge(n1, n2));
	dg.addEdge(graph::DirectedEdge(n1, n2, "multi-edge"));
	dg.addEdge(graph::DirectedEdge(n2, n3));
	dg.addEdge(graph::DirectedEdge(n3, n4));
	dg.addEdge(graph::DirectedEdge(n3, n6));
	dg.addEdge(graph::DirectedEdge(n4, n1));
	dg.addEdge(graph::DirectedEdge(n4, n5));

	counter = 0;
	graph::traverse::Dfs::dfs(dg, n1, [&](const graph::Node &) {
		counter++;
		return true;
	});

	CPPUNIT_ASSERT_EQUAL(6, counter);

	// Undirected
	graph::UndirectedGraph ug;
	ug.addEdge(graph::UndirectedEdge(n1, n2));
	ug.addEdge(graph::UndirectedEdge(n1, n2, "multi-edge"));
	ug.addEdge(graph::UndirectedEdge(n2, n3));
	ug.addEdge(graph::UndirectedEdge(n3, n4));
	ug.addEdge(graph::UndirectedEdge(n3, n6));
	ug.addEdge(graph::UndirectedEdge(n4, n1));
	ug.addEdge(graph::UndirectedEdge(n4, n5));

	counter = 0;
	graph::traverse::Dfs::dfs(ug, n1, [&](const graph::Node &) {
		counter++;
		return true;
	});

	CPPUNIT_ASSERT_EQUAL(6, counter);
}

void GraphDfsTest::testEarlyReturn()
{
	// Common
	int counter;
	graph::Node n1("n1");
	graph::Node n2("n2");
	graph::Node n3("n3");
	graph::Node n4("n4");

	// Directed
	graph::DirectedGraph dg;
	dg.addEdge(graph::DirectedEdge(n1, n2));
	dg.addEdge(graph::DirectedEdge(n1, n2, "multi-edge"));
	dg.addEdge(graph::DirectedEdge(n2, n3));
	dg.addEdge(graph::DirectedEdge(n3, n4));

	counter = 0;
	graph::traverse::Dfs::dfs(dg, n1, [&](const graph::Node &) {
		counter++;
		if (counter == 2) {
			return false;
		}
		return true;
	});

	CPPUNIT_ASSERT_EQUAL(2, counter);

	// Undirected
	graph::UndirectedGraph ug;
	ug.addEdge(graph::UndirectedEdge(n1, n2));
	ug.addEdge(graph::UndirectedEdge(n1, n2, "multi-edge"));
	ug.addEdge(graph::UndirectedEdge(n2, n3));
	ug.addEdge(graph::UndirectedEdge(n3, n4));

	counter = 0;
	graph::traverse::Dfs::dfs(ug, n1, [&](const graph::Node &) {
		counter++;
		if (counter == 2) {
			return false;
		}
		return true;
	});

	CPPUNIT_ASSERT_EQUAL(2, counter);
}

void GraphDfsTest::testDisconnectedGraph()
{
	// Common
	int counter;
	graph::Node n1("n1");
	graph::Node n2("n2");
	graph::Node n3("n3");
	graph::Node n4("n4");
	graph::Node n5("n5");

	// Directed
	graph::DirectedGraph dg;
	dg.addNode(n5);
	dg.addEdge(graph::DirectedEdge(n1, n2));
	dg.addEdge(graph::DirectedEdge(n1, n2, "multi-edge"));
	dg.addEdge(graph::DirectedEdge(n2, n3));
	dg.addEdge(graph::DirectedEdge(n3, n4));

	counter = 0;
	graph::traverse::Dfs::dfs(dg, n1, [&](const graph::Node &) {
		counter++;
		return true;
	});

	CPPUNIT_ASSERT_EQUAL(4, counter);

	counter = 0;
	graph::traverse::Dfs::dfs(dg, n5, [&](const graph::Node &) {
		counter++;
		return true;
	});

	CPPUNIT_ASSERT_EQUAL(1, counter);

	// Undirected
	graph::UndirectedGraph ug;
	ug.addNode(n5);
	ug.addEdge(graph::UndirectedEdge(n1, n2));
	ug.addEdge(graph::UndirectedEdge(n1, n2, "multi-edge"));
	ug.addEdge(graph::UndirectedEdge(n2, n3));
	ug.addEdge(graph::UndirectedEdge(n3, n4));

	counter = 0;
	graph::traverse::Dfs::dfs(ug, n1, [&](const graph::Node &) {
		counter++;
		return true;
	});

	CPPUNIT_ASSERT_EQUAL(4, counter);

	counter = 0;
	graph::traverse::Dfs::dfs(ug, n5, [&](const graph::Node &) {
		counter++;
		return true;
	});

	CPPUNIT_ASSERT_EQUAL(1, counter);
}

void GraphDfsTest::testDfs2()
{
	// Common
	int counter;
	std::unordered_map<graph::Node, int> opened;
	std::unordered_map<graph::Node, int> closed;
	std::unordered_map<graph::Node, graph::Node> predecessors;

	graph::Node n1("n1");
	graph::Node n2("n2");
	graph::Node n3("n3");
	graph::Node n4("n4");
	graph::Node n5("n5");

	// Directed
	graph::DirectedGraph dg;
	dg.addEdge(graph::DirectedEdge(n1, n2));
	dg.addEdge(graph::DirectedEdge(n1, n3));
	dg.addEdge(graph::DirectedEdge(n1, n4));
	dg.addEdge(graph::DirectedEdge(n2, n5));

	counter = 0;
	opened.clear();
	closed.clear();
	predecessors.clear();

	graph::traverse::Dfs::dfs(dg, n1, [&](const graph::Node &n, const graph::Node &p, int o, int c) {
		counter++;
		opened[n] = o;
		closed[n] = c;
		predecessors.insert(std::make_pair(n, p));
		return true;
	});

	CPPUNIT_ASSERT_EQUAL(5, counter);
	CPPUNIT_ASSERT_EQUAL(4, closed[n5]);

	CPPUNIT_ASSERT_EQUAL(n1, predecessors.find(n1)->second);
	CPPUNIT_ASSERT_EQUAL(n1, predecessors.find(n2)->second);
	CPPUNIT_ASSERT_EQUAL(n1, predecessors.find(n3)->second);
	CPPUNIT_ASSERT_EQUAL(n1, predecessors.find(n4)->second);
	CPPUNIT_ASSERT_EQUAL(n2, predecessors.find(n5)->second);

	// Undirected
	graph::UndirectedGraph ug;
	ug.addEdge(graph::UndirectedEdge(n1, n2));
	ug.addEdge(graph::UndirectedEdge(n1, n3));
	ug.addEdge(graph::UndirectedEdge(n1, n4));
	ug.addEdge(graph::UndirectedEdge(n2, n5));

	counter = 0;
	opened.clear();
	closed.clear();
	predecessors.clear();

	graph::traverse::Dfs::dfs(ug, n1, [&](const graph::Node &n, const graph::Node &p, int o, int c) {
		counter++;
		opened[n] = o;
		closed[n] = c;
		predecessors.insert(std::make_pair(n, p));
		return true;
	});

	CPPUNIT_ASSERT_EQUAL(5, counter);
	CPPUNIT_ASSERT_EQUAL(4, closed[n5]);

	CPPUNIT_ASSERT_EQUAL(n1, predecessors.find(n1)->second);
	CPPUNIT_ASSERT_EQUAL(n1, predecessors.find(n2)->second);
	CPPUNIT_ASSERT_EQUAL(n1, predecessors.find(n3)->second);
	CPPUNIT_ASSERT_EQUAL(n1, predecessors.find(n4)->second);
	CPPUNIT_ASSERT_EQUAL(n2, predecessors.find(n5)->second);
}

