/*
 * DFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "DFA.h"
#include "../AutomatonException.h"
#include <ostream>
#include <sstream>

#include "../../sax/FromXMLParserHelper.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace automaton {

DFA::DFA(State initialState) : SingleInitialState(std::move(initialState)) {

}

AutomatonBase* DFA::clone() const {
	return new DFA(*this);
}

AutomatonBase* DFA::plunder() && {
	return new DFA(std::move(*this));
}

bool DFA::removeState(const State& state) {
	if (initialState == state) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (const std::pair<std::pair<State, alphabet::Symbol>, State>& t : transitions) {
		if (t.first.first == state || t.second == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool DFA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<std::pair<State, alphabet::Symbol>, State>& transition : transitions) {
		if (transition.first.second == symbol)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used.");
	}

	return inputAlphabet.erase(symbol);

}

bool DFA::addTransition(State from, alphabet::Symbol input, State to) {
	if (states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");

	if (inputAlphabet.find(input) == inputAlphabet.end())
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");

	if (states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");

	std::pair<State, alphabet::Symbol> key = std::make_pair(std::move(from), std::move(input));

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == to)
			return false;
		else
			throw AutomatonException(
				"Transition from this state and symbol already exists (\"" + (std::string) key.first.getName() + "\", \"" + (std::string) key.second + "\") -> \"" + (std::string) to.getName()
						+ "\".");
	}

	transitions.insert(std::make_pair(std::move(key), std::move(to)));
	return true;
}

bool DFA::removeTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::pair<State, alphabet::Symbol> key = std::make_pair(from, input);

	if (transitions.find(key) == transitions.end())
		return false;

	if(transitions.find(key)->second != to)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	transitions.erase(key);
	return true;
}

const std::map<std::pair<State, alphabet::Symbol>, State>& DFA::getTransitions() const {
	return transitions;
}

std::map<std::pair<State, alphabet::Symbol>, State> DFA::getTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::map<std::pair<State, alphabet::Symbol>, State> transitionsFromState;
	for (const std::pair<std::pair<State, alphabet::Symbol>, State>& transition : transitions) {
		if (transition.first.first == from) {
			transitionsFromState.insert(transition);
		}
	}

	return transitionsFromState;
}

std::map<std::pair<State, alphabet::Symbol>, State> DFA::getTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<std::pair<State, alphabet::Symbol>, State> transitionsToState;
	for (const std::pair<std::pair<State, alphabet::Symbol>, State>& transition : transitions) {
		if (transition.second == to) {
			transitionsToState.insert(transition);
		}
	}

	return transitionsToState;
}

bool DFA::isTotal() const {
	return transitions.size() == inputAlphabet.size() * states.size();
}

int DFA::compare(const DFA& other) const {
	auto first = std::tie(states, inputAlphabet, initialState, finalStates, transitions);
	auto second = std::tie(other.states, other.inputAlphabet, other.initialState, other.finalStates, other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void DFA::operator>>(std::ostream& out) const {
	out << "(DFA "
		<< " states = " << states
		<< " inputAlphabet = " << inputAlphabet
		<< " initialState = " << initialState
		<< " finalStates = " << finalStates
		<< " transitions = " << transitions
		<< ")";
}

DFA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string DFA::XML_TAG_NAME = "DFA";

DFA DFA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, DFA::XML_TAG_NAME);

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::Symbol> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet(input);
	State initialState = AutomatonFromXMLParser::parseInitialState(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	DFA automaton(std::move(initialState));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<DFA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, DFA::XML_TAG_NAME);
	return automaton;
}

void DFA::parseTransition(std::deque<sax::Token>::iterator& input, DFA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	State from = AutomatonFromXMLParser::parseTransitionFrom(input);
	alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
	State to = AutomatonFromXMLParser::parseTransitionTo(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(to));
}

void DFA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(DFA::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(DFA::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void DFA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

xmlApi<automaton::Automaton>::ParserRegister<automaton::DFA> DFAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::DFA>(automaton::DFA::XML_TAG_NAME, automaton::DFA::parse);
xmlApi<alib::Object>::ParserRegister<automaton::DFA> DFAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::DFA>(automaton::DFA::XML_TAG_NAME, automaton::DFA::parse);

} /* namespace alib */
