/*
 * EpsilonNFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "EpsilonNFA.h"
#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"
#include "../AutomatonException.h"
#include "../Automaton.h"
#include <ostream>
#include <sstream>

#include "../../sax/FromXMLParserHelper.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"
#include "../../CastApi.hpp"

namespace automaton {

EpsilonNFA::EpsilonNFA(State initialState) : SingleInitialState(std::move(initialState)) {

}

EpsilonNFA::EpsilonNFA(const MultiInitialStateNFA& other) : SingleInitialState(automaton::createUniqueState(automaton::State("q0"), other.getStates())) {
	inputAlphabet = other.getInputAlphabet();
	states = other.getStates();
	finalStates = other.getFinalStates();
	for(const auto& transition : other.getTransitions()) {
		std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> > key = std::make_pair(transition.first.first, std::variant<string::Epsilon, alphabet::Symbol>(transition.first.second));
		transitions[key] = transition.second;
	}
	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> > key = std::make_pair(this->getInitialState(), std::variant<string::Epsilon, alphabet::Symbol>(string::Epsilon::EPSILON));
	transitions[key] = other.getInitialStates();

}

EpsilonNFA::EpsilonNFA(const NFA& other) : SingleInitialState(other.getInitialState()) {
	inputAlphabet = other.getInputAlphabet();
	states = other.getStates();
	finalStates = other.getFinalStates();
	for(const auto& transition : other.getTransitions()) {
		std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> > key = std::make_pair(transition.first.first, std::variant<string::Epsilon, alphabet::Symbol>(transition.first.second));
		transitions[key] = transition.second;
	}
}

EpsilonNFA::EpsilonNFA(const DFA& other) : SingleInitialState(other.getInitialState()) {
	inputAlphabet = other.getInputAlphabet();
	states = other.getStates();
	finalStates = other.getFinalStates();
	for(const auto& transition : other.getTransitions()) {
		std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> > key = std::make_pair(transition.first.first, std::variant<string::Epsilon, alphabet::Symbol>(transition.first.second));
		transitions[key].insert(transition.second);
	}
}

AutomatonBase* EpsilonNFA::clone() const {
	return new EpsilonNFA(*this);
}

AutomatonBase* EpsilonNFA::plunder() && {
	return new EpsilonNFA(std::move(*this));
}

bool EpsilonNFA::removeState(const State& state) {
	if (initialState == state) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& transition : transitions) {
		if (transition.first.first == state || transition.second.find(state) != transition.second.end())
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool EpsilonNFA::removeInputSymbol(const alphabet::Symbol& symbol) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(symbol);
	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& transition : transitions) {
		if (transition.first.second.is<alphabet::Symbol>() && transition.first.second.get<alphabet::Symbol>() == symbol)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used.");
	}

	return inputAlphabet.erase(symbol);
}

bool EpsilonNFA::addTransition(State from, std::variant<string::Epsilon, alphabet::Symbol> input, State to) {
	if (states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");

	if (input.is<alphabet::Symbol>() && inputAlphabet.find(input.get<alphabet::Symbol>()) == inputAlphabet.end())
		throw AutomatonException("Input symbol \"" + (std::string) input.get<alphabet::Symbol>() + "\" doesn't exist.");

	if (states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");

	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> > key = std::make_pair(std::move(from), std::move(input));

	return transitions[std::move(key)].insert(std::move(to)).second;
}

bool EpsilonNFA::addTransition(State from, alphabet::Symbol input, State to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(std::move(input));
	return addTransition(std::move(from), std::move(inputVariant), std::move(to));
}

bool EpsilonNFA::addTransition(State from, State to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	return addTransition(std::move(from), std::move(inputVariant), std::move(to));
}

bool EpsilonNFA::removeTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(input);
	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> > key = std::make_pair(from, inputVariant);

	return transitions[key].erase(to);
}

bool EpsilonNFA::removeTransition(const State& from, const State& to) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant(string::Epsilon::EPSILON);
	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> > key = std::make_pair(from, inputVariant);

	return transitions[key].erase(to);
}

const std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& EpsilonNFA::getTransitions() const {
	return transitions;
}

std::map<State, std::set<State> > EpsilonNFA::getEpsilonTransitions() const {
	std::map<State, std::set<State> > result;
	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& transition : transitions) {
		if (transition.first.second.is<string::Epsilon>())
			result[transition.first.first].insert(transition.second.begin(), transition.second.end());
	}
	return result;
}

std::map<std::pair<State, alphabet::Symbol>, std::set<State> > EpsilonNFA::getSymbolTransitions() const {
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > result;
	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& transition : transitions) {
		if (transition.first.second.is<alphabet::Symbol>())
			result[std::pair<State, alphabet::Symbol>(transition.first.first, transition.first.second.get<alphabet::Symbol>())].insert(transition.second.begin(), transition.second.end());
	}
	return result;
}

std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> > EpsilonNFA::getTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> > transitionsFromState;
	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& transition : transitions) {
		if (transition.first.first == from)
			transitionsFromState[transition.first].insert(transition.second.begin(), transition.second.end());
	}

	return transitionsFromState;
}

std::map<State, std::set<State> > EpsilonNFA::getEpsilonTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> > key(from, std::variant<string::Epsilon, alphabet::Symbol>(string::Epsilon::EPSILON));
	std::map<State, std::set<State>> res;
	auto transition = transitions.find(key);
	if(transition != transitions.end())
		res.insert(std::make_pair(from, transition->second));
	return res;
}

std::map<std::pair<State, alphabet::Symbol>, std::set<State> > EpsilonNFA::getSymbolTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > transitionsFromState;
	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& transition : transitions) {
		if (transition.first.first == from && transition.first.second.is<alphabet::Symbol>())
			transitionsFromState[std::pair<State, alphabet::Symbol>(transition.first.first, transition.first.second.get<alphabet::Symbol>())].insert(transition.second.begin(), transition.second.end());
	}

	return transitionsFromState;
}

std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> > EpsilonNFA::getTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> > transitionsToState;
	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& transition : transitions) {
		if (transition.second.find(to) != transition.second.end())
			transitionsToState[transition.first].insert(transition.second.begin(), transition.second.end());
	}

	return transitionsToState;
}

std::map<State, std::set<State> > EpsilonNFA::getEpsilonTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<State, std::set<State> > transitionsToState;
	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& transition : transitions) {
		if (transition.second.find(to) != transition.second.end() && transition.first.second.is<string::Epsilon>())
			transitionsToState[transition.first.first].insert(transition.second.begin(), transition.second.end());
	}

	return transitionsToState;
}

std::map<std::pair<State, alphabet::Symbol>, std::set<State> > EpsilonNFA::getSymbolTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > transitionsToState;
	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& transition : transitions) {
		if (transition.second.find(to) != transition.second.end() && transition.first.second.is<alphabet::Symbol>())
			transitionsToState[std::pair<State, alphabet::Symbol>(transition.first.first, transition.first.second.get<alphabet::Symbol>())].insert(transition.second.begin(), transition.second.end());
	}

	return transitionsToState;
}

bool EpsilonNFA::isEpsilonFree() const {
	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& transition : transitions) {
		if (transition.first.second.is<string::Epsilon>())
			return false;
	}
	return true;
}

bool EpsilonNFA::isDeterministic() const {
	for (const std::pair<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& transition : transitions) {
		if (transition.second.size() != 1 || transition.second.size() != 0)
			return false;
	}

	return true;
}

bool EpsilonNFA::isTotal() const {
	return isDeterministic() && transitions.size() == inputAlphabet.size() * states.size();
}

int EpsilonNFA::compare(const EpsilonNFA& other) const {
	auto first = std::tie(states, inputAlphabet, initialState, finalStates, transitions);
	auto second = std::tie(other.states, other.inputAlphabet, other.initialState, other.finalStates, other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void EpsilonNFA::operator>>(std::ostream& out) const {
	out << "(EpsilonNFA "
		<< " states = " << states
		<< " inputAlphabet = " << inputAlphabet
		<< " initialState = " << initialState
		<< " finalStates = " << finalStates
		<< " transitions = " << transitions
		<< ")";
}

EpsilonNFA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string EpsilonNFA::XML_TAG_NAME = "EpsilonNFA";

EpsilonNFA EpsilonNFA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, EpsilonNFA::XML_TAG_NAME);

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::Symbol> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet(input);
	State initialState = AutomatonFromXMLParser::parseInitialState(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	EpsilonNFA automaton(std::move(initialState));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<EpsilonNFA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, EpsilonNFA::XML_TAG_NAME);
	return automaton;
}

void EpsilonNFA::parseTransition(std::deque<sax::Token>::iterator& input, EpsilonNFA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	State from = AutomatonFromXMLParser::parseTransitionFrom(input);
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol(input);
	State to = AutomatonFromXMLParser::parseTransitionTo(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputVariant), std::move(to));
}

void EpsilonNFA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(EpsilonNFA::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(EpsilonNFA::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void EpsilonNFA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		for(const auto& targetState: transition.second) {
			out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, targetState);

			out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
		}
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

xmlApi<automaton::Automaton>::ParserRegister<automaton::EpsilonNFA> epsilonNFAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::EpsilonNFA>(automaton::EpsilonNFA::XML_TAG_NAME, automaton::EpsilonNFA::parse);
xmlApi<alib::Object>::ParserRegister<automaton::EpsilonNFA> epsilonNFAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::EpsilonNFA>(automaton::EpsilonNFA::XML_TAG_NAME, automaton::EpsilonNFA::parse);

auto EpsilonNFAFromDFA = castApi::CastRegister<automaton::EpsilonNFA, automaton::DFA>();
auto EpsilonNFAFromNFA = castApi::CastRegister<automaton::EpsilonNFA, automaton::NFA>();
auto EpsilonNFAFromMultiInitialStateNFA = castApi::CastRegister<automaton::EpsilonNFA, automaton::MultiInitialStateNFA>();
auto EpsilonNFACastBinder = castApi::CastPoolStringBinder<automaton::EpsilonNFA>(automaton::EpsilonNFA::XML_TAG_NAME);

} /* namespace alib */
