/*
 * EpsilonNFA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef EPSILON_NFA_H_
#define EPSILON_NFA_H_

#include <map>
#include <variant>
#include "../AutomatonBase.h"
#include "../common/SingleInitialState.h"
#include "../common/InputAlphabet.h"
#include "../../alphabet/Symbol.h"
#include "../../string/Epsilon.h"
#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"

namespace automaton {

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton with epsilon transitions.
 */
class EpsilonNFA : public std::acceptor<EpsilonNFA, VisitableAutomatonBase, std::acceptor<EpsilonNFA, alib::VisitableObjectBase, AutomatonBase> >, public SingleInitialState, public InputAlphabet {
protected:
	std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> > transitions;
public:
	explicit EpsilonNFA(State initialState);
	explicit EpsilonNFA(const MultiInitialStateNFA& other);
	explicit EpsilonNFA(const NFA& other);
	explicit EpsilonNFA(const DFA& other);

	virtual AutomatonBase* clone() const;

	virtual AutomatonBase* plunder() &&;

	/**
	 * @copydoc Automaton::removeState(const State&)
	 */
	virtual bool removeState(const State& state);

	/**
	 * @copydoc Automaton::removeInputSymbol(const Symbol&)
	 */
	virtual bool removeInputSymbol(const alphabet::Symbol& symbol);

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition(State from, std::variant<string::Epsilon, alphabet::Symbol> input, State to);

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition(State from, alphabet::Symbol input, State to);

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition(State from, State to);

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition(const State& current, const alphabet::Symbol& input, const State& next);

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition(const State& current, const State& next);

	/**
	 * @return automaton transitions
	 */
	const std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> >& getTransitions() const;

	/**
	 * @return automaton epsilon transitions
	 */
	std::map<State, std::set<State> > getEpsilonTransitions() const;

	/**
	 * @return automaton transitions reading some symbol
	 */
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > getSymbolTransitions() const;


	/**
	 * @return automaton transitions from state
	 */
	std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> > getTransitionsFromState(const State& from) const;

	/**
	 * @return automaton epsilon transitions from state
	 */
	std::map<State, std::set<State> > getEpsilonTransitionsFromState(const State& from) const;

	/**
	 * @return automaton transitions from state reading some symbols
	 */
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > getSymbolTransitionsFromState(const State& from) const;

	/**
	 * @return automaton transitions to state
	 */
	std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol> >, std::set<State> > getTransitionsToState(const State& from) const;

	/**
	 * @return automaton epsilon transitions to state
	 */
	std::map<State, std::set<State> > getEpsilonTransitionsToState(const State& to) const;

	/**
	 * @return automaton transitions to state reading some symbols
	 */
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > getSymbolTransitionsToState(const State& to) const;

	/**
	 * Determines whether NFA is without epsilon transitions.
	 * @return true when automaton doesn't contain epsilon transitions, false otherwise
	 */
	bool isEpsilonFree() const;

	/**
	 * Determines whether NFA is deterministic.
	 * FA is deterministic if and only if:
	 * \li \c is epsilon free.
	 * \li \c size of transition function \delta (from state, input symbol) \leq 1
	 * @return true when automaton is deterministic, false otherwise
	 */
	bool isDeterministic() const;

	/**
	 * Determines whether NFA is total deterministic
	 * FA is deterministic if and only if
	 * \li \c it is deterministic
	 * \li \c size of transition function \forall states and \forall input symbols \delta (from state, input symbol) = 1
	 * @return true when automaton is total deterministic, false otherwise
	 */
	bool isTotal() const;

	virtual int compare(const ObjectBase& other) const {
		return ObjectBase::compare(*this, other);
	}

	virtual int compare(const EpsilonNFA& other) const;

	virtual void operator>>(std::ostream& os) const;

	virtual explicit operator std::string() const;

	virtual long long selfTypeId() const {
		return typeId(*this);
	}

	const static std::string XML_TAG_NAME;

	static EpsilonNFA parse(std::deque<sax::Token>::iterator& input);
	static void parseTransition(std::deque<sax::Token>::iterator& input, EpsilonNFA& automaton);

	void compose(std::deque<sax::Token>& out) const;
	void composeTransitions(std::deque<sax::Token>& out) const;
};

} /* namespace automaton */

namespace std {

template<>
struct compare<automaton::EpsilonNFA> {
	int operator()(const automaton::EpsilonNFA& first, const automaton::EpsilonNFA& second) const {
		return first.compare(second);
	}
};

} /* namespace std */

#endif /* EPSILON_NFA_H_ */
