/*
 * SinglePopNPDA.h
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef SINGLE_POP_NPDA_H_
#define SINGLE_POP_NPDA_H_

#include <set>
#include <map>
#include <vector>
#include <variant>
#include "../AutomatonBase.h"
#include "../common/SingleInitialState.h"
#include "../common/SingleInitialSymbolPushdownStoreAlphabet.h"
#include "../common/InputAlphabet.h"
#include "../../alphabet/Symbol.h"
#include "../../string/Epsilon.h"

namespace automaton {

/**
 * Push Down Automaton
 */
class SinglePopNPDA: public std::acceptor<SinglePopNPDA, VisitableAutomatonBase, std::acceptor<SinglePopNPDA, alib::VisitableObjectBase, AutomatonBase> >, public SingleInitialSymbolPushdownStoreAlphabet, public SingleInitialState, public InputAlphabet {
protected:
	std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::set<std::pair<State, std::vector<alphabet::Symbol> > > > transitions;
public:
	explicit SinglePopNPDA(State initialState, alphabet::Symbol initialPushdownSymbol);

	virtual AutomatonBase* clone() const;

	virtual AutomatonBase* plunder() &&;

	/**
	 * @copydoc Automaton::removeState(const State&)
	 */
	virtual bool removeState(const State& state);

	/**
	 * @copydoc Automaton::removeInputSymbol(const Symbol&)
	 */
	virtual bool removeInputSymbol(const alphabet::Symbol& symbol);

	/**
	 * @copydoc Automaton::removeStackSymbol(const Symbol&)
	 */
	virtual bool removeStackSymbol(const alphabet::Symbol& symbol);

	/**
	 * Adds transition to the SinglePopNPDA.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the SinglePopNPDA (state, input symbol, stack symbol) or when transition already exists
	 */
	bool addTransition(State from, std::variant<string::Epsilon, alphabet::Symbol> input, alphabet::Symbol pop, State to, std::vector<alphabet::Symbol> push);

	bool addTransition(State from, alphabet::Symbol input, alphabet::Symbol pop, State to, std::vector<alphabet::Symbol> push);

	bool addTransition(State from, alphabet::Symbol pop, State to, std::vector<alphabet::Symbol> push);

	/**
	 * Removes the transition from the SinglePopNPDA.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the SinglePopNPDA
	 */
	bool removeTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const alphabet::Symbol& pop, const State& to, const std::vector<alphabet::Symbol>& push);

	bool removeTransition(const State& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& to, const std::vector<alphabet::Symbol>& push);

	bool removeTransition(const State& from, const alphabet::Symbol& pop, const State& to, const std::vector<alphabet::Symbol>& push);

	/**
	 * @return SinglePopNPDA transitions
	 */
	const std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::set<std::pair<State, std::vector<alphabet::Symbol> > > >& getTransitions() const;

	virtual int compare(const ObjectBase& other) const {
		return ObjectBase::compare(*this, other);
	}

	virtual int compare(const SinglePopNPDA& other) const;

	virtual void operator>>(std::ostream& os) const;

	virtual explicit operator std::string() const;

	virtual long long selfTypeId() const {
		return typeId(*this);
	}

	const static std::string XML_TAG_NAME;

	static SinglePopNPDA parse(std::deque<sax::Token>::iterator& input);
	static void parseTransition(std::deque<sax::Token>::iterator& input, SinglePopNPDA& automaton);

	void compose(std::deque<sax::Token>& out) const;
	void composeTransitions(std::deque<sax::Token>& out) const;
};

} /* namespace automaton */

namespace std {

template<>
struct compare<automaton::SinglePopNPDA> {
	int operator()(const automaton::SinglePopNPDA& first, const automaton::SinglePopNPDA& second) const {
		return first.compare(second);
	}
};

} /* namespace std */

#endif /* SINGLE_POP_NPDA_H_ */
