/*
 * DFTA.h
 *
 *  Created on: Apr 14, 2015
 *      Author: Stepan Plachy
 */

#ifndef DFTA_H_
#define DFTA_H_

#include <map>
#include <vector>
#include "../AutomatonBase.h"
#include "../common/States.h"
//#include "../common/InputAlphabet.h"
#include "../../alphabet/RankedSymbol.h"

namespace automaton {

/**
 * Represents Finite Tree Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class DFTA : public std::acceptor<DFTA, VisitableAutomatonBase, std::acceptor<DFTA, alib::VisitableObjectBase, AutomatonBase> >, public States {
protected:
	std::set<alphabet::RankedSymbol> inputAlphabet;
	std::map<std::pair<alphabet::RankedSymbol, std::vector<State> >, State > transitions;
public:
	explicit DFTA();

	virtual AutomatonBase* clone() const;
	
	virtual AutomatonBase* plunder() &&;

	/**
	 * @copydoc Automaton::removeState(const State&)
	 */
	virtual bool removeState(const State& state);

	/**
	 * @copydoc Automaton::removeInputSymbol(const Symbol&)
	 */
	virtual bool removeInputSymbol(const alphabet::RankedSymbol& symbol);
	
	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current symbol in node
	 * @param children states of children of current node
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition(const alphabet::RankedSymbol & current, const std::vector<State> & children, const State & next);

	/**
	 * Removes transition from the automaton.
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition(const alphabet::RankedSymbol symbol, const std::vector<State> & states, const State & next);

	/**
	 * @return automaton transitions
	 */
	const std::map<std::pair<alphabet::RankedSymbol, std::vector<State> >, State > & getTransitions() const {return transitions;}

	std::set<State> getTransitionRightSide(const alphabet::RankedSymbol symbol, const std::vector<State> & states);

	unsigned transitionsSize() const;

	const std::set<alphabet::RankedSymbol> & getInputAlphabet() const {return inputAlphabet;}

	void setInputSymbols(const std::set<alphabet::RankedSymbol> & symbols) {inputAlphabet = symbols;}

	void addInputSymbol(const alphabet::RankedSymbol & symbol) {inputAlphabet.insert(symbol);}

	virtual int compare(const ObjectBase& other) const {
		return ObjectBase::compare(*this, other);
	}

	virtual int compare(const DFTA& other) const;

	virtual void operator>>(std::ostream& os) const;

	virtual operator std::string() const;

	virtual long long selfTypeId() const {
		return typeId(*this);
	}

	const static std::string XML_TAG_NAME;

	static DFTA parse(std::deque<sax::Token>::iterator& input);
	static void parseTransition(std::deque<sax::Token>::iterator& input, DFTA& automaton);

	void compose(std::deque<sax::Token>& out) const;
	void composeTransitions(std::deque<sax::Token>& out) const;
};

} /* namespace automaton */

namespace std {

template<>
struct compare<automaton::DFTA> {
	int operator()(const automaton::DFTA& first, const automaton::DFTA& second) const {
		return first.compare(second);
	}
};

} /* namespace std */

#endif /* DFTA_H_ */
