/*
 * NFTA.cpp
 *
 *  Created on: Mar 21, 2015
 *      Author: Stepan Plachy
 */

#include "NFTA.h"
#include "../AutomatonException.h"
#include <ostream>
#include <sstream>

#include "../../sax/FromXMLParserHelper.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace automaton {

NFTA::NFTA() {

}

NFTA::NFTA(const DFTA& other) {
	inputAlphabet = other.getInputAlphabet();
	states = other.getStates();
	finalStates = other.getFinalStates();
	for(const auto& transition : other.getTransitions()) {
		transitions[transition.first].insert(transition.second);
	}
}

AutomatonBase* NFTA::clone() const {
	return new NFTA(*this);
}

AutomatonBase* NFTA::plunder() && {
	return new NFTA(std::move(*this));
}

bool NFTA::removeState(const State& state) {
	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	AutomatonException ex ("State \"" + (std::string) state.getName() + "\" is used in transition.");
	for (const std::pair<std::pair<alphabet::RankedSymbol, std::vector<State> >, std::set<State> >& t : transitions) {
		const std::vector<State>& states = t . first.second;
		for (const State& it : states) {
			if (it == state) throw ex;
		}
		if (t . second.find(state) != t . second.end()) throw ex;
	}

	return states.erase(state);
}

bool NFTA::removeInputSymbol(const alphabet::RankedSymbol& symbol) {
	for (const std::pair<std::pair<alphabet::RankedSymbol, std::vector<State> >, std::set<State> >& t : transitions) {
		if (t . first.first == symbol)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used.");
	}

	return inputAlphabet.erase(symbol);
}

bool NFTA::addTransition(const alphabet::RankedSymbol & symbol, const std::vector<State> & prevStates, const State & next) {

	if (prevStates.size() != symbol.getRank().getData())
		throw AutomatonException("Number of states doesn't match rank of the symbol");

	if (inputAlphabet.find(symbol) == inputAlphabet.end())
		throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" doesn't exist.");

	if (states.find(next) == states.end())
		throw AutomatonException("State \"" + (std::string) next.getName() + "\" doesn't exist.");

	for (const State& it : prevStates) {
		if (states.find(it) == states.end())
			throw AutomatonException("State \"" + (std::string) it . getName() + "\" doesn't exist.");
	}

	std::pair<alphabet::RankedSymbol, std::vector<State> > key = std::make_pair(symbol, prevStates);
	return transitions[key].insert(next).second;
}

bool NFTA::removeTransition(const alphabet::RankedSymbol symbol, const std::vector<State> & states, const State & next) {
	std::pair<alphabet::RankedSymbol, std::vector<State> > key = std::make_pair(symbol, states);
	return transitions[key].erase(next);
}

bool NFTA::isDeterministic() const {
	for (const std::pair<std::pair<alphabet::RankedSymbol, std::vector<State> >, std::set<State> >& t : transitions) {
		if (t . second.size() != 1 || t . second.size() != 0) {
			return false;
		}
	}

	return true;
}

unsigned NFTA::transitionsSize() const {
	int res = 0;
	for(const auto& transition : transitions ){
		res += transition.second.size();
	}
	return res;
}

int NFTA::compare(const NFTA& other) const {
	auto first = std::tie(states, inputAlphabet, finalStates, transitions);
	auto second = std::tie(other.states, other.inputAlphabet, other.finalStates, other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void NFTA::operator>>(std::ostream& out) const {
	out << "(NFTA "
		<< " states = " << states
		<< " inputAlphabet = " << inputAlphabet
		<< " finalStates = " << finalStates
		<< " transitions = " << transitions
		<< ")";
}

NFTA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string NFTA::XML_TAG_NAME = "NFTA";

NFTA NFTA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, NFTA::XML_TAG_NAME);

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::RankedSymbol> inputSymbols = AutomatonFromXMLParser::parseRankedInputAlphabet(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	NFTA automaton;
	automaton.setStates(std::move(states));
	automaton.setInputSymbols(std::move(inputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<NFTA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, NFTA::XML_TAG_NAME);
	return automaton;
}

void NFTA::parseTransition(std::deque<sax::Token>::iterator& input, NFTA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	alphabet::RankedSymbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputRankedSymbol(input);
	std::vector<State> from = AutomatonFromXMLParser::parseTransitionFromMultiple(input);
	State to = AutomatonFromXMLParser::parseTransitionTo(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(inputSymbol), std::move(from), std::move(to));
}

void NFTA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(NFTA::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeRankedInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(NFTA::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void NFTA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::START_ELEMENT));
	for(const auto& transition : this->getTransitions()) {
		for(const auto& targetState: transition.second) {
			out.emplace_back(sax::Token("transition", sax::Token::TokenType::START_ELEMENT));

			AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, targetState);

			out.emplace_back(sax::Token("transition", sax::Token::TokenType::END_ELEMENT));
		}
	}

	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::END_ELEMENT));
}

} /* namespace automaton */

namespace alib {

xmlApi<automaton::Automaton>::ParserRegister<automaton::NFTA> NFTAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::NFTA>(automaton::NFTA::XML_TAG_NAME, automaton::NFTA::parse);
xmlApi<alib::Object>::ParserRegister<automaton::NFTA> NFTAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::NFTA>(automaton::NFTA::XML_TAG_NAME, automaton::NFTA::parse);

} /* namespace alib */
