/*
 * multipleDispatch.hpp
 *
 *  Created on: Apr 05, 2014
 *	  Author: Jan Travnicek
 */

#ifndef MULTIPLE_DISPATCH_H_
#define MULTIPLE_DISPATCH_H_

#include <stdexcept>
#include <functional>
#include <map>
#include <iostream>

#include "../CastApi.hpp"

namespace std {

template<class ReturnType, class FirstParameterType>
class SingleDispatch {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval(const FirstParameterType&) = 0;

	};

private:
	typedef SingleDispatch<ReturnType, FirstParameterType> SelfType;

	std::map<long long, typename SelfType::RegistratorWrapperBase*> registeredFunctions;

public:
	template<class RealReturnType, class RealFirstParameterType>
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function<RealReturnType(const RealFirstParameterType&)> callback;

	public:
		ReturnType eval(const FirstParameterType& first) {
			return ReturnType(callback((const RealFirstParameterType&) first));
		}

		RegistratorWrapper(SelfType& pool, RealReturnType(*callback)(const RealFirstParameterType&)) : callback(callback) {
			int a = 0;
			pool.registeredFunctions.insert(std::make_pair(RealFirstParameterType::typeId((const RealFirstParameterType&) a), this));
		}
	};

	ReturnType dispatch(const FirstParameterType& first) {
		typename std::map<long long, RegistratorWrapperBase*>::iterator callback = registeredFunctions.find(first.selfTypeId());
		if(callback == registeredFunctions.end()) throw std::bad_function_call();
		return callback->second->eval(first);
	}

};

template<class ReturnType, class StaticParamType, class FirstParameterType>
class SingleDispatchFirstStaticParam {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval(StaticParamType&, const FirstParameterType&) = 0;

	};

private:
	typedef SingleDispatchFirstStaticParam<ReturnType, StaticParamType, FirstParameterType> SelfType;

	std::map<long long, typename SelfType::RegistratorWrapperBase*> registeredFunctions;

public:
	template<class RealReturnType, class RealFirstParameterType>
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function<RealReturnType(StaticParamType&, const RealFirstParameterType&)> callback;

	public:
		ReturnType eval(StaticParamType& res, const FirstParameterType& first) {
			return ReturnType(callback(res, (const RealFirstParameterType&) first));
		}

		RegistratorWrapper(SelfType& pool, RealReturnType(*callback)(StaticParamType&, const RealFirstParameterType&)) : callback(callback) {
			int a = 0;
			pool.registeredFunctions.insert(std::make_pair(RealFirstParameterType::typeId((const RealFirstParameterType&) a), this));
		}
	};

	ReturnType dispatch(StaticParamType& res, const FirstParameterType& first) {
		typename std::map<long long, RegistratorWrapperBase*>::iterator callback = registeredFunctions.find(first.selfTypeId());
		if(callback == registeredFunctions.end()) throw std::bad_function_call();
		return callback->second->eval(res, first);
	}

};

template<class ReturnType, class FirstParameterType, class StaticParamType>
class SingleDispatchLastStaticParam {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval(const FirstParameterType&, StaticParamType&) = 0;

	};

private:
	typedef SingleDispatchLastStaticParam<ReturnType, FirstParameterType, StaticParamType> SelfType;

	std::map<long long, typename SelfType::RegistratorWrapperBase*> registeredFunctions;

public:
	template<class RealReturnType, class RealFirstParameterType>
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function<RealReturnType(const RealFirstParameterType&, StaticParamType&)> callback;

	public:
		ReturnType eval(const FirstParameterType& first, StaticParamType& res) {
			return ReturnType(callback((const RealFirstParameterType&) first, res));
		}

		RegistratorWrapper(SelfType& pool, RealReturnType(*callback)(const RealFirstParameterType&, StaticParamType&)) : callback(callback) {
			int a = 0;
			pool.registeredFunctions.insert(std::make_pair(RealFirstParameterType::typeId((const RealFirstParameterType&) a), this));
		}
	};

	ReturnType dispatch(const FirstParameterType& first, StaticParamType& res) {
		typename std::map<long long, RegistratorWrapperBase*>::iterator callback = registeredFunctions.find(first.selfTypeId());
		if(callback == registeredFunctions.end()) throw std::bad_function_call();
		return callback->second->eval(first, res);
	}

};

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template<class ReturnType, class ParametersType>
class PromotingDoubleDispatch {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval(bool firstAttempt, const ParametersType&, const ParametersType&) = 0;
		virtual bool available(bool firstAttempt, long long first, long long second) = 0;
	};

private:
	typedef PromotingDoubleDispatch<ReturnType, ParametersType> SelfType;

	std::map<long long, typename SelfType::RegistratorWrapperBase*> registeredFunctions;

public:
	template<class RealReturnType, class RealParametersType>
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function<RealReturnType(const RealParametersType&, const RealParametersType&)> callback;

	public:
		ReturnType eval(bool firstAttempt, const ParametersType& first, const ParametersType& second) {
			if(firstAttempt) {
				alib::Object casted = alib::castApi::getCastPool(first.selfTypeId()).cast(second);
				return ReturnType(callback((const RealParametersType&) first, (const RealParametersType&) casted.getData()));
			} else {
				alib::Object casted = alib::castApi::getCastPool(first.selfTypeId()).cast(first);
				return ReturnType(callback((const RealParametersType&) casted.getData(), (const RealParametersType&) second));
			}
		}

		bool available(bool firstAttempt, long long first, long long second) {
			if(firstAttempt)
				return alib::castApi::castAvailable(first, second);
			else
				return alib::castApi::castAvailable(second, first);
		}

		RegistratorWrapper(SelfType& pool, RealReturnType(*callback)(const RealParametersType&, const RealParametersType&)) : callback(callback) {
			int a = 0;
			pool.registeredFunctions.insert(std::make_pair(RealParametersType::typeId((const RealParametersType&) a), this));
		}
	};

	ReturnType dispatch(const ParametersType& first, const ParametersType& second) {
		typename std::map<long long, RegistratorWrapperBase*>::iterator callback = registeredFunctions.find(first.selfTypeId());
		if(callback != registeredFunctions.end() && callback->second->available(true, first.selfTypeId(), second.selfTypeId())) return callback->second->eval(true, first, second);

		callback = registeredFunctions.find(second.selfTypeId());
		if(callback != registeredFunctions.end() && callback->second->available(false, first.selfTypeId(), second.selfTypeId())) return callback->second->eval(false, first, second);

		throw std::bad_function_call();
	}

};

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template<class ReturnType, class FirstParameterType, class SecondParameterType>
class DoubleDispatch {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval(const FirstParameterType&, const SecondParameterType&) = 0;

	};

private:
	typedef DoubleDispatch<ReturnType, FirstParameterType, SecondParameterType> SelfType;

	std::map<std::pair<long long, long long>, typename SelfType::RegistratorWrapperBase*> registeredFunctions;

public:
	template<class RealReturnType, class RealFirstParameterType, class RealSecondParameterType>
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function<RealReturnType(const RealFirstParameterType&, const RealSecondParameterType&)> callback;

	public:
		ReturnType eval(const FirstParameterType& first, const SecondParameterType& second) {
			return ReturnType(callback((const RealFirstParameterType&) first, (const RealSecondParameterType&) second));
		}

		RegistratorWrapper(SelfType& pool, RealReturnType(*callback)(const RealFirstParameterType&, const RealSecondParameterType&)) : callback(callback) {
			int a = 0;
			pool.registeredFunctions.insert(std::make_pair(std::make_pair(RealFirstParameterType::typeId((const RealFirstParameterType&) a), RealSecondParameterType::typeId((const RealSecondParameterType&) a)), this));
		}
	};

	ReturnType dispatch(const FirstParameterType& first, const SecondParameterType& second) {
		typename std::map<std::pair<long long, long long>, RegistratorWrapperBase*>::iterator callback = registeredFunctions.find(std::make_pair(first.selfTypeId(), second.selfTypeId()));
		if(callback == registeredFunctions.end()) throw std::bad_function_call();
		return callback->second->eval(first, second);
	}

};

template<class ReturnType, class StaticParamType, class FirstParameterType, class SecondParameterType>
class DoubleDispatchFirstStaticParam {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval(StaticParamType&, const FirstParameterType&, const SecondParameterType&) = 0;

	};

private:
	typedef DoubleDispatchFirstStaticParam<ReturnType, StaticParamType, FirstParameterType, SecondParameterType> SelfType;

	std::map<std::pair<long long, long long>, typename SelfType::RegistratorWrapperBase*> registeredFunctions;

public:
	template<class RealReturnType, class RealFirstParameterType, class RealSecondParameterType>
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function<RealReturnType(StaticParamType&, const RealFirstParameterType&, const RealSecondParameterType&)> callback;

	public:
		ReturnType eval(StaticParamType& res, const FirstParameterType& first, const SecondParameterType& second) {
			return ReturnType(callback(res, (const RealFirstParameterType&) first, (const SecondParameterType&) second));
		}

		RegistratorWrapper(SelfType& pool, RealReturnType(*callback)(StaticParamType&, const RealFirstParameterType&, const RealSecondParameterType&)) : callback(callback) {
			int a = 0;
			pool.registeredFunctions.insert(std::make_pair(std::make_pair(RealFirstParameterType::typeId((const RealFirstParameterType&) a), RealSecondParameterType::typeId((const RealSecondParameterType&) a)), this));
		}
	};

	ReturnType dispatch(StaticParamType& res, const FirstParameterType& first, const SecondParameterType& second) {
		typename std::map<std::pair<long long, long long>, RegistratorWrapperBase*>::iterator callback = registeredFunctions.find(std::make_pair(first.selfTypeId(), first.selfTypeId()));
		if(callback == registeredFunctions.end()) throw std::bad_function_call();
		return callback->second->eval(res, first, second);
	}

};

} /* namespace std */

#endif /* MULTIPLE_DISPATCH_H_ */
