/*
 * GraphFeatures.h
 *
 *  Created on: Jun 19, 2014
 *      Author: David Rosca
 */

#include "DirectedGraph.h"
#include "IDirectedGraph.h"
#include "AdjacencyListDirectedGraph.h"
#include "AdjacencyMatrixDirectedGraph.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../../sax/FromXMLParserHelper.h"
#include "../common/GraphFromXMLParser.h"
#include "../common/GraphToXMLComposer.h"
#include "../Graph.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace graph {

DirectedGraph::DirectedGraph(REPRESENTATION representation)
	: representation(representation)
{
	init();
}

DirectedGraph::~DirectedGraph() noexcept
{
	delete impl;
}

DirectedGraph::DirectedGraph(const DirectedGraph &other)
	: representation(other.representation)
	, nodeValues(other.nodeValues)
	, edgeValues(other.edgeValues)
{
	init();
	impl->copy(other.impl);
}

DirectedGraph::DirectedGraph(DirectedGraph &&other) noexcept
	: representation(other.representation)
	, impl(other.impl)
	, nodeValues(std::move(other.nodeValues))
	, edgeValues(std::move(other.edgeValues))
{
	other.impl = 0;
}

DirectedGraph &DirectedGraph::operator=(const DirectedGraph &other)
{
	if (this == &other) {
		return *this;
	}

	*this = DirectedGraph(other);
	return *this;
}

DirectedGraph &DirectedGraph::operator=(DirectedGraph &&other) noexcept
{
	this->representation = other.representation;
	std::swap(this->impl, other.impl);
	std::swap(this->nodeValues, other.nodeValues);
	std::swap(this->edgeValues, other.edgeValues);
	return *this;
}

GraphBase *DirectedGraph::clone() const
{
	return new DirectedGraph(*this);
}

GraphBase *DirectedGraph::plunder() &&
{
	return new DirectedGraph(std::move(*this));
}

REPRESENTATION DirectedGraph::getRepresentation() const
{
	return representation;
}

std::set<Node> DirectedGraph::getNodes() const
{
	return impl->allNodes();
}

bool DirectedGraph::hasNode(const Node &node) const
{
	return impl->hasNode(node);
}

bool DirectedGraph::hasEdge(const DirectedEdge &edge) const
{
	return impl->hasEdge(edge);
}

std::set<DirectedEdge> DirectedGraph::getEdges() const
{
	return impl->allEdges();
}

bool DirectedGraph::hasEdge(const Node &from, const Node &to) const
{
	return impl->hasEdge(from, to);
}

std::set<DirectedEdge> DirectedGraph::findEdges(const Node &from, const Node &to) const
{
	return impl->findEdges(from, to);
}

std::set<Node> DirectedGraph::neighbors(const Node &node) const
{
	return impl->neighbors(node);
}

std::set<DirectedEdge> DirectedGraph::neighborEdges(const Node &node) const
{
	return impl->neighborEdges(node);
}

bool DirectedGraph::addNode(const Node &node)
{
	return impl->addNode(node);
}

bool DirectedGraph::addNode(const Node &node, int value)
{
	bool success = addNode(node);
	if (success) {
		setNodeValue(node, value);
	}
	return success;
}

bool DirectedGraph::removeNode(const Node &node)
{
	bool success = impl->removeNode(node);
	if (success) {
		nodeValues.erase(node);
	}
	return success;
}

bool DirectedGraph::addEdge(const DirectedEdge &edge)
{
	return impl->addEdge(edge);
}

bool DirectedGraph::addEdge(const DirectedEdge &edge, int value)
{
	bool success = addEdge(edge);
	if (success) {
		setEdgeValue(edge, value);
	}
	return success;
}

bool DirectedGraph::removeEdge(const DirectedEdge &edge)
{
	bool success = impl->removeEdge(edge);
	if (success) {
		edgeValues.erase(edge);
	}
	return success;
}

int DirectedGraph::getNodeValue(const Node &node) const
{
	return nodeValues.at(node);
}

void DirectedGraph::setNodeValue(const Node &node, int value)
{
	nodeValues.insert({node, value});
}

int DirectedGraph::getEdgeValue(const DirectedEdge &edge) const
{
	return edgeValues.at(edge);
}

void DirectedGraph::setEdgeValue(const DirectedEdge &edge, int value)
{
	edgeValues.insert({edge, value});
}

void DirectedGraph::operator>>(std::ostream &out) const
{
	out << "(DirectedGraph " << static_cast<std::string>(*impl) << ")";
}

int DirectedGraph::compare(const DirectedGraph &other) const
{
	if (representation != other.representation) {
		return static_cast<int>(representation) - static_cast<int>(other.representation);
	}

	auto first = std::tie(nodeValues, edgeValues);
	auto second = std::tie(other.nodeValues, other.edgeValues);

	std::compare<decltype(first)> comp;
	int res = comp(first, second);

	if (res != 0) {
		return res;
	}

	return impl->compare(other.impl);
}

DirectedGraph::operator std::string() const
{
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

void DirectedGraph::init()
{
	switch (representation) {
	case REPRESENTATION::ADJACENCY_LIST:
		impl = new AdjacencyListDirectedGraph();
		break;

	case REPRESENTATION::ADJACENCY_MATRIX:
		impl = new AdjacencyMatrixDirectedGraph();
		break;

	case REPRESENTATION::INCIDENCE_MATRIX:

	default:
		impl = 0;
	}
}

const std::string DirectedGraph::XML_TAG_NAME = "DirectedGraph";

DirectedGraph DirectedGraph::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, DirectedGraph::XML_TAG_NAME);

	DirectedGraph graph(GraphFromXMLParser::parseRepresentation(input));
	GraphFromXMLParser::parseNodes(input, graph);
	GraphFromXMLParser::parseDirectedEdges(input, graph);
	GraphFromXMLParser::parseNodeValues(input, graph);
	GraphFromXMLParser::parseDirectedEdgeValues(input, graph);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, DirectedGraph::XML_TAG_NAME);
	return graph;
}

void DirectedGraph::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(DirectedGraph::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	GraphToXMLComposer::composeRepresentation(out, representationToString(this->getRepresentation()));
	GraphToXMLComposer::composeNodes(out, this->getNodes());
	GraphToXMLComposer::composeEdges(out, this->getEdges());
	GraphToXMLComposer::composeNodeValues(out, *this);
	GraphToXMLComposer::composeEdgeValues(out, *this);

	out.emplace_back(DirectedGraph::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

} /* namespace graph */

namespace alib {

xmlApi<graph::Graph>::ParserRegister<graph::DirectedGraph> directedGraphParserRegister = xmlApi<graph::Graph>::ParserRegister<graph::DirectedGraph>(graph::DirectedGraph::XML_TAG_NAME, graph::DirectedGraph::parse);
xmlApi<alib::Object>::ParserRegister<graph::DirectedGraph> directedGraphParserRegister2 = xmlApi<alib::Object>::ParserRegister<graph::DirectedGraph>(graph::DirectedGraph::XML_TAG_NAME, graph::DirectedGraph::parse);

} /* namespace alib */
