/*
 * GraphFeatures.h
 *
 *  Created on: Jun 19, 2014
 *      Author: David Rosca
 */

#include "AdjacencyMatrixUndirectedGraph.h"
#include <compare>
#include "utils.h"

namespace graph {

std::set<Node> AdjacencyMatrixUndirectedGraph::allNodes() const
{
	std::set<Node> out;

	for (auto i : adj) {
		out.insert(i.first);
	}

	return out;
}

std::set<UndirectedEdge> AdjacencyMatrixUndirectedGraph::allEdges() const
{
	return edges;
}

bool AdjacencyMatrixUndirectedGraph::hasEdge(const Node &first, const Node &second) const
{
	auto search = adj.find(first);
	if (search == adj.end()) {
		return false;
	}
	return search->second.find(second) != search->second.end();
}

bool AdjacencyMatrixUndirectedGraph::hasNode(const Node &node) const
{
	return adj.find(node) != adj.end();
}

bool AdjacencyMatrixUndirectedGraph::hasEdge(const UndirectedEdge &edge) const
{
	return edges.find(edge) != edges.end();
}

std::set<UndirectedEdge> AdjacencyMatrixUndirectedGraph::findEdges(const Node &first, const Node &second) const
{
	std::set<UndirectedEdge> out;

	for (const UndirectedEdge &e : edges) {
		if (edgeMatch(e, first, second)) {
			out.insert(e);
		}
	}

	return out;
}

std::set<Node> AdjacencyMatrixUndirectedGraph::neighbors(const Node &node) const
{
	std::set<Node> out;

	auto search = adj.find(node);
	if (search == adj.end()) {
		return out;
	}

	for (auto i : search->second) {
		if (i.second > 0) {
			out.insert(i.first);
		}
	}

	return out;
}

std::set<UndirectedEdge> AdjacencyMatrixUndirectedGraph::neighborEdges(const Node &node) const
{
	std::set<UndirectedEdge> out;
	const std::set<Node> &nghbrs = neighbors(node);

	for (const UndirectedEdge &e : edges) {
		if (edgeMatch(e, node, nghbrs)) {
			out.insert(e);
		}
	}

	return out;
}

bool AdjacencyMatrixUndirectedGraph::addNode(const Node &node)
{
	if (adj.find(node) != adj.end()) {
		return false;
	}

	adj[node]; // insert key
	return true;
}

bool AdjacencyMatrixUndirectedGraph::removeNode(const Node &node)
{
	auto search = adj.find(node);
	if (search == adj.end()) {
		return false;
	}

	adj.erase(search);

	for (auto& i : edges) {
		if (i.getFirstNode() == node) {
			adj[i.getSecondNode()].erase(node);
			edges.erase(i);
		} else if (i.getSecondNode() == node) {
			adj[i.getFirstNode()].erase(node);
			edges.erase(i);
		}
	}

	return true;
}

bool AdjacencyMatrixUndirectedGraph::addEdge(const UndirectedEdge &edge)
{
	if (edges.find(edge) != edges.end()) {
		return false;
	}

	edges.insert(edge);

	adj[edge.getFirstNode()][edge.getSecondNode()] += 1;
	adj[edge.getSecondNode()][edge.getFirstNode()] += 1;

	return true;
}

bool AdjacencyMatrixUndirectedGraph::removeEdge(const UndirectedEdge &edge)
{
	auto search = edges.find(edge);
	if (search == edges.end()) {
		return false;
	}

	edges.erase(search);

	auto &map1 = adj[edge.getFirstNode()];
	int &val1 = map1[edge.getSecondNode()];
	if (--val1 <= 0) {
		map1.erase(edge.getSecondNode());
		if (map1.empty()) {
			adj.erase(edge.getFirstNode());
		}
	}

	auto &map2 = adj[edge.getSecondNode()];
	int &val2 = map2[edge.getFirstNode()];
	if (--val2 <= 0) {
		map2.erase(edge.getFirstNode());
		if (map2.empty()) {
			adj.erase(edge.getSecondNode());
		}
	}

	return true;
}

void AdjacencyMatrixUndirectedGraph::copy(IUndirectedGraph *other)
{
	AdjacencyMatrixUndirectedGraph *o = static_cast<AdjacencyMatrixUndirectedGraph*>(other);

	edges = o->edges;
	adj = o->adj;
}

int AdjacencyMatrixUndirectedGraph::compare(IUndirectedGraph *other) const
{
	AdjacencyMatrixUndirectedGraph *o = static_cast<AdjacencyMatrixUndirectedGraph*>(other);

	auto first = std::tie(edges, adj);
	auto second = std::tie(o->edges, o->adj);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

} // namespace graph
