/*
 * FormalRegExpAlternation.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "FormalRegExpAlternation.h"
#include "../../exception/AlibException.h"
#include "../unbounded/UnboundedRegExpAlternation.h"
#include <sstream>

namespace regexp {

FormalRegExpAlternation::FormalRegExpAlternation(FormalRegExpElement&& left, FormalRegExpElement&& right) {
	this->left = std::move(left).plunder();
	this->right = std::move(right).plunder();
}

FormalRegExpAlternation::FormalRegExpAlternation(const FormalRegExpElement& left, const FormalRegExpElement& right) {
	this->left = left.clone();
	this->right = right.clone();
}

FormalRegExpAlternation::FormalRegExpAlternation(const FormalRegExpAlternation& other) {
	this->left = other.left->clone();
	this->right = other.right->clone();
}

FormalRegExpAlternation::FormalRegExpAlternation(FormalRegExpAlternation&& other) noexcept : left(std::move(other.left)), right(std::move(other.right)) {
	other.left = NULL;
	other.right = NULL;
	this->attachRegExp(NULL);
}

FormalRegExpAlternation& FormalRegExpAlternation::operator=(const FormalRegExpAlternation& other) {
	if (this == &other) {
		return *this;
	}

	*this = FormalRegExpAlternation(other);

	return *this;
}

FormalRegExpAlternation& FormalRegExpAlternation::operator=(FormalRegExpAlternation&& other) noexcept {
	std::swap(this->left, other.left);
	std::swap(this->right, other.right);

	this->left->attachRegExp(this->parentRegExp);
	this->right->attachRegExp(this->parentRegExp);

	return *this;
}

FormalRegExpAlternation::~FormalRegExpAlternation() noexcept {
	delete left;
	delete right;
}

const FormalRegExpElement & FormalRegExpAlternation::getLeftElement() const {
	return *left;
}

const FormalRegExpElement & FormalRegExpAlternation::getRightElement() const {
	return *right;
}

FormalRegExpElement & FormalRegExpAlternation::getLeftElement() {
	return *left;
}

FormalRegExpElement & FormalRegExpAlternation::getRightElement() {
	return *right;
}

void FormalRegExpAlternation::setLeftElement(const FormalRegExpElement& element) {
	FormalRegExpElement* elem = element.clone();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet.");
	delete left;
	this->left = elem;
}

void FormalRegExpAlternation::setLeftElement(FormalRegExpElement&& element) {
	FormalRegExpElement* elem = std::move(element).plunder();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet."); 
	delete left;
	this->left = elem;
}

void FormalRegExpAlternation::setRightElement(const FormalRegExpElement& element) {
	FormalRegExpElement* elem = element.clone();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet.");
	delete right;
	this->right = elem;
}

void FormalRegExpAlternation::setRightElement(FormalRegExpElement&& element) {
	FormalRegExpElement* elem = std::move(element).plunder();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet."); 
	delete right;
	this->right = elem;
}

FormalRegExpElement* FormalRegExpAlternation::clone() const {
	return new FormalRegExpAlternation(*this);
}

FormalRegExpElement* FormalRegExpAlternation::plunder() && {
	return new FormalRegExpAlternation(std::move(*this));
}

UnboundedRegExpElement* FormalRegExpAlternation::cloneAsUnbounded() const {
	UnboundedRegExpAlternation* res = new UnboundedRegExpAlternation();
	
	UnboundedRegExpElement* leftTmp = left->cloneAsUnbounded();
	res->appendElement(std::move(*leftTmp));
	delete leftTmp;
	
	UnboundedRegExpElement* rightTmp = right->cloneAsUnbounded();
	res->appendElement(std::move(*rightTmp));
	delete rightTmp;

	return res;
}

int FormalRegExpAlternation::compare(const FormalRegExpAlternation& other) const {
	int res = left->compare(*other.left);
	if(res == 0) res = right->compare(*other.right);
	return res;
}

void FormalRegExpAlternation::operator>>(std::ostream& out) const {
	out << "(FormalRegExpAlternation";
	out << " " << *left;
	out << " " << *right;
	out << ")";
}

bool FormalRegExpAlternation::testSymbol( const alphabet::Symbol & symbol ) const {
	if(left->testSymbol(symbol)) return true;
	if(right->testSymbol(symbol)) return true;
	return false;
}

bool FormalRegExpAlternation::attachRegExp(const RegExpAlphabet * regexp ) {
	if(this->parentRegExp == regexp) return true;
	
	this->parentRegExp = regexp;
	if(!left->attachRegExp(regexp)) return false;
	if(!right->attachRegExp(regexp)) return false;
	return true;
}

void FormalRegExpAlternation::computeMinimalAlphabet( std::set<alphabet::Symbol>& alphabet ) const {
	left->computeMinimalAlphabet(alphabet);
	right->computeMinimalAlphabet(alphabet);
}

FormalRegExpAlternation::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

} /* namespace regexp */
