/*
 * UnboundedRegExp.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNBOUNDED_REG_EXP_H_
#define UNBOUNDED_REG_EXP_H_

#include <vector>
#include <list>
#include <string>
#include <set>
#include "UnboundedRegExpElement.h"
#include "../RegExpBase.h"
#include "../common/RegExpAlphabet.h"

namespace regexp {

class UnboundedRegExpElement;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
class UnboundedRegExp : public std::acceptor<UnboundedRegExp, VisitableRegExpBase, std::acceptor<UnboundedRegExp, alib::VisitableObjectBase, RegExpBase> >, public RegExpAlphabet {
protected:
	UnboundedRegExpElement* regExp;

public:
	/**
	 * @copydoc UnboundedRegExpElement::clone() const
	 */
	virtual RegExpBase* clone() const;

	/**
	 * @copydoc UnboundedRegExpElement::plunder() const
	 */
	virtual RegExpBase* plunder() &&;

	explicit UnboundedRegExp();
	explicit UnboundedRegExp(const FormalRegExp& other);
	explicit UnboundedRegExp(std::set<alphabet::Symbol> alphabet, const UnboundedRegExpElement& regExp);
	explicit UnboundedRegExp(std::set<alphabet::Symbol> alphabet, UnboundedRegExpElement&& regExp);
	explicit UnboundedRegExp(const UnboundedRegExpElement& regExp);
	explicit UnboundedRegExp(UnboundedRegExpElement&& regExp);

	/**
	 * Copy constructor.
	 * @param other RegExp to copy
	 */
	UnboundedRegExp(const UnboundedRegExp& other);
	UnboundedRegExp(UnboundedRegExp&& other) noexcept;
	UnboundedRegExp& operator =(const UnboundedRegExp& other);
	UnboundedRegExp& operator =(UnboundedRegExp&& other) noexcept;
	~UnboundedRegExp() noexcept;

	/**
	 * @return Root node of the regular expression tree
	 */
	const UnboundedRegExpElement& getRegExp() const;

	/**
	 * @return Root node of the regular expression tree
	 */
	UnboundedRegExpElement& getRegExp();

	/**
	 * Sets the root node of the regular expression tree. Doesn't perform copy of the regExp param,
	 * just stores it!
	 * @param regExp root node to set
	 */
	void setRegExp(const UnboundedRegExpElement& regExp);

	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRegExp(UnboundedRegExpElement&& regExp);

	virtual const std::set<alphabet::Symbol>& getAlphabet() const {
		return RegExpAlphabet::getAlphabet();
	}

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet(const alphabet::Symbol & symbol);

	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 * @param regexp RegExp to print
	 */
	virtual void operator>>(std::ostream& out) const;

	virtual int compare(const ObjectBase& other) const {
		return ObjectBase::compare(*this, other);
	}

	virtual int compare(const UnboundedRegExp& other) const;

	virtual explicit operator std::string() const;

	virtual long long selfTypeId() const {
		return typeId(*this);
	}

	const static std::string XML_TAG_NAME;

	static UnboundedRegExp parse(std::deque<sax::Token>::iterator& input);

	void compose(std::deque<sax::Token>& out) const;
};

} /* namespace regexp */

namespace std {

template<>
struct compare<regexp::UnboundedRegExp> {
	int operator()(const regexp::UnboundedRegExp& first, const regexp::UnboundedRegExp& second) const {
		return first.compare(second);
	}
};

} /* namespace std */

#endif /* UNBOUNDED_REG_EXP_H_ */
