/*
 * PrefixRankedTree.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "PrefixRankedTree.h"
#include "../../exception/AlibException.h"

#include <sstream>
#include <algorithm>
#include <deque>

#include "RankedTree.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../Tree.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"
#include "../../CastApi.hpp"

namespace tree {

PrefixRankedTree::PrefixRankedTree(std::set<alphabet::RankedSymbol> alphabet, std::vector<alphabet::RankedSymbol> data) {
	this->alphabet = std::move(alphabet);
	setContent(std::move(data));
}

PrefixRankedTree::PrefixRankedTree(std::vector<alphabet::RankedSymbol> data) {
	arityChecksum(data);

	alphabet = std::set<alphabet::RankedSymbol>(data.begin(), data.end());
	m_Data = std::move(data);
}

PrefixRankedTree::PrefixRankedTree(const RankedTree& tree) {
	alphabet = tree.getAlphabet();
	std::deque<const RankedNode*> queue {&tree.getRoot()};
	while(!queue.empty()) {
		const RankedNode* elem = queue.back();
		queue.pop_back();

		m_Data.push_back(elem->getSymbol());
		for(const RankedNode* child : elem->getChildren()) {
			queue.push_back(child);
		}
	}
}

TreeBase* PrefixRankedTree::clone() const {
	return new PrefixRankedTree(*this);
}

TreeBase* PrefixRankedTree::plunder() && {
	return new PrefixRankedTree(std::move(*this));
}

bool PrefixRankedTree::removeSymbolFromAlphabet(const alphabet::RankedSymbol & symbol) {
	if(std::any_of(m_Data.begin(), m_Data.end(), [&](const alphabet::RankedSymbol & s) { return s == symbol; } ) )
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is used.");

	return alphabet.erase(symbol);
}

const std::vector<alphabet::RankedSymbol>& PrefixRankedTree::getContent() const {
	return this->m_Data;
}

void PrefixRankedTree::setContent(std::vector<alphabet::RankedSymbol> data) {
	arityChecksum(data);

	std::set<alphabet::RankedSymbol> minimalAlphabet(data.begin(), data.end());
	std::set<alphabet::RankedSymbol> unknownSymbols;
	std::set_difference(minimalAlphabet.begin(), minimalAlphabet.end(), alphabet.begin(), alphabet.end(), std::inserter(unknownSymbols, unknownSymbols.end()));

	if(unknownSymbols.size() > 0)
		throw exception::AlibException("Input symbols not in the alphabet.");

	this->m_Data = std::move(data);
}

void PrefixRankedTree::arityChecksum(const std::vector<alphabet::RankedSymbol>& data) {
	int arityChecksum = 1;
	for(const alphabet::RankedSymbol& symbol : data) {
		arityChecksum += symbol.getRank().getData();
		arityChecksum -= 1;
	}

	if(arityChecksum != 0) throw exception::AlibException("The string does not form a tree");
}

bool PrefixRankedTree::isEmpty() const {
	return this->m_Data.size() == 0;
}

int PrefixRankedTree::compare(const PrefixRankedTree& other) const {
	auto first = std::tie(m_Data, alphabet);
	auto second = std::tie(other.m_Data, other.alphabet);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void PrefixRankedTree::operator >>(std::ostream& out) const {
	if( this->isEmpty() ) {
		out << "(Epsilon)";
	} else {
		out << "(PrefixRankedTree ";
		for(const alphabet::RankedSymbol& symbol : this->m_Data)
			out << symbol;
		out << ")";
	}
}

PrefixRankedTree::operator std::string () const {
	std::stringstream ss;
	if( this->isEmpty() ) {
		ss << "E";
	} else {
		ss << "\"";
		for(const alphabet::RankedSymbol& symbol : this->m_Data)
			ss << (std::string) symbol;
		ss << "\"";
	}
	return std::move(ss).str();
}

const std::string PrefixRankedTree::XML_TAG_NAME = "PrefixRankedTree";

PrefixRankedTree PrefixRankedTree::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, PrefixRankedTree::XML_TAG_NAME);
	std::set<alphabet::RankedSymbol> rankedAlphabet = TreeFromXMLParser::parseRankedAlphabet(input);
	std::vector<alphabet::RankedSymbol> data = TreeFromXMLParser::parseContentData(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, PrefixRankedTree::XML_TAG_NAME);
	return PrefixRankedTree(std::move(rankedAlphabet), std::move(data));
}

void PrefixRankedTree::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(PrefixRankedTree::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);
	TreeToXMLComposer::composeAlphabet(out, alphabet);
	TreeToXMLComposer::composeContent(out, m_Data);
	out.emplace_back(PrefixRankedTree::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

} /* namespace tree */

namespace alib {

xmlApi<tree::Tree>::ParserRegister<tree::PrefixRankedTree> prefixRankedTreeParserRegister = xmlApi<tree::Tree>::ParserRegister<tree::PrefixRankedTree>(tree::PrefixRankedTree::XML_TAG_NAME, tree::PrefixRankedTree::parse);
xmlApi<alib::Object>::ParserRegister<tree::PrefixRankedTree> prefixRankedTreeParserRegister2 = xmlApi<alib::Object>::ParserRegister<tree::PrefixRankedTree>(tree::PrefixRankedTree::XML_TAG_NAME, tree::PrefixRankedTree::parse);

auto PrefixRankedTreeFromRankedTree = castApi::CastRegister<tree::PrefixRankedTree, tree::RankedTree>();
auto PrefixRankedTreeCastBinder = castApi::CastPoolStringBinder<tree::PrefixRankedTree>(tree::PrefixRankedTree::XML_TAG_NAME);

} /* namespace alib */
