/*
 * RankedPattern.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "RankedPattern.h"
#include "../../exception/AlibException.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../unranked/UnrankedPattern.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../Tree.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace tree {

RankedPattern::RankedPattern(const UnrankedPattern& other) : RankedPatternAlphabet(alphabet::RankedSymbol(other.getSubtreeWildcard(), 0)) {
	this->pattern = NULL;
	RankedNode* element = other.getRoot().cloneAsRanked();
	element->computeMinimalAlphabet(alphabet);
	setTree(std::move(*element));
	delete element;
}

RankedPattern::RankedPattern(alphabet::RankedSymbol subtreeWildcard, std::set<alphabet::RankedSymbol> alphabet, RankedNode pattern) : RankedPatternAlphabet(std::move(subtreeWildcard)) {
	setAlphabet(std::move(alphabet));
	this->pattern = NULL;
	setTree(std::move(pattern));
}

RankedPattern::RankedPattern(alphabet::RankedSymbol subtreeWildcard, RankedNode pattern) : RankedPatternAlphabet(std::move(subtreeWildcard)) {
	pattern.computeMinimalAlphabet(alphabet);
	this->pattern = NULL;
	setTree(std::move(pattern));
}

RankedPattern::RankedPattern(const RankedPattern& other) : RankedPatternAlphabet(other), pattern(other.pattern->clone()) {
	this->pattern->attachTree(this);
}

RankedPattern::RankedPattern(RankedPattern&& other) noexcept : RankedPatternAlphabet(other), pattern(other.pattern) {
	this->pattern->attachTree(this);
	other.pattern = NULL;
}

TreeBase* RankedPattern::clone() const {
	return new RankedPattern(*this);
}

TreeBase* RankedPattern::plunder() && {
	return new RankedPattern(std::move(*this));
}

RankedPattern& RankedPattern::operator=(const RankedPattern& other) {
	if (this == &other) {
		return *this;
	}

	*this = RankedPattern(other);

	return *this;
}

RankedPattern& RankedPattern::operator=(RankedPattern&& other) noexcept {
	std::swap(this->pattern, other.pattern);
	std::swap(this->alphabet, other.alphabet);
	std::swap(this->subtreeWildcard, other.subtreeWildcard);
	return *this;
}

RankedPattern::~RankedPattern() noexcept {
	delete pattern;
}

const RankedNode& RankedPattern::getRoot() const {
	return *pattern;
}

RankedNode& RankedPattern::getRoot() {
	return *pattern;
}

void RankedPattern::setTree(RankedNode pattern) {
	delete this->pattern;
	this->pattern = std::move(pattern).plunder();
	if(!this->pattern->attachTree(this)) {
		delete this->pattern;
		throw exception::AlibException("Input symbols not in the alphabet.");
	}
}

bool RankedPattern::removeSymbolFromAlphabet(const alphabet::RankedSymbol & symbol) {
	if(this->pattern->testSymbol(symbol))
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is used.");

	if(this->subtreeWildcard == symbol)
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is subtreeWildcard.");

	return alphabet.erase(symbol);
}

void RankedPattern::operator >>(std::ostream& out) const {
	out << "(RankedPattern " << *(this->pattern) << ")";
}

int RankedPattern::compare(const RankedPattern& other) const {
	int res = pattern->compare(*other.pattern);
	if(res == 0) {
		std::compare<alphabet::RankedSymbol> comp;
		res = comp(subtreeWildcard, other.subtreeWildcard);
	}
	if(res == 0) {
		std::compare<std::set<alphabet::RankedSymbol>> comp;
		res = comp(alphabet, other.alphabet);
	}
	return res;
}

void RankedPattern::nicePrint(std::ostream & os) const {
	pattern -> nicePrint(os);
}

RankedPattern::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string RankedPattern::XML_TAG_NAME = "RankedPattern";

RankedPattern RankedPattern::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, RankedPattern::XML_TAG_NAME);
	alphabet::RankedSymbol subtreeWildcardSymbol = TreeFromXMLParser::parseSubtreeWildcardRankedSymbol(input);
	std::set<alphabet::RankedSymbol> rankedAlphabet = TreeFromXMLParser::parseRankedAlphabet(input);
	RankedNode * root = TreeFromXMLParser::parseRankedNode(input);
	RankedPattern tree(std::move(subtreeWildcardSymbol), std::move(rankedAlphabet), std::move(*root));
	delete root;
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, RankedPattern::XML_TAG_NAME);
	return tree;
}

void RankedPattern::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(RankedPattern::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);
	TreeToXMLComposer::composeSubtreeWildcard(out, subtreeWildcard);
	TreeToXMLComposer::composeAlphabet(out, alphabet);
	TreeToXMLComposer::composeNode(out, *pattern);
	out.emplace_back(RankedPattern::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

} /* namespace tree */

namespace alib {

xmlApi<tree::Tree>::ParserRegister<tree::RankedPattern> rankedPatternParserRegister = xmlApi<tree::Tree>::ParserRegister<tree::RankedPattern>(tree::RankedPattern::XML_TAG_NAME, tree::RankedPattern::parse);
xmlApi<alib::Object>::ParserRegister<tree::RankedPattern> rankedPatternParserRegister2 = xmlApi<alib::Object>::ParserRegister<tree::RankedPattern>(tree::RankedPattern::XML_TAG_NAME, tree::RankedPattern::parse);

} /* namespace alib */
