/*
 * UnrankedPattern.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNRANKED_PATTERN_H_
#define UNRANKED_PATTERN_H_

#include <vector>
#include <list>
#include <string>
#include <set>
#include "UnrankedNode.h"
#include "../TreeBase.h"
#include "../common/UnrankedPatternAlphabet.h"

namespace tree {

class UnrankedNode;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a pattern of RegExpElement.
 */
class UnrankedPattern : public std::acceptor<UnrankedPattern, VisitableTreeBase, std::acceptor<UnrankedPattern, alib::VisitableObjectBase, TreeBase> >, public UnrankedPatternAlphabet {
protected:
	UnrankedNode* pattern;

public:
	/**
	 * @copydoc UnrankedNode::clone() const
	 */
	virtual TreeBase* clone() const;

	/**
	 * @copydoc UnrankedNode::plunder() const
	 */
	virtual TreeBase* plunder() &&;

	explicit UnrankedPattern(const RankedPattern& other);
	explicit UnrankedPattern(alphabet::Symbol subtreeWildcard, std::set<alphabet::Symbol> alphabet, UnrankedNode pattern);
	explicit UnrankedPattern(alphabet::Symbol subtreeWildcard, UnrankedNode pattern);

	/**
	 * Copy constructor.
	 * @param other pattern to copy
	 */
	UnrankedPattern(const UnrankedPattern& other);
	UnrankedPattern(UnrankedPattern&& other) noexcept;
	UnrankedPattern& operator =(const UnrankedPattern& other);
	UnrankedPattern& operator =(UnrankedPattern&& other) noexcept;
	~UnrankedPattern() noexcept;

	/**
	 * @return Root node of the regular expression pattern
	 */
	const UnrankedNode& getRoot() const;

	/**
	 * @return Root node of the regular expression pattern
	 */
	UnrankedNode& getRoot();

	/**
	 * Sets the root node of the regular expression pattern
	 * @param pattern root node to set
	 */
	void setTree(UnrankedNode pattern);

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet(const alphabet::Symbol & symbol);

	/**
	 * Prints XML representation of the pattern to the output stream.
	 * @param out output stream to which print the pattern
	 * @param pattern pattern to print
	 */
	virtual void operator>>(std::ostream& out) const;

	virtual int compare(const ObjectBase& other) const {
		return ObjectBase::compare(*this, other);
	}

	virtual int compare(const UnrankedPattern& other) const;

	void nicePrint(std::ostream & os) const;

	virtual explicit operator std::string() const;

	virtual long long selfTypeId() const {
		return typeId(*this);
	}

	const static std::string XML_TAG_NAME;

	static UnrankedPattern parse(std::deque<sax::Token>::iterator& input);

	void compose(std::deque<sax::Token>& out) const;
};

} /* namespace tree */

namespace std {

template<>
struct compare<tree::UnrankedPattern> {
	int operator()(const tree::UnrankedPattern& first, const tree::UnrankedPattern& second) const {
		return first.compare(second);
	}
};

} /* namespace std */

#endif /* UNRANKED_PATTERN_H_ */
