/*
 * EpsilonRemoverIncoming.cpp
 *
 *  Created on: 24. 11. 2014
 *	  Author: Jan Travnicek
 */

#include "EpsilonRemoverIncoming.h"

#include "../../properties/efficient/AllEpsilonClosure.h"

namespace automaton {

namespace simplify {

namespace efficient {

automaton::DFA EpsilonRemoverIncoming::remove(const automaton::DFA& origFSM)
{
	return origFSM;
}

automaton::MultiInitialStateNFA EpsilonRemoverIncoming::remove(const automaton::MultiInitialStateNFA& origFSM)
{
	return origFSM;
}

automaton::NFA EpsilonRemoverIncoming::remove(const automaton::NFA& origFSM)
{
	return origFSM;
}

automaton::NFA EpsilonRemoverIncoming::remove( const automaton::EpsilonNFA & origFSM ) {
	automaton::NFA fsm(origFSM.getInitialState());

	fsm.setStates( origFSM.getStates() );
	fsm.setInputAlphabet( origFSM.getInputAlphabet() );

	std::map<automaton::State, std::set<automaton::State>> closures = automaton::properties::efficient::AllEpsilonClosure::allEpsilonClosure( origFSM );

	std::map<std::pair<automaton::State, alphabet::Symbol>, std::set<automaton::State>> origTransitions = origFSM.getSymbolTransitions();

	/**
	 * Step 1 from Melichar 2.41
	 */
	for( const auto & from : origFSM.getStates( ) )
		for( const auto & fromClosure : closures[from] )
			for( const auto & symbol : origFSM.getInputAlphabet() ) {
				auto transitions = origTransitions.find(std::make_pair(fromClosure, symbol));
				if(transitions == origTransitions.end()) continue;

				for( const auto & to : transitions->second )
					fsm.addTransition( from, symbol, to );
			}

	/**
	 * Step 2 from Melichar 2.41
	 */
	const std::set<automaton::State> & F = origFSM.getFinalStates( );
	for( const auto & q : fsm.getStates( ) )
	{
		const std::set<automaton::State> & cl = closures[ q ];
		std::set<automaton::State> intersect;

		set_intersection( cl.begin(), cl.end(), F.begin(), F.end(), std::inserter( intersect, intersect.begin() ) );
		if( intersect.size( ) != 0 )
			fsm.addFinalState( q );
	}

	return fsm;
}

automaton::Automaton EpsilonRemoverIncoming::remove(const automaton::Automaton& automaton) {
	automaton::Automaton* out = NULL;
	automaton.getData().Accept((void*) &out, EpsilonRemoverIncoming::EPSILON_REMOVER_INCOMING);
	automaton::Automaton res = std::move(*out);
	delete out;
	return res;
}

void EpsilonRemoverIncoming::Visit(void* data, const automaton::EpsilonNFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->remove(automaton));
}

void EpsilonRemoverIncoming::Visit(void* data, const automaton::MultiInitialStateNFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->remove(automaton));
}

void EpsilonRemoverIncoming::Visit(void* data, const automaton::NFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->remove(automaton));
}

void EpsilonRemoverIncoming::Visit(void* data, const automaton::DFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->remove(automaton));
}

void EpsilonRemoverIncoming::Visit(void*, const DFTA&) const {
	throw exception::AlibException("Unsupported automaton type DFTA");
}

void EpsilonRemoverIncoming::Visit(void*, const NFTA&) const {
	throw exception::AlibException("Unsupported automaton type NFTA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::ExtendedNFA& ) const {
	throw exception::AlibException("Unsupported automaton type ExtendedNFA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::CompactNFA& ) const {
	throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::DPDA&) const {
	throw exception::AlibException("Unsupported automaton type DPDA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::InputDrivenDPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenDPDA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::VisiblyPushdownDPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::RealTimeHeightDeterministicDPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::NPDA&) const {
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void EpsilonRemoverIncoming::Visit(void*, const automaton::OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

const EpsilonRemoverIncoming EpsilonRemoverIncoming::EPSILON_REMOVER_INCOMING;

} /* namespace efficient */

} /* namespace simplify */

} /* namespace automaton */
