/*
 * UnreachableStatesRemover.cpp
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "UnreachableStatesRemover.h"

#include <exception/AlibException.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include "../../properties/efficient/ReachableStates.h"

#include "automaton/common/State.h"
#include "automaton/Automaton.h"

namespace automaton {

namespace simplify {

namespace efficient {

template<class T>
T UnreachableStatesRemover::remove( const T & fsm ) {
	// 1a
	std::set<automaton::State> Qa = automaton::properties::efficient::ReachableStates::reachableStates( fsm );

	// 2
	T M(fsm.getInitialState());

	for( const auto & q : Qa )
		M.addState( q );

	for( const auto & a : fsm.getInputAlphabet( ) )
		M.addInputSymbol( a );

	for( const auto & transition : fsm.getTransitions( ) )
		if( Qa.count( transition.first.first ) )
			for(const auto& to : transition.second )
				M.addTransition( transition.first.first, transition.first.second, to );

	std::set<automaton::State> intersect;
	std::set_intersection( fsm.getFinalStates( ).begin(), fsm.getFinalStates( ).end(), Qa.begin( ), Qa.end( ), std::inserter( intersect, intersect.begin( ) ) );
	for( auto const & state : intersect )
		M.addFinalState( state );

	return M;
}

template automaton::EpsilonNFA UnreachableStatesRemover::remove( const automaton::EpsilonNFA & fsm );
template automaton::NFA UnreachableStatesRemover::remove( const automaton::NFA & fsm );
template automaton::CompactNFA UnreachableStatesRemover::remove( const automaton::CompactNFA & fsm );
template automaton::ExtendedNFA UnreachableStatesRemover::remove( const automaton::ExtendedNFA & fsm );

template<>
automaton::DFA UnreachableStatesRemover::remove( const automaton::DFA & fsm ) {
	// 1a
	std::set<automaton::State> Qa = automaton::properties::efficient::ReachableStates::reachableStates( fsm );

	// 2
	automaton::DFA M(fsm.getInitialState() );

	for( const auto & q : Qa )
		M.addState( q );

	for( const auto & a : fsm.getInputAlphabet( ) )
		M.addInputSymbol( a );

	for( const auto & transition : fsm.getTransitions( ) )
		if( Qa.count( transition.first.first ) )
			M.addTransition( transition.first.first, transition.first.second, transition.second );

	std::set<automaton::State> intersect;
	std::set_intersection( fsm.getFinalStates( ).begin(), fsm.getFinalStates( ).end(), Qa.begin( ), Qa.end( ), std::inserter( intersect, intersect.begin( ) ) );
	for( auto const & state : intersect )
		M.addFinalState( state );

	return M;
}

template<>
automaton::MultiInitialStateNFA UnreachableStatesRemover::remove( const automaton::MultiInitialStateNFA & fsm ) {
	// 1a
	std::set<automaton::State> Qa = automaton::properties::efficient::ReachableStates::reachableStates( fsm );

	// 2
	automaton::MultiInitialStateNFA M;

	for( const auto & q : Qa )
		M.addState( q );

	M.setInitialStates( fsm.getInitialStates() );

	for( const auto & a : fsm.getInputAlphabet( ) )
		M.addInputSymbol( a );

	for( const auto & transition : fsm.getTransitions( ) )
		if( Qa.count( transition.first.first ) )
			for(const auto& to : transition.second )
				M.addTransition( transition.first.first, transition.first.second, to );

	std::set<automaton::State> intersect;
	std::set_intersection( fsm.getFinalStates( ).begin(), fsm.getFinalStates( ).end(), Qa.begin( ), Qa.end( ), std::inserter( intersect, intersect.begin( ) ) );
	for( auto const & state : intersect )
		M.addFinalState( state );

	return M;
}

automaton::Automaton UnreachableStatesRemover::remove(const automaton::Automaton& automaton) {
	automaton::Automaton* out = NULL;
	automaton.getData().Accept((void*) &out, UnreachableStatesRemover::UNREACHABLE_STATES_REMOVER);
	automaton::Automaton res = std::move(*out);
	delete out;
	return res;
}

void UnreachableStatesRemover::Visit(void* data, const automaton::EpsilonNFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->remove(automaton));
}

void UnreachableStatesRemover::Visit(void* data, const automaton::MultiInitialStateNFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->remove(automaton));
}

void UnreachableStatesRemover::Visit(void* data, const automaton::NFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->remove(automaton));
}

void UnreachableStatesRemover::Visit(void* data, const automaton::DFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->remove(automaton));
}

void UnreachableStatesRemover::Visit(void* data, const automaton::ExtendedNFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->remove(automaton));
}

void UnreachableStatesRemover::Visit(void* data, const automaton::CompactNFA& automaton) const {
	automaton::Automaton* & out = *((automaton::Automaton**) data);
	out = new automaton::Automaton(this->remove(automaton));
}

void UnreachableStatesRemover::Visit(void*, const DFTA&) const {
	throw exception::AlibException("Unsupported automaton type DFTA");
}

void UnreachableStatesRemover::Visit(void*, const NFTA&) const {
	throw exception::AlibException("Unsupported automaton type NFTA");
}

void UnreachableStatesRemover::Visit(void*, const automaton::DPDA&) const {
	throw exception::AlibException("Unsupported automaton type DPDA");
}

void UnreachableStatesRemover::Visit(void*, const automaton::SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void UnreachableStatesRemover::Visit(void*, const automaton::InputDrivenDPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenDPDA");
}

void UnreachableStatesRemover::Visit(void*, const automaton::InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void UnreachableStatesRemover::Visit(void*, const automaton::VisiblyPushdownDPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void UnreachableStatesRemover::Visit(void*, const automaton::VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void UnreachableStatesRemover::Visit(void*, const automaton::RealTimeHeightDeterministicDPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void UnreachableStatesRemover::Visit(void*, const automaton::RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void UnreachableStatesRemover::Visit(void*, const automaton::NPDA&) const {
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void UnreachableStatesRemover::Visit(void*, const automaton::SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void UnreachableStatesRemover::Visit(void*, const automaton::OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

const UnreachableStatesRemover UnreachableStatesRemover::UNREACHABLE_STATES_REMOVER;

} /* namespace efficient */

} /* namespace simplify */

} /* namespace automaton */
