/*
 * aminimize.cpp
 *
 *  Created on: 24. 2. 2014
 *      Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>

#include <exception/AlibException.h>
#include <factory/XmlDataFactory.hpp>

#include "automaton/simplify/Minimize.h"
#include "automaton/simplify/MinimizeBrzozowski.h"

int main(int argc, char** argv) {

	try {
		TCLAP::CmdLine cmd("Automaton minimize binary", ' ', "0.01");

		TCLAP::ValueArg<std::string> input(	"i",	"input",	"Automaton to minimize",	false,	"-",		"file");
		cmd.add( input );

		std::vector<std::string> algorithms {"hopcroft", "brzozowski" };
		TCLAP::ValuesConstraint<std::string> algorithmVals( algorithms );
		TCLAP::ValueArg<std::string> algorithm(	"a",	"algorithm",	"Use specific algorithm. Hopcroft default.",	false,	"hopcroft",	&algorithmVals);
		cmd.add( algorithm );

		TCLAP::SwitchArg measure(		"m",	"measure",	"Measure times",		false);
		cmd.add( measure );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc, argv);

		std::chrono::measurements::start("Overal", std::chrono::measurements::Type::OVERALL);
		std::chrono::measurements::start("Input read", std::chrono::measurements::Type::AUXILARY);

		std::deque<sax::Token> tokens;
		if(input.isSet()) {
			if(input.getValue() == "-") {
				sax::SaxParseInterface::parseStdin(tokens);
			} else {
				sax::SaxParseInterface::parseFile(input.getValue(), tokens);
			}
		} else {
			sax::SaxParseInterface::parseStdin(tokens);
		}

		automaton::Automaton automaton = alib::XmlDataFactory::fromTokens<automaton::Automaton>(tokens);

		std::chrono::measurements::end();
		std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

		if(algorithm.getValue() == "hopcroft") {
			automaton::Automaton res = automaton::simplify::Minimize::minimize(automaton);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout(res);
		} else if(algorithm.getValue() == "brzozowski") {
			automaton::Automaton res = automaton::simplify::MinimizeBrzozowski::minimize(automaton);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout(res);
		} else {
			throw exception::AlibException("Unknown algorithm selected.");
		}

		std::chrono::measurements::end();
		std::chrono::measurements::end();

		if(measure.getValue()) std::clog << std::chrono::measurements::results() << std::endl;

		return 0;
	} catch (const exception::AlibException& exception) {
		alib::XmlDataFactory::toStdout(exception);
		return 1;
	} catch (const TCLAP::ArgException& exception) {
		std::cout << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch (...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
