/*
 * arun.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>
#include <vector>

#include <factory/XmlDataFactory.hpp>
#include <exception/AlibException.h>
#include <string/String.h>
#include <primitive/Bool.h>
#include <object/Object.h>
#include <tree/ranked/RankedTree.h>
#include <automaton/Automaton.h>
#include <automaton/run/Accept.h>
#include <automaton/run/Result.h>
#include <automaton/run/Occurrences.h>
#include <container/ObjectsSet.h>

#include <iostream>

int main(int argc, char* argv[]) {
	try {
		TCLAP::CmdLine cmd("Automaton run binary", ' ', "0.01");

		std::vector<std::string> allowed;
		allowed.push_back("occurrences");
		allowed.push_back("accept");
		allowed.push_back("result");
		TCLAP::ValuesConstraint<std::string> allowedVals( allowed );

		TCLAP::ValueArg<std::string> type(	"t",	"type",		"Run type",		false,	"accept",	&allowedVals);
		cmd.add( type );

		TCLAP::ValueArg<std::string> input(	"i",	"input",	"Input from file",	false,	"-",		"file");
		cmd.add( input );

		TCLAP::ValueArg<std::string> automaton(	"a",	"automaton",	"Automaton to run",	false,	"-",		"file");
		cmd.add( automaton );

		TCLAP::SwitchArg measure(		"m",	"measure",	"Measure times",		false);
		cmd.add( measure );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc, argv);

		std::chrono::measurements::start("Overal", std::chrono::measurements::Type::OVERALL);
		std::chrono::measurements::start("Input read", std::chrono::measurements::Type::AUXILARY);

		std::deque<sax::Token> inputTokens;
		if(input.isSet()) {
			if(input.getValue() == "-") {
				sax::SaxParseInterface::parseStdin(inputTokens);
			} else {
				sax::SaxParseInterface::parseFile(input.getValue(), inputTokens);
			}
		} else {
			sax::SaxParseInterface::parseStdin(inputTokens);
		}

		std::deque<sax::Token> automatonTokens;
		if(automaton.isSet()) {
			if(automaton.getValue() == "-") {
				sax::SaxParseInterface::parseStdin(automatonTokens);
			} else {
				sax::SaxParseInterface::parseFile(automaton.getValue(), automatonTokens);
			}
		} else {
			sax::SaxParseInterface::parseStdin(automatonTokens);
		}

		alib::Object inputData = alib::XmlDataFactory::fromTokens<alib::Object>(inputTokens);
		automaton::Automaton automatonData = alib::XmlDataFactory::fromTokens<automaton::Automaton>(automatonTokens);

		std::chrono::measurements::end();
		std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

		if( type.getValue() == "occurrences") {
			std::set<unsigned> res = automaton::run::Occurrences::occurrences(automatonData, inputData);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout( res );
		} else if( type.getValue() == "accept") {
			bool res = automaton::run::Accept::accept(automatonData, inputData);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout( res );
		} else if( type.getValue() == "result") {
			label::Label res = automaton::run::Result::result(automatonData, inputData);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout( res );
		} else {
			throw exception::AlibException( "Invalid run type" );
		}

		std::chrono::measurements::end();
		std::chrono::measurements::end();

		if(measure.getValue()) std::clog << std::chrono::measurements::results() << std::endl;

		return 0;
	} catch ( const exception::AlibException & exception ) {
		alib::XmlDataFactory::toStdout( exception );
		return 1;
	} catch (const TCLAP::ArgException& exception) {
		std::cout << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch (...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
