/*
 * astringology.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>
#include <vector>

#include <factory/XmlDataFactory.hpp>
#include <exception/AlibException.h>
#include <string/String.h>
#include <automaton/Automaton.h>
#include <container/Container.h>
#include <container/ObjectsSet.h>
#include <container/ObjectsVector.h>

#include <stringology/exact/ExactFactorMatch.h>
#include <stringology/exact/BoyerMooreHorspool.h>
#include <stringology/exact/ExactMatchingAutomaton.h>
#include <stringology/exact/ExactFactorAutomaton.h>
#include <stringology/exact/ExactSubsequenceAutomaton.h>
#include <stringology/exact/ExactNondeterministicSubsequenceAutomaton.h>
#include <stringology/exact/ExactMultiNondeterministicSubsequenceAutomaton.h>
#include <stringology/exact/BorderArray.h>

int main(int argc, char* argv[]) {
	try {
		TCLAP::CmdLine cmd("Stringology algorithm access binary", ' ', "0.01");

		std::vector<std::string> allowed;
		allowed.push_back("exactMatchingAutomaton");
		allowed.push_back("exactFactorAutomaton");
		allowed.push_back("exactSubsequenceAutomaton");
		allowed.push_back("exactNondeterministicSubsequenceAutomaton");
		allowed.push_back("exactMultiNondeterministicSubsequenceAutomaton");
		allowed.push_back("exactFactorMatch");
		allowed.push_back("boyerMooreHorspool");
		allowed.push_back("borderArray");
		TCLAP::ValuesConstraint<std::string> allowedVals( allowed );

		TCLAP::ValueArg<std::string> algorithm(	"a",	"algorithm",	"Execute algorithm",		false,	"exactFactorMatch",	&allowedVals);
		cmd.add(algorithm);

		TCLAP::MultiArg<std::string> subject(	"s",	"subject",	"Subject string from file",	false,	"file");
		cmd.add( subject );

		TCLAP::MultiArg<std::string> pattern(	"p",	"pattern",	"Pattern string from file",	false,	"file");
		cmd.add( pattern );

		TCLAP::SwitchArg measure(		"m",	"measure",	"Measure times",		false);
		cmd.add( measure );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc,argv);

		int needPattern = 0;
		int needSubject = 0;
		if( algorithm.getValue() == "exactFactorMatch") {
			needPattern = needSubject = 1;
		} else if( algorithm.getValue() == "boyerMooreHorspool") {
			needPattern = needSubject = 1;
		} else if( algorithm.getValue() == "exactMatchingAutomaton") {
			needPattern = 1;
		} else if( algorithm.getValue() == "exactFactorAutomaton") {
			needSubject = 1;
		} else if( algorithm.getValue() == "exactSubsequenceAutomaton") {
			needSubject = 1;
		} else if( algorithm.getValue() == "exactNondeterministicSubsequenceAutomaton") {
			needSubject = 1;
		} else if( algorithm.getValue() == "exactMultiNondeterministicSubsequenceAutomaton") {
			needSubject = 2;
		} else if( algorithm.getValue() == "borderArray") {
			needSubject = 1;
		} else {
		}

		std::chrono::measurements::start("Overal", std::chrono::measurements::Type::OVERALL);
		std::chrono::measurements::start("Input read", std::chrono::measurements::Type::AUXILARY);

		std::deque<std::deque<sax::Token>> subjectTokens;
		if(subject.isSet()) {
			for(const std::string& fileName : subject.getValue()) {
				std::deque<sax::Token> tmp;
				if(fileName == "-") {
					sax::SaxParseInterface::parseStdin(tmp);
				} else {
					sax::SaxParseInterface::parseFile(fileName, tmp);
				}
				subjectTokens.emplace_back(std::move(tmp));
			}
		} else if(needSubject) {
			std::deque<sax::Token> tmp;
			sax::SaxParseInterface::parseStdin(tmp);
			subjectTokens.emplace_back(std::move(tmp));
		}

		std::deque<std::deque<sax::Token>> patternTokens;
		if(pattern.isSet()) {
			for(const std::string& fileName : pattern.getValue()) {
				std::deque<sax::Token> tmp;
				if(fileName == "-") {
					sax::SaxParseInterface::parseStdin(tmp);
				} else {
					sax::SaxParseInterface::parseFile(fileName, tmp);
				}
				patternTokens.emplace_back(std::move(tmp));
			}
		} else if(needPattern) {
			std::deque<sax::Token> tmp;
			sax::SaxParseInterface::parseStdin(tmp);
			patternTokens.emplace_back(std::move(tmp));
		}

		if( algorithm.getValue() == "exactFactorMatch") {
			string::String subject = alib::XmlDataFactory::fromTokens<string::String>(subjectTokens.front());
			string::String pattern = alib::XmlDataFactory::fromTokens<string::String>(patternTokens.front());

			std::chrono::measurements::end();
			std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

			std::set<unsigned> res = stringology::exact::ExactFactorMatch::match(subject, pattern);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout(res);
		} else if( algorithm.getValue() == "boyerMooreHorspool") {
			string::String subject = alib::XmlDataFactory::fromTokens<string::String>(subjectTokens.front());
			string::String pattern = alib::XmlDataFactory::fromTokens<string::String>(patternTokens.front());

			std::chrono::measurements::end();
			std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

			std::set<unsigned> res = stringology::exact::BoyerMooreHorspool::match(subject, pattern);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout(res);
		} else if( algorithm.getValue() == "exactMatchingAutomaton") {
			string::String pattern = alib::XmlDataFactory::fromTokens<string::String>(patternTokens.front());

			std::chrono::measurements::end();
			std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

			automaton::Automaton automaton = stringology::exact::ExactMatchingAutomaton::construct(pattern);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout(automaton);
		} else if( algorithm.getValue() == "exactFactorAutomaton") {
			string::String subject = alib::XmlDataFactory::fromTokens<string::String>(subjectTokens.front());

			std::chrono::measurements::end();
			std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

			automaton::Automaton automaton = stringology::exact::ExactFactorAutomaton::construct(subject);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout(automaton);
		} else if( algorithm.getValue() == "exactSubsequenceAutomaton") {
			string::String subject = alib::XmlDataFactory::fromTokens<string::String>(subjectTokens.front());

			std::chrono::measurements::end();
			std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

			automaton::Automaton automaton = stringology::exact::ExactSubsequenceAutomaton::construct(subject);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout(automaton);
		} else if( algorithm.getValue() == "exactNondeterministicSubsequenceAutomaton") {
			string::String subject = alib::XmlDataFactory::fromTokens<string::String>(subjectTokens.front());

			std::chrono::measurements::end();
			std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

			automaton::Automaton automaton = stringology::exact::ExactNondeterministicSubsequenceAutomaton::construct(subject);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout(automaton);
		} else if( algorithm.getValue() == "exactMultiNondeterministicSubsequenceAutomaton") {
			std::set<string::String> subjects;
			for(std::deque<sax::Token>& tokens : subjectTokens) {
				subjects.insert(alib::XmlDataFactory::fromTokens<string::String>(tokens));
			}

			std::chrono::measurements::end();
			std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

			automaton::Automaton automaton = stringology::exact::ExactMultiNondeterministicSubsequenceAutomaton::construct(subjects);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout(automaton);
		} else if( algorithm.getValue() == "borderArray") {
			string::String subject = alib::XmlDataFactory::fromTokens<string::String>(subjectTokens.front());

			std::chrono::measurements::end();
			std::chrono::measurements::start("Algorithm", std::chrono::measurements::Type::MAIN);

			std::vector<unsigned> borderArray = stringology::exact::BorderArray::construct(subject);

			std::chrono::measurements::end();
			std::chrono::measurements::start("Output write", std::chrono::measurements::Type::AUXILARY);

			alib::XmlDataFactory::toStdout(borderArray);
		} else {
			throw exception::AlibException( "Invalid algorithm" );
		}

		std::chrono::measurements::end();
		std::chrono::measurements::end();

		if(measure.getValue()) std::clog << std::chrono::measurements::results() << std::endl;

		return 0;
	} catch( const exception::AlibException & exception ) {
		alib::XmlDataFactory::toStdout( exception );
		return 1;
	} catch(const TCLAP::ArgException& exception) {
		std::cout << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch(...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
